/* Ofront 1.4 -xtspkaG */
#include "SYSTEM.h"
#include "Display.h"
#include "Files.h"
#include "Input.h"
#include "Math.h"
#include "Oberon.h"
#include "Printer.h"
#include "TextFrames.h"
#include "TextPrinter.h"
#include "Texts.h"
#include "Viewers.h"

struct ClockElems__1 { /* Display_FrameDesc */
	Display_Frame dsc, next;
	INTEGER X, Y, W, H;
	Display_Handler handle;
	SHORTINT col;
};

typedef
	struct ClockElems__1 *ClockElems_Frame;

typedef
	struct ClockElems_NotifyMsg { /* Display_FrameMsg */
		char _prvt0[1];
	} ClockElems_NotifyMsg;

typedef
	struct ClockElems_Time {
		INTEGER sec, min, hour;
		LONGINT timeStamp;
	} ClockElems_Time;


static REAL ClockElems_sin[60], ClockElems_cos[60];
static LONGINT ClockElems_wakeUp;
static ClockElems_Time ClockElems_old, ClockElems_new;
static Oberon_Task ClockElems_Task;
static void (*ClockElems_Line)(INTEGER, INTEGER, INTEGER, INTEGER, INTEGER, INTEGER);
static void (*ClockElems_Circle)(INTEGER, INTEGER, INTEGER, INTEGER, INTEGER);

export LONGINT *ClockElems_Time__typ;
export LONGINT *ClockElems__1__typ;
export LONGINT *ClockElems_NotifyMsg__typ;

static void ClockElems_Clock (void);
static void ClockElems_Draw (ClockElems_Time *time, LONGINT *time__typ, INTEGER x0, INTEGER y0, INTEGER r, INTEGER color);
static void ClockElems_GetTime (ClockElems_Time *time, LONGINT *time__typ);
static void ClockElems_HandleElem (Texts_Elem E, Texts_ElemMsg *msg, LONGINT *msg__typ);
static void ClockElems_HandleFrame (Display_Frame F, Display_FrameMsg *M, LONGINT *M__typ);
static void ClockElems_Init (void);
export void ClockElems_Insert (void);
static void ClockElems_Line2 (INTEGER ang, INTEGER x0, INTEGER y0, INTEGER r1, INTEGER r2, INTEGER color);
export void ClockElems_New (void);
static void ClockElems_PCircle (INTEGER x0, INTEGER y0, INTEGER r, INTEGER color, INTEGER mode);
static void ClockElems_PLine (INTEGER x1, INTEGER y1, INTEGER x2, INTEGER y2, INTEGER color, INTEGER mode);
static void ClockElems_SCircle (INTEGER x0, INTEGER y0, INTEGER r, INTEGER color, INTEGER mode);
static void ClockElems_SLine (INTEGER x1, INTEGER y1, INTEGER x2, INTEGER y2, INTEGER color, INTEGER mode);
static void ClockElems_Update (ClockElems_Time *old, LONGINT *old__typ, ClockElems_Time *new, LONGINT *new__typ, INTEGER x0, INTEGER y0, INTEGER r, INTEGER color);


static void ClockElems_Init (void)
{
	INTEGER i = 0;
	i = 0;
	while (i < 60) {
		ClockElems_sin[__X(i, 60)] = Math_sin(  1.0471976e-001 * i);
		ClockElems_cos[__X(i, 60)] = Math_cos(  1.0471976e-001 * i);
		i += 1;
	}
}

static void ClockElems_GetTime (ClockElems_Time *time, LONGINT *time__typ)
{
	LONGINT t = 0, d = 0;
	Oberon_GetClock(&t, &d);
	(*time).sec = (int)__MASK(t, -64);
	(*time).min = (int)__MASK(__ASHR(t, 6), -64);
	(*time).hour = (int)__ASHR(t, 12) * 5 + __DIV((*time).min, 12);
	(*time).timeStamp = t;
}

static struct SCircle__12 {
	struct SCircle__12 *lnk;
} *SCircle__12_s;

static void Dot4__13 (INTEGER x1, INTEGER x2, INTEGER y1, INTEGER y2, INTEGER color, INTEGER mode);

static void Dot4__13 (INTEGER x1, INTEGER x2, INTEGER y1, INTEGER y2, INTEGER color, INTEGER mode)
{
	Display_Dot(color, x1, y1, mode);
	Display_Dot(color, x1, y2, mode);
	Display_Dot(color, x2, y1, mode);
	Display_Dot(color, x2, y2, mode);
}

static void ClockElems_SCircle (INTEGER x0, INTEGER y0, INTEGER r, INTEGER color, INTEGER mode)
{
	INTEGER x = 0, y = 0, dx = 0, dy = 0, d = 0;
	struct SCircle__12 _s;
	_s.lnk = SCircle__12_s;
	SCircle__12_s = &_s;
	x = r;
	y = 0;
	dx = __ASHL(x - 1, 3);
	dy = __ASHL(y, 3) + 4;
	d = 1 - __ASHL(r, 2);
	while (x > y) {
		Dot4__13(x0 - x, x0 + x, y0 - y, y0 + y, color, mode);
		Dot4__13(x0 - y, x0 + y, y0 - x, y0 + x, color, mode);
		d += dy;
		dy += 8;
		y += 1;
		if (d >= 0) {
			d -= dx;
			dx -= 8;
			x -= 1;
		}
	}
	if (x == y) {
		Dot4__13(x0 - x, x0 + x, y0 - y, y0 + y, color, mode);
	}
	SCircle__12_s = _s.lnk;
}

static void ClockElems_SLine (INTEGER x1, INTEGER y1, INTEGER x2, INTEGER y2, INTEGER color, INTEGER mode)
{
	INTEGER x = 0, y = 0, dx = 0, dy = 0, d = 0, inc = 0;
	if (y2 - y1 < x1 - x2) {
		x = x1;
		x1 = x2;
		x2 = x;
		y = y1;
		y1 = y2;
		y2 = y;
	}
	dx = __ASHL(x2 - x1, 1);
	dy = __ASHL(y2 - y1, 1);
	x = x1;
	y = y1;
	inc = 1;
	if (dy > dx) {
		d = __ASHR(dy, 1);
		if (dx < 0) {
			inc = -1;
			dx = -dx;
		}
		while (y <= y2) {
			Display_Dot(color, x, y, mode);
			y += 1;
			d -= dx;
			if (d < 0) {
				d += dy;
				x += inc;
			}
		}
	} else {
		d = __ASHR(dx, 1);
		if (dy < 0) {
			inc = -1;
			dy = -dy;
		}
		while (x <= x2) {
			Display_Dot(color, x, y, mode);
			x += 1;
			d -= dy;
			if (d < 0) {
				d += dx;
				y += inc;
			}
		}
	}
}

static void ClockElems_PCircle (INTEGER x0, INTEGER y0, INTEGER r, INTEGER color, INTEGER mode)
{
	Printer_Circle(x0, y0, r);
}

static void ClockElems_PLine (INTEGER x1, INTEGER y1, INTEGER x2, INTEGER y2, INTEGER color, INTEGER mode)
{
	Printer_Line(x1, y1, x2, y2);
}

static void ClockElems_Line2 (INTEGER ang, INTEGER x0, INTEGER y0, INTEGER r1, INTEGER r2, INTEGER color)
{
	INTEGER x1 = 0, y1 = 0, x2 = 0, y2 = 0;
	ang = (int)__MOD(15 - ang, 60);
	x1 = (int)__ENTIER(r1 * ClockElems_cos[__X(ang, 60)] +   5.0000000e-001);
	y1 = (int)__ENTIER(r1 * ClockElems_sin[__X(ang, 60)] +   5.0000000e-001);
	x2 = (int)__ENTIER(r2 * ClockElems_cos[__X(ang, 60)] +   5.0000000e-001);
	y2 = (int)__ENTIER(r2 * ClockElems_sin[__X(ang, 60)] +   5.0000000e-001);
	(*ClockElems_Line)(x0 + x1, y0 + y1, x0 + x2, y0 + y2, color, 2);
}

static void ClockElems_Draw (ClockElems_Time *time, LONGINT *time__typ, INTEGER x0, INTEGER y0, INTEGER r, INTEGER color)
{
	INTEGER rh = 0, rm = 0, rs = 0, i = 0;
	if (r >= 12) {
		rh = __DIV(7 * r, 11);
		rm = __DIV(9 * r, 11);
		rs = __DIV(10 * r, 11);
		i = 0;
		while (i < 60) {
			ClockElems_Line2(i, x0, y0, rm, r, color);
			i += 5;
		}
		ClockElems_Line2((*time).sec, x0, y0, rm - r, rs, color);
		ClockElems_Line2((*time).min, x0, y0, 0, rm, color);
		ClockElems_Line2((*time).hour, x0, y0, 0, rh, color);
		(*ClockElems_Circle)(x0, y0, 2, color, 0);
	}
	(*ClockElems_Circle)(x0, y0, r, color, 0);
}

static void ClockElems_Update (ClockElems_Time *old, LONGINT *old__typ, ClockElems_Time *new, LONGINT *new__typ, INTEGER x0, INTEGER y0, INTEGER r, INTEGER color)
{
	INTEGER rh = 0, rm = 0, rs = 0;
	if (r >= 12) {
		rh = __DIV(7 * r, 11);
		rm = __DIV(9 * r, 11);
		rs = __DIV(10 * r, 11);
		if ((*old).sec != (*new).sec) {
			ClockElems_Line2((*old).sec, x0, y0, rm - r, rs, color);
			ClockElems_Line2((*new).sec, x0, y0, rm - r, rs, color);
		}
		if ((*old).min != (*new).min) {
			ClockElems_Line2((*old).min, x0, y0, 0, rm, color);
			ClockElems_Line2((*new).min, x0, y0, 0, rm, color);
		}
		if ((*old).hour != (*new).hour) {
			ClockElems_Line2((*old).hour, x0, y0, 0, rh, color);
			ClockElems_Line2((*new).hour, x0, y0, 0, rh, color);
		}
	}
}

static void ClockElems_HandleFrame (Display_Frame F, Display_FrameMsg *M, LONGINT *M__typ)
{
	INTEGER r = 0;
	if (__IS(M__typ, ClockElems_NotifyMsg, 1)) {
		ClockElems_Line = ClockElems_SLine;
		ClockElems_Circle = ClockElems_SCircle;
		r = __ASHR(F->W, 1);
		ClockElems_Update(&ClockElems_old, ClockElems_Time__typ, &ClockElems_new, ClockElems_Time__typ, F->X + r, F->Y + r, r, __GUARDA(F, ClockElems__1, 1)->col);
	} else if (__IS(M__typ, Oberon_InputMsg, 1)) {
		if (__IS(M__typ, Oberon_InputMsg, 1)) {
			Oberon_InputMsg *M__ = (void*)M;
			if ((*M__).id == 1) {
				Oberon_DrawCursor(&Oberon_Mouse, Oberon_Cursor__typ, &Oberon_Arrow, Oberon_Marker__typ, (*M__).X, (*M__).Y);
			}
		} else __WITHCHK;
	}
}

static void ClockElems_HandleElem (Texts_Elem E, Texts_ElemMsg *msg, LONGINT *msg__typ)
{
	Texts_Elem e = NIL;
	ClockElems_Frame f = NIL;
	INTEGER r = 0;
	CHAR ch = 0;
	if (__IS(msg__typ, TextFrames_DisplayMsg, 1)) {
		if (__IS(msg__typ, TextFrames_DisplayMsg, 1)) {
			TextFrames_DisplayMsg *msg__ = (void*)msg;
			if (!(*msg__).prepare) {
				ClockElems_Line = ClockElems_SLine;
				ClockElems_Circle = ClockElems_SCircle;
				r = (int)__ASHR(__DIV(E->W, (LONGINT)TextFrames_Unit) - 1, 1);
				ClockElems_Draw(&ClockElems_new, ClockElems_Time__typ, (*msg__).X0 + r, (*msg__).Y0 + r, r, (*msg__).col);
				__NEW(f, ClockElems__1);
				f->X = (*msg__).X0;
				f->Y = (*msg__).Y0;
				f->W = __ASHL(r, 1) + 1;
				f->H = f->W;
				f->handle = ClockElems_HandleFrame;
				f->col = (*msg__).col;
				(*msg__).elemFrame = (Display_Frame)f;
			}
		} else __WITHCHK;
	} else if (__IS(msg__typ, TextPrinter_PrintMsg, 1)) {
		if (__IS(msg__typ, TextPrinter_PrintMsg, 1)) {
			TextPrinter_PrintMsg *msg__ = (void*)msg;
			if (!(*msg__).prepare) {
				ClockElems_Line = ClockElems_PLine;
				ClockElems_Circle = ClockElems_PCircle;
				r = (int)__ASHR(__DIV(E->W, 3048) - 1, 1);
				ClockElems_Draw(&ClockElems_new, ClockElems_Time__typ, (*msg__).X0 + r, (*msg__).Y0 + r, r, (*msg__).col);
			}
		} else __WITHCHK;
	} else if (__IS(msg__typ, Texts_CopyMsg, 1)) {
		__NEW(e, Texts_ElemDesc);
		Texts_CopyElem(E, e);
		__GUARDR(msg, Texts_CopyMsg, 1).e = e;
	} else if (__IS(msg__typ, Texts_IdentifyMsg, 1)) {
		if (__IS(msg__typ, Texts_IdentifyMsg, 1)) {
			Texts_IdentifyMsg *msg__ = (void*)msg;
			__MOVE("ClockElems", (*msg__).mod, 11);
			__MOVE("New", (*msg__).proc, 4);
		} else __WITHCHK;
	} else if (__IS(msg__typ, Texts_FileMsg, 1)) {
		if (__IS(msg__typ, Texts_FileMsg, 1)) {
			Texts_FileMsg *msg__ = (void*)msg;
			if ((*msg__).id == 0) {
				Files_Read(&(*msg__).r, Files_Rider__typ, (void*)&ch);
			} else if ((*msg__).id == 1) {
				Files_Write(&(*msg__).r, Files_Rider__typ, 0x00);
			}
		} else __WITHCHK;
	}
}

static void ClockElems_Clock (void)
{
	ClockElems_NotifyMsg msg = {0};
	ClockElems_old = ClockElems_new;
	ClockElems_GetTime(&ClockElems_new, ClockElems_Time__typ);
	if (ClockElems_old.timeStamp != ClockElems_new.timeStamp) {
		ClockElems_Task->time = Input_Time() + Input_TimeUnit;
		Viewers_Broadcast((void*)&msg, ClockElems_NotifyMsg__typ);
	} else {
		ClockElems_Task->time = Input_Time() + 10;
	}
}

void ClockElems_New (void)
{
	__NEW(Texts_new, Texts_ElemDesc);
	Texts_new->handle = ClockElems_HandleElem;
}

void ClockElems_Insert (void)
{
	Texts_Scanner S = {0};
	LONGINT w = 0;
	Texts_Elem E = NIL;
	TextFrames_InsertElemMsg M = {0};
	Texts_OpenScanner(&S, Texts_Scanner__typ, Oberon_Par->text, Oberon_Par->pos);
	Texts_Scan(&S, Texts_Scanner__typ);
	if ((S.line == 0 && S.class == 3) && S.i > 0) {
		w = (__ASHL(S.i, 1) + 1) * Display_Unit;
	} else {
		w = 65 * Display_Unit;
	}
	__NEW(E, Texts_ElemDesc);
	E->W = w;
	E->H = w;
	E->handle = ClockElems_HandleElem;
	M.e = E;
	Viewers_Broadcast((void*)&M, TextFrames_InsertElemMsg__typ);
}

static void EnumPtrs(void (*P)(void*))
{
	P(ClockElems_Task);
}

__TDESC(ClockElems_Time, 1, 0) = {__TDFLDS("Time", 12), {-4}};
__TDESC(ClockElems__1, 1, 2) = {__TDFLDS("", 24), {0, 4, -12}};
__TDESC(ClockElems_NotifyMsg, 1, 0) = {__TDFLDS("NotifyMsg", 1), {-4}};

export void *ClockElems__init(void)
{
	__DEFMOD;
	__IMPORT(Display);
	__IMPORT(Files);
	__IMPORT(Input);
	__IMPORT(Math);
	__IMPORT(Oberon);
	__IMPORT(Printer);
	__IMPORT(TextFrames);
	__IMPORT(TextPrinter);
	__IMPORT(Texts);
	__IMPORT(Viewers);
	__REGMOD(ClockElems, EnumPtrs);
	__REGCMD("Insert", ClockElems_Insert);
	__REGCMD("New", ClockElems_New);
	__INITYP(ClockElems_Time, ClockElems_Time, 0);
	__INITYP(ClockElems__1, Display_FrameDesc, 1);
	__INITYP(ClockElems_NotifyMsg, Display_FrameMsg, 1);
/* BEGIN */
	ClockElems_Init();
	ClockElems_wakeUp = 0;
	ClockElems_GetTime(&ClockElems_new, ClockElems_Time__typ);
	__NEW(ClockElems_Task, Oberon_TaskDesc);
	ClockElems_Task->safe = 1;
	ClockElems_Task->time = 0;
	ClockElems_Task->handle = ClockElems_Clock;
	Oberon_Install(ClockElems_Task);
	__ENDMOD;
}
