/* Ofront 1.4 -xtspkaG */
#include "SYSTEM.h"
#include "Dialogs.h"
#include "Display.h"
#include "Files.h"
#include "Input.h"
#include "MenuViewers.h"
#include "Oberon.h"
#include "TextFrames.h"
#include "Texts.h"
#include "Viewers.h"

typedef
	struct DialogFrames_FrameDesc *DialogFrames_Frame;

typedef
	struct DialogFrames_FrameDesc { /* Display_FrameDesc */
		Display_Frame dsc, next;
		INTEGER X, Y, W, H;
		Display_Handler handle;
		INTEGER col;
		Dialogs_Panel panel;
		INTEGER grid;
		LONGINT pat;
	} DialogFrames_FrameDesc;

static void DialogFrames_Frame_DrawObject (DialogFrames_Frame f, Dialogs_Object o, BOOLEAN drawmode);
static void DialogFrames_Frame_Extend (DialogFrames_Frame f, INTEGER newY);
static void DialogFrames_Frame_MarkMenu (DialogFrames_Frame f);
static void DialogFrames_Frame_Modify (DialogFrames_Frame f, INTEGER id, INTEGER dY, INTEGER y, INTEGER h);
export void DialogFrames_Frame_Open (DialogFrames_Frame f, Display_Handler handle, Dialogs_Panel p);
static void DialogFrames_Frame_Reduce (DialogFrames_Frame f, INTEGER newY);
export void DialogFrames_Frame_Restore (DialogFrames_Frame f);
static void DialogFrames_Frame_Send (DialogFrames_Frame f, INTEGER x, INTEGER y, Display_FrameMsg *m, LONGINT *m__typ, BOOLEAN *cond);
static void DialogFrames_Frame_TrackMouse (DialogFrames_Frame f, INTEGER x, INTEGER y, SET keys);
#define __DialogFrames_Frame_DrawObject(f, o, drawmode) __SEND(__TYPEOF(f), 0, void(*)(DialogFrames_Frame, Dialogs_Object, BOOLEAN), (f, o, drawmode))
#define __DialogFrames_Frame_Extend(f, newY) __SEND(__TYPEOF(f), 1, void(*)(DialogFrames_Frame, INTEGER), (f, newY))
#define __DialogFrames_Frame_MarkMenu(f) __SEND(__TYPEOF(f), 2, void(*)(DialogFrames_Frame), (f))
#define __DialogFrames_Frame_Modify(f, id, dY, y, h) __SEND(__TYPEOF(f), 3, void(*)(DialogFrames_Frame, INTEGER, INTEGER, INTEGER, INTEGER), (f, id, dY, y, h))
#define __DialogFrames_Frame_Open(f, handle, p) __SEND(__TYPEOF(f), 4, void(*)(DialogFrames_Frame, Display_Handler, Dialogs_Panel), (f, handle, p))
#define __DialogFrames_Frame_Reduce(f, newY) __SEND(__TYPEOF(f), 5, void(*)(DialogFrames_Frame, INTEGER), (f, newY))
#define __DialogFrames_Frame_Restore(f) __SEND(__TYPEOF(f), 6, void(*)(DialogFrames_Frame), (f))
#define __DialogFrames_Frame_Send(f, x, y, m, m__typ, cond) __SEND(__TYPEOF(f), 7, void(*)(DialogFrames_Frame, INTEGER, INTEGER, Display_FrameMsg*, LONGINT *, BOOLEAN*), (f, x, y, m, m__typ, cond))
#define __DialogFrames_Frame_TrackMouse(f, x, y, keys) __SEND(__TYPEOF(f), 8, void(*)(DialogFrames_Frame, INTEGER, INTEGER, SET), (f, x, y, keys))

typedef
	struct DialogFrames_GetFrameMsg { /* Display_FrameMsg */
		Dialogs_Panel p;
		DialogFrames_Frame f;
	} DialogFrames_GetFrameMsg;

typedef
	struct DialogFrames_SetCaretMsg { /* Display_FrameMsg */
		Dialogs_Panel p;
		INTEGER x, y;
	} DialogFrames_SetCaretMsg;


static Texts_Writer DialogFrames_w0;
static INTEGER DialogFrames_left, DialogFrames_right, DialogFrames_top, DialogFrames_bot;

export LONGINT *DialogFrames_FrameDesc__typ;
export LONGINT *DialogFrames_GetFrameMsg__typ;
export LONGINT *DialogFrames_SetCaretMsg__typ;

export void DialogFrames_FindObject (Dialogs_Object *o, Dialogs_Panel *p);
export void DialogFrames_GetCaretPosition (Dialogs_Panel *p, INTEGER *xpos, INTEGER *ypos);
export void DialogFrames_Handle (Display_Frame f, Display_FrameMsg *m, LONGINT *m__typ);
static INTEGER DialogFrames_Min (INTEGER x, INTEGER y);
export void DialogFrames_OpenPanel (CHAR *name, LONGINT name__len, INTEGER x, INTEGER y, Dialogs_Panel *p);
export void DialogFrames_SetCaretAtObject (Dialogs_Object o);
static void DialogFrames_box (Dialogs_Object obj, BOOLEAN *done);


static INTEGER DialogFrames_Min (INTEGER x, INTEGER y)
{
	if (x < y) {
		return x;
	} else {
		return y;
	}
	__RETCHK;
}

static void DialogFrames_Frame_MarkMenu (DialogFrames_Frame f)
{
	Texts_Reader r = {0};
	Viewers_Viewer v = NIL;
	Texts_Text t = NIL;
	CHAR ch = 0;
	v = Viewers_This(f->X, f->Y);
	if ((__ISP(v, MenuViewers_ViewerDesc, 2) && __ISP(v->dsc, TextFrames_FrameDesc, 1)) && f != (void *) v->dsc) {
		t = __GUARDP(v->dsc, TextFrames_FrameDesc, 1)->text;
		if (t->len > 0) {
			Texts_OpenReader(&r, Texts_Reader__typ, t, t->len - 1);
			Texts_Read(&r, Texts_Reader__typ, &ch);
		} else {
			ch = 0x00;
		}
		if (ch != '!') {
			Texts_Write(&DialogFrames_w0, Texts_Writer__typ, '!');
			Texts_Append(t, DialogFrames_w0.buf);
		}
	}
}

void DialogFrames_Frame_Restore (DialogFrames_Frame f)
{
	Oberon_RemoveMarks(f->X, f->Y, f->W, f->H);
	if (f->pat != 32767) {
		Display_ReplPatternC((void*)f, f->col, f->pat, f->X, f->Y, f->W, f->H, f->X, f->Y, 0);
	} else {
		Display_ReplConstC((void*)f, f->col, f->X, f->Y, f->W, f->H, 0);
	}
	__Dialogs_Panel_Draw(f->panel, f->X, f->Y + f->H, (void*)f);
}

static void DialogFrames_Frame_DrawObject (DialogFrames_Frame f, Dialogs_Object o, BOOLEAN drawmode)
{
	INTEGER x = 0, y = 0, ox = 0, oy = 0, ow = 0, oh = 0;
	LONGINT i = 0;
	__Dialogs_Object_GetDim(o, &ox, &oy, &ow, &oh);
	x = f->X + ox;
	y = (f->Y + f->H) + oy;
	Oberon_RemoveMarks(x, y, ow, oh);
	if (!drawmode) {
		Display_ReplConstC((void*)f, f->col, x, y, ow, oh, 1);
		if (f->pat != 32767) {
			Display_ReplPatternC((void*)f, f->col, f->pat, x, y, ow, oh, f->X, f->Y, 0);
		}
	} else {
		__Dialogs_Object_Draw(o, x, y, (void*)f);
	}
}

static void DialogFrames_Frame_TrackMouse (DialogFrames_Frame f, INTEGER x, INTEGER y, SET keys)
{
	Oberon_DrawCursor(&Oberon_Mouse, Oberon_Cursor__typ, &Oberon_Arrow, Oberon_Marker__typ, x, y);
	while (keys != 0x0) {
		Input_Mouse(&keys, &x, &y);
		Oberon_DrawCursor(&Oberon_Mouse, Oberon_Cursor__typ, &Oberon_Arrow, Oberon_Marker__typ, x, y);
	}
}

static void DialogFrames_Frame_Send (DialogFrames_Frame f, INTEGER x, INTEGER y, Display_FrameMsg *m, LONGINT *m__typ, BOOLEAN *cond)
{
	Dialogs_Object o = NIL;
	o = __Dialogs_Panel_ThisObject(f->panel, x - f->X, (y - f->Y) - f->H);
	if (o != NIL) {
		__Dialogs_Object_Handle(o, (void*)f, &*m, m__typ);
		*cond = 1;
	} else {
		*cond = 0;
	}
}

static void DialogFrames_Frame_Extend (DialogFrames_Frame f, INTEGER newY)
{
	INTEGER dY = 0, newH = 0;
	dY = f->Y - newY;
	if (f->pat != 32767) {
		Display_ReplPattern(f->col, f->pat, f->X, newY, f->W, f->Y - newY, 0);
	} else {
		Display_ReplConst(f->col, f->X, newY, f->W, f->Y - newY, 0);
	}
	f->H = (f->H + f->Y) - newY;
	f->Y = newY;
	__Dialogs_Panel_Draw(f->panel, f->X, f->Y + f->H, (void*)f);
}

static void DialogFrames_Frame_Reduce (DialogFrames_Frame f, INTEGER newY)
{
	f->H = (f->H + f->Y) - newY;
	f->Y = newY;
}

static void DialogFrames_Frame_Modify (DialogFrames_Frame f, INTEGER id, INTEGER dY, INTEGER y, INTEGER h)
{
	Oberon_RemoveMarks(f->X, f->Y, f->W, f->H);
	__Dialogs_Panel_RemoveSelections(f->panel);
	if (id == 0) {
		if (dY > 0) {
			if (f->pat != 32767) {
				Display_ReplPattern(f->col, f->pat, f->X, f->Y + dY, f->W, f->H, 0);
			} else {
				Display_ReplConst(f->col, f->X, f->Y + dY, f->W, f->H, 0);
			}
			f->Y += dY;
		}
		__DialogFrames_Frame_Extend(f, y);
	} else if (id == 1) {
		__DialogFrames_Frame_Reduce(f, y + dY);
		if (dY > 0) {
			Display_CopyBlock(f->X, f->Y, f->W, f->H, f->X, y, 0);
			f->Y = y;
		}
	}
}

static struct Handle__12 {
	struct Handle__12 *lnk;
} *Handle__12_s;

static BOOLEAN IsIn__13 (Display_Frame f, INTEGER x, INTEGER y);

static BOOLEAN IsIn__13 (Display_Frame f, INTEGER x, INTEGER y)
{
	if (((x >= f->X && x <= f->X + f->W) && y > f->Y) && y <= f->Y + f->H) {
		return 1;
	} else {
		return 0;
	}
	__RETCHK;
}

void DialogFrames_Handle (Display_Frame f, Display_FrameMsg *m, LONGINT *m__typ)
{
	BOOLEAN cond = 0;
	DialogFrames_Frame copy = NIL;
	struct Handle__12 _s;
	_s.lnk = Handle__12_s;
	Handle__12_s = &_s;
	if (__ISP(f, DialogFrames_FrameDesc, 1)) {
		if (__IS(m__typ, Oberon_InputMsg, 1)) {
			Oberon_InputMsg *m__ = (void*)m;
			if ((*m__).id == 1) {
				if (IsIn__13((void*)(*(DialogFrames_Frame*)&f), (*m__).X, (*m__).Y)) {
					__DialogFrames_Frame_Send((*(DialogFrames_Frame*)&f), (*m__).X, (*m__).Y, (void*)&*m__, m__typ, &cond);
					if (!cond) {
						__DialogFrames_Frame_TrackMouse((*(DialogFrames_Frame*)&f), (*m__).X, (*m__).Y, (*m__).keys);
					}
				}
			} else {
				__Dialogs_Panel_Broadcast((*(DialogFrames_Frame*)&f)->panel, (void*)(*(DialogFrames_Frame*)&f), (void*)&*m__, m__typ);
			}
		} else if (__IS(m__typ, MenuViewers_ModifyMsg, 1)) {
			MenuViewers_ModifyMsg *m__ = (void*)m;
			__DialogFrames_Frame_Modify((*(DialogFrames_Frame*)&f), (*m__).id, (*m__).dY, (*m__).Y, (*m__).H);
			__Dialogs_Panel_Broadcast((*(DialogFrames_Frame*)&f)->panel, (void*)(*(DialogFrames_Frame*)&f), (void*)&*m__, m__typ);
		} else if (__IS(m__typ, Oberon_CopyMsg, 1)) {
			Oberon_CopyMsg *m__ = (void*)m;
			__NEW(copy, DialogFrames_FrameDesc);
			__GUARDEQP(copy, DialogFrames_FrameDesc) = *(*(DialogFrames_Frame*)&f);
			(*m__).F = (Display_Frame)copy;
		} else if (__IS(m__typ, Dialogs_NotifyMsg, 1)) {
			Dialogs_NotifyMsg *m__ = (void*)m;
			if ((*m__).id == 0) {
				if (__Dialogs_Panel_Contains((*(DialogFrames_Frame*)&f)->panel, (*m__).obj)) {
					__DialogFrames_Frame_DrawObject((*(DialogFrames_Frame*)&f), (*m__).obj, 1);
				}
			} else if ((*m__).id == 1) {
				if (__Dialogs_Panel_Contains((*(DialogFrames_Frame*)&f)->panel, (*m__).obj)) {
					__DialogFrames_Frame_DrawObject((*(DialogFrames_Frame*)&f), (*m__).obj, 0);
				}
			} else if ((*m__).id == 2) {
				if ((*m__).p == (*(DialogFrames_Frame*)&f)->panel) {
					__DialogFrames_Frame_MarkMenu((*(DialogFrames_Frame*)&f));
				}
			} else if ((*m__).id == 3) {
				if ((*m__).p == (*(DialogFrames_Frame*)&f)->panel) {
					__DialogFrames_Frame_Restore((*(DialogFrames_Frame*)&f));
				}
			}
		} else if (__IS(m__typ, DialogFrames_SetCaretMsg, 1)) {
			DialogFrames_SetCaretMsg *m__ = (void*)m;
			if ((*m__).p == (*(DialogFrames_Frame*)&f)->panel) {
				Oberon_RemoveMarks((*(DialogFrames_Frame*)&f)->X, (*(DialogFrames_Frame*)&f)->Y, (*(DialogFrames_Frame*)&f)->W, (*(DialogFrames_Frame*)&f)->H);
				Oberon_Pointer.X = (*m__).x + (*(DialogFrames_Frame*)&f)->X;
				Oberon_Pointer.Y = ((*m__).y + (*(DialogFrames_Frame*)&f)->Y) + (*(DialogFrames_Frame*)&f)->H;
			}
		} else if (__IS(m__typ, DialogFrames_GetFrameMsg, 1)) {
			DialogFrames_GetFrameMsg *m__ = (void*)m;
			if ((*(DialogFrames_Frame*)&f)->panel == (*m__).p) {
				(*m__).f = (*(DialogFrames_Frame*)&f);
			}
		} else {
			__Dialogs_Panel_Broadcast((*(DialogFrames_Frame*)&f)->panel, (void*)(*(DialogFrames_Frame*)&f), &*m, m__typ);
		}
	} else __WITHCHK;
	Handle__12_s = _s.lnk;
}

void DialogFrames_Frame_Open (DialogFrames_Frame f, Display_Handler handle, Dialogs_Panel p)
{
	f->handle = handle;
	f->panel = p;
	f->col = 13;
	f->grid = 1;
	f->pat = 32767;
}

void DialogFrames_GetCaretPosition (Dialogs_Panel *p, INTEGER *xpos, INTEGER *ypos)
{
	INTEGER x = 0, y = 0;
	DialogFrames_Frame f = NIL;
	Viewers_Viewer v = NIL;
	x = Oberon_Pointer.X;
	y = Oberon_Pointer.Y;
	v = Viewers_This(x, y);
	if ((v->dsc != NIL && v->dsc->next != NIL) && __ISP(v->dsc->next, DialogFrames_FrameDesc, 1)) {
		f = __GUARDP(v->dsc->next, DialogFrames_FrameDesc, 1);
		*p = f->panel;
		*xpos = x - f->X;
		*ypos = (y - f->Y) - f->H;
	} else {
		*p = NIL;
	}
}

static void DialogFrames_box (Dialogs_Object obj, BOOLEAN *done)
{
	INTEGER x = 0, y = 0, w = 0, h = 0;
	__Dialogs_Object_GetDim(obj, &x, &y, &w, &h);
	if (x < DialogFrames_left) {
		DialogFrames_left = x;
	}
	if (y < DialogFrames_bot) {
		DialogFrames_bot = y;
	}
	if (x + w > DialogFrames_right) {
		DialogFrames_right = x + w;
	}
	if (y + h > DialogFrames_top) {
		DialogFrames_top = y + h;
	}
}

void DialogFrames_OpenPanel (CHAR *name, LONGINT name__len, INTEGER x, INTEGER y, Dialogs_Panel *p)
{
	DialogFrames_Frame f = NIL;
	Files_File file = NIL;
	Files_Rider r = {0};
	INTEGER h = 0, res = 0;
	Viewers_Viewer v = NIL, vmax = NIL;
	TextFrames_Frame m = NIL;
	Texts_Text t = NIL;
	Texts_Buffer buf = NIL;
	__DUP(name, name__len, CHAR);
	file = Files_Old(name, name__len);
	__NEW(*p, Dialogs_PanelDesc);
	if (file != NIL) {
		Files_Set(&r, Files_Rider__typ, file, 0);
		__Dialogs_Panel_Load(*p, &r, Files_Rider__typ);
	}
	__NEW(f, DialogFrames_FrameDesc);
	__DialogFrames_Frame_Open(f, DialogFrames_Handle, *p);
	v = Viewers_This(x, 0);
	vmax = NIL;
	h = 0;
	while (v->state > 1) {
		if (v->H > h) {
			vmax = v;
			h = v->H;
		}
		v = Viewers_Next(v);
	}
	if (vmax != NIL) {
		DialogFrames_left = 32767;
		DialogFrames_right = -32768;
		DialogFrames_bot = 32767;
		DialogFrames_top = -32768;
		__Dialogs_Panel_Enumerate(*p, DialogFrames_box);
		y = DialogFrames_Min(((vmax->Y + __ABS(DialogFrames_bot)) + 10) + TextFrames_menuH, ((vmax->Y + vmax->H) - TextFrames_menuH) - 2);
	}
	m = TextFrames_NewMenu(name, name__len, (CHAR*)"^Dialog.Menu.Text", (LONGINT)18);
	v = (Viewers_Viewer)MenuViewers_New((void*)m, (void*)f, TextFrames_menuH, x, y);
	if ((*p)->cmd[0] != 0x00) {
		Dialogs_cmdPanel = *p;
		Oberon_Call((*p)->cmd, 64, Oberon_Par, 0, &res);
	}
	__DEL(name);
}

void DialogFrames_FindObject (Dialogs_Object *o, Dialogs_Panel *p)
{
	INTEGER x = 0, y = 0;
	DialogFrames_GetCaretPosition(&*p, &x, &y);
	if (*p != NIL) {
		*o = __Dialogs_Panel_ThisObject(*p, x, y);
		if (*o != NIL) {
			Dialogs_res = 0;
		} else {
			Dialogs_res = 3;
		}
	} else {
		Dialogs_res = 5;
	}
}

void DialogFrames_SetCaretAtObject (Dialogs_Object o)
{
	DialogFrames_SetCaretMsg msg = {0};
	INTEGER x = 0, y = 0, w = 0, h = 0;
	__Dialogs_Object_GetDim(o, &x, &y, &w, &h);
	msg.p = o->panel;
	msg.x = x;
	msg.y = y;
	Viewers_Broadcast((void*)&msg, DialogFrames_SetCaretMsg__typ);
}

static void EnumPtrs(void (*P)(void*))
{
	__ENUMR(&DialogFrames_w0, Texts_Writer__typ, 36, 1, P);
}

__TDESC(DialogFrames_FrameDesc, 10, 3) = {__TDFLDS("FrameDesc", 36), {0, 4, 24, -16}};
__TDESC(DialogFrames_GetFrameMsg, 1, 2) = {__TDFLDS("GetFrameMsg", 8), {0, 4, -12}};
__TDESC(DialogFrames_SetCaretMsg, 1, 1) = {__TDFLDS("SetCaretMsg", 8), {0, -8}};

export void *DialogFrames__init(void)
{
	__DEFMOD;
	__IMPORT(Dialogs);
	__IMPORT(Display);
	__IMPORT(Files);
	__IMPORT(Input);
	__IMPORT(MenuViewers);
	__IMPORT(Oberon);
	__IMPORT(TextFrames);
	__IMPORT(Texts);
	__IMPORT(Viewers);
	__REGMOD(DialogFrames, EnumPtrs);
	__INITYP(DialogFrames_FrameDesc, Display_FrameDesc, 1);
	__INITBP(DialogFrames_FrameDesc, DialogFrames_Frame_DrawObject, 0);
	__INITBP(DialogFrames_FrameDesc, DialogFrames_Frame_Extend, 1);
	__INITBP(DialogFrames_FrameDesc, DialogFrames_Frame_MarkMenu, 2);
	__INITBP(DialogFrames_FrameDesc, DialogFrames_Frame_Modify, 3);
	__INITBP(DialogFrames_FrameDesc, DialogFrames_Frame_Open, 4);
	__INITBP(DialogFrames_FrameDesc, DialogFrames_Frame_Reduce, 5);
	__INITBP(DialogFrames_FrameDesc, DialogFrames_Frame_Restore, 6);
	__INITBP(DialogFrames_FrameDesc, DialogFrames_Frame_Send, 7);
	__INITBP(DialogFrames_FrameDesc, DialogFrames_Frame_TrackMouse, 8);
	__INITYP(DialogFrames_GetFrameMsg, Display_FrameMsg, 1);
	__INITYP(DialogFrames_SetCaretMsg, Display_FrameMsg, 1);
/* BEGIN */
	Texts_OpenWriter(&DialogFrames_w0, Texts_Writer__typ);
	__ENDMOD;
}
