/* Ofront 1.4 -xtspkaG */
#include "SYSTEM.h"
#include "DialogFrames.h"
#include "DialogTexts.h"
#include "Dialogs.h"
#include "Display.h"
#include "Files.h"
#include "GraphicFrames.h"
#include "Graphics.h"
#include "In.h"
#include "MenuViewers.h"
#include "Oberon.h"
#include "TextFrames.h"
#include "TextPrinter.h"
#include "Texts.h"
#include "Viewers.h"

typedef
	struct DialogGraphics_FrameDesc *DialogGraphics_Frame;

typedef
	struct DialogGraphics_ItemDesc *DialogGraphics_Item;

typedef
	struct DialogGraphics_FrameDesc { /* GraphicFrames_FrameDesc */
		Display_Frame dsc, next;
		INTEGER X, Y, W, H;
		Display_Handler handle;
		Graphics_Graph graph;
		INTEGER Xg, Yg, X1, Y1, x, y, col;
		BOOLEAN marked, ticked;
		GraphicFrames_LocDesc mark;
		DialogGraphics_Item elem;
	} DialogGraphics_FrameDesc;

typedef
	struct DialogGraphics_ItemDesc { /* Dialogs_ObjectDesc */
		LONGINT _prvt0;
		char _prvt1[16];
		CHAR name[16];
		CHAR cmd[32];
		CHAR par[32];
		BOOLEAN selected, overlapping;
		Dialogs_Panel panel;
		char _prvt2[4];
		Graphics_Graph graph;
		BOOLEAN empty;
		INTEGER xg, yg;
	} DialogGraphics_ItemDesc;

export void DialogGraphics_Item_Copy (DialogGraphics_Item g, Dialogs_Object *dup);
export void DialogGraphics_Item_Draw (DialogGraphics_Item g, INTEGER x, INTEGER y, Display_Frame f);
export void DialogGraphics_Item_Edit (DialogGraphics_Item g);
export void DialogGraphics_Item_Handle (DialogGraphics_Item g, Display_Frame f, Display_FrameMsg *m, LONGINT *m__typ);
export void DialogGraphics_Item_Load (DialogGraphics_Item g, Files_Rider *r, LONGINT *r__typ);
static void DialogGraphics_Item_Open (DialogGraphics_Item g);
export void DialogGraphics_Item_Print (DialogGraphics_Item g, INTEGER x, INTEGER y);
export void DialogGraphics_Item_Store (DialogGraphics_Item g, Files_Rider *r, LONGINT *r__typ);
#define __DialogGraphics_Item_Open(g) __SEND(__TYPEOF(g), 24, void(*)(DialogGraphics_Item), (g))


static INTEGER DialogGraphics_x0, DialogGraphics_x1, DialogGraphics_y0, DialogGraphics_y1;
static Texts_Writer DialogGraphics_w;
static INTEGER DialogGraphics_dUnit, DialogGraphics_GripW, DialogGraphics_GripH;

export LONGINT *DialogGraphics_ItemDesc__typ;
export LONGINT *DialogGraphics_FrameDesc__typ;

static Graphics_Graph DialogGraphics_CopyGraph (Graphics_Graph g);
static void DialogGraphics_HandleFrame (Display_Frame f0, Display_FrameMsg *msg, LONGINT *msg__typ);
export void DialogGraphics_Insert (void);
static void DialogGraphics_MarkMenu (DialogGraphics_Frame f);
static INTEGER DialogGraphics_Max (INTEGER x, INTEGER y);
export void DialogGraphics_Update (void);
static void DialogGraphics_box (Graphics_Object obj, BOOLEAN *done);


static void DialogGraphics_box (Graphics_Object obj, BOOLEAN *done)
{
	if (obj->x < DialogGraphics_x0) {
		DialogGraphics_x0 = obj->x;
	}
	if (DialogGraphics_x1 < obj->x + obj->w) {
		DialogGraphics_x1 = obj->x + obj->w;
	}
	if (obj->y < DialogGraphics_y0) {
		DialogGraphics_y0 = obj->y;
	}
	if (DialogGraphics_y1 < obj->y + obj->h) {
		DialogGraphics_y1 = obj->y + obj->h;
	}
}

static INTEGER DialogGraphics_Max (INTEGER x, INTEGER y)
{
	if (x > y) {
		return x;
	} else {
		return y;
	}
	__RETCHK;
}

static Graphics_Graph DialogGraphics_CopyGraph (Graphics_Graph g)
{
	Graphics_Graph g1 = NIL;
	Graphics_SelectArea(g, -32768, -32768, 32767, 32767);
	__NEW(g1, Graphics_GraphDesc);
	Graphics_Copy(g, g1, 0, 0);
	Graphics_Deselect(g1);
	return g1;
}

void DialogGraphics_Item_Draw (DialogGraphics_Item g, INTEGER x, INTEGER y, Display_Frame f)
{
	INTEGER x0 = 0, y0 = 0, w = 0, h = 0, mode = 0;
	GraphicFrames_Frame f1 = NIL;
	__Dialogs_Object_GetDim((void*)g, &x0, &y0, &w, &h);
	if (g->selected) {
		mode = 2;
	} else {
		mode = 0;
	}
	if ((((g->empty || x < f->X) || y < f->Y) || x + w > f->X + f->W) || y + h > f->Y + f->H) {
		Display_ReplPatternC(f, 15, Display_grey1, x, y, w, h, 0, 0, mode);
	} else {
		__NEW(f1, GraphicFrames_FrameDesc);
		GraphicFrames_Open(f1, g->graph, g->xg, g->yg, __GUARDP(f, DialogFrames_FrameDesc, 1)->col, 0);
		f1->X = x;
		f1->Y = y;
		f1->W = w;
		f1->H = h;
		GraphicFrames_Restore(f1);
	}
}

void DialogGraphics_Item_Print (DialogGraphics_Item g, INTEGER x, INTEGER y)
{
	INTEGER x0 = 0, y0 = 0, w = 0, h = 0;
	__Dialogs_Object_GetPDim((void*)g, &x0, &y0, &w, &h);
	if (!g->empty) {
		Graphics_Print(g->graph, (int)((int)__ASHL(g->xg, 2)) + x, ((int)((int)__ASHL(g->yg, 2)) + y) + h);
	}
}

void DialogGraphics_Item_Copy (DialogGraphics_Item g, Dialogs_Object *dup)
{
	DialogGraphics_Item x = NIL;
	if (*dup == NIL) {
		__NEW(x, DialogGraphics_ItemDesc);
		*dup = (Dialogs_Object)x;
	} else {
		x = __GUARDP(*dup, DialogGraphics_ItemDesc, 1);
	}
	Dialogs_Object_Copy((void*)g, &*dup);
	x->graph = DialogGraphics_CopyGraph(g->graph);
	x->xg = g->xg;
	x->yg = g->yg;
	x->empty = g->empty;
}

void DialogGraphics_Item_Load (DialogGraphics_Item g, Files_Rider *r, LONGINT *r__typ)
{
	Dialogs_Object_Load((void*)g, &*r, r__typ);
	Files_ReadBool(&*r, r__typ, &g->empty);
	Files_ReadInt(&*r, r__typ, &g->xg);
	Files_ReadInt(&*r, r__typ, &g->yg);
	__NEW(g->graph, Graphics_GraphDesc);
	Graphics_Load(g->graph, &*r, r__typ);
}

void DialogGraphics_Item_Store (DialogGraphics_Item g, Files_Rider *r, LONGINT *r__typ)
{
	Dialogs_Object_Store((void*)g, &*r, r__typ);
	Files_WriteBool(&*r, r__typ, g->empty);
	Files_WriteInt(&*r, r__typ, g->xg);
	Files_WriteInt(&*r, r__typ, g->yg);
	Graphics_Store(g->graph, &*r, r__typ);
}

void DialogGraphics_Item_Handle (DialogGraphics_Item g, Display_Frame f, Display_FrameMsg *m, LONGINT *m__typ)
{
	Texts_Text t = NIL;
	Dialogs_Object_Handle((void*)g, f, &*m, m__typ);
	if (__ISP(f, DialogFrames_FrameDesc, 1)) {
		if (__IS(m__typ, Oberon_InputMsg, 1)) {
			Oberon_InputMsg *m__ = (void*)m;
			if (((*m__).id == 1 && (((*m__).keys == 0x02 || (*m__).keys == 0x04) || (*m__).keys == 0x01)) && g->cmd[0] != 0x00) {
				(*DialogTexts_GetParText)(g->par, 32, g->panel, &t);
				__Dialogs_Object_CallCmd((void*)g, (void*)(*(DialogFrames_Frame*)&f), Viewers_This((*m__).X, (*m__).Y), t);
			}
		}
	}
}

static void DialogGraphics_MarkMenu (DialogGraphics_Frame f)
{
	Texts_Reader r = {0};
	Viewers_Viewer v = NIL;
	Texts_Text t = NIL;
	CHAR ch = 0;
	v = Viewers_This(f->X, f->Y);
	if (__ISP(v, MenuViewers_ViewerDesc, 2)) {
		t = __GUARDP(v->dsc, TextFrames_FrameDesc, 1)->text;
		if (t->len > 0) {
			Texts_OpenReader(&r, Texts_Reader__typ, t, t->len - 1);
			Texts_Read(&r, Texts_Reader__typ, &ch);
		} else {
			ch = 0x00;
		}
		if (ch != '!') {
			Texts_Write(&DialogGraphics_w, Texts_Writer__typ, '!');
			Texts_Append(t, DialogGraphics_w.buf);
		}
	}
}

static void DialogGraphics_HandleFrame (Display_Frame f0, Display_FrameMsg *msg, LONGINT *msg__typ)
{
	DialogGraphics_Frame f = NIL, f1 = NIL;
	f = __GUARDP(f0, DialogGraphics_FrameDesc, 2);
	if (__IS(msg__typ, Oberon_InputMsg, 1)) {
		GraphicFrames_Handle((void*)f, &*msg, msg__typ);
		if (__IS(msg__typ, Oberon_InputMsg, 1)) {
			Oberon_InputMsg *msg__ = (void*)msg;
			if ((*msg__).id == 0 || (*msg__).id == 1 && (*msg__).keys != 0x0) {
				DialogGraphics_MarkMenu(f);
			}
		} else __WITHCHK;
	} else if (__IS(msg__typ, Oberon_CopyMsg, 1)) {
		__NEW(f1, DialogGraphics_FrameDesc);
		GraphicFrames_Open((void*)f1, f->graph, f->Xg, f->Yg, f->col, f->ticked);
		f1->handle = f->handle;
		f1->elem = f->elem;
		__GUARDR(msg, Oberon_CopyMsg, 1).F = (Display_Frame)f1;
	} else {
		GraphicFrames_Handle((void*)f, &*msg, msg__typ);
	}
}

void DialogGraphics_Item_Edit (DialogGraphics_Item g)
{
	Viewers_Viewer v = NIL;
	DialogGraphics_Frame f = NIL;
	INTEGER x = 0, y = 0;
	__NEW(f, DialogGraphics_FrameDesc);
	GraphicFrames_Open((void*)f, DialogGraphics_CopyGraph(g->graph), 0, 0, 0, 1);
	f->elem = g;
	f->handle = DialogGraphics_HandleFrame;
	Oberon_AllocateSystemViewer(Oberon_Mouse.X, &x, &y);
	v = (Viewers_Viewer)MenuViewers_New((void*)TextFrames_NewMenu((CHAR*)"DialogGraphics.Graph", (LONGINT)21, (CHAR*)"^DialogDraw.Menu.Text", (LONGINT)22), (void*)f, TextFrames_menuH, x, y);
}

static void DialogGraphics_Item_Open (DialogGraphics_Item g)
{
	DialogGraphics_x0 = 32767;
	DialogGraphics_x1 = -32768;
	DialogGraphics_y0 = 32767;
	DialogGraphics_y1 = -32768;
	Graphics_Enumerate(g->graph, DialogGraphics_box);
	if (DialogGraphics_x0 == 32767) {
		DialogGraphics_x0 = 0;
		DialogGraphics_y0 = 0;
		g->empty = 1;
		g->xg = 0;
		g->yg = 0;
	} else {
		g->xg = -DialogGraphics_x0;
		g->yg = -DialogGraphics_y1;
		DialogGraphics_x1 -= DialogGraphics_x0;
		DialogGraphics_y1 -= DialogGraphics_y0;
		g->empty = 0;
	}
}

void DialogGraphics_Insert (void)
{
	INTEGER w = 0, h = 0, x = 0, y = 0, xh = 0, yh = 0;
	DialogGraphics_Item g = NIL;
	Dialogs_Panel p = NIL;
	CHAR name[64];
	__NEW(g, DialogGraphics_ItemDesc);
	DialogFrames_GetCaretPosition(&p, &x, &y);
	if (p != NIL) {
		__Dialogs_Object_Init((void*)g);
		In_Open();
		In_Name((void*)name, 64);
		if (!In_Done) {
			__COPY("", name, 64);
			In_Open();
		}
		__Dialogs_Object_SetName((void*)g, name, 64);
		__NEW(g->graph, Graphics_GraphDesc);
		Graphics_Open(g->graph, name, 64);
		__DialogGraphics_Item_Open(g);
		In_Int(&xh);
		In_Int(&yh);
		In_Int(&w);
		In_Int(&h);
		if (In_Done) {
			w = DialogGraphics_Max(DialogGraphics_x1, w);
			h = DialogGraphics_Max(DialogGraphics_y1, h);
			x = xh;
			y = yh;
		} else if (g->empty) {
			w = 100;
			h = 100;
		} else {
			w = DialogGraphics_x1;
			h = DialogGraphics_y1;
		}
		__Dialogs_Object_SetDim((void*)g, x, y, w, h, 0);
		__Dialogs_Panel_Insert(p, (void*)g, 0);
	} else {
		Dialogs_res = 5;
	}
	if (Dialogs_res != 0) {
		Dialogs_Error((CHAR*)"DialogGraphics", (LONGINT)15);
	}
}

void DialogGraphics_Update (void)
{
	Viewers_Viewer v = NIL;
	DialogGraphics_Frame f = NIL;
	Texts_Reader r = {0};
	Texts_Text t = NIL;
	CHAR ch = 0;
	INTEGER x = 0, y = 0, w = 0, h = 0;
	v = Oberon_Par->vwr;
	f = __GUARDP(v->dsc->next, DialogGraphics_FrameDesc, 2);
	t = __GUARDP(v->dsc, TextFrames_FrameDesc, 1)->text;
	GraphicFrames_Deselect((void*)f);
	f->elem->graph = DialogGraphics_CopyGraph(f->graph);
	__DialogGraphics_Item_Open(f->elem);
	__Dialogs_Object_GetDim((void*)f->elem, &x, &y, &w, &h);
	w = DialogGraphics_Max(DialogGraphics_x1, w);
	h = DialogGraphics_Max(DialogGraphics_y1, h);
	__Dialogs_Object_SetDim((void*)f->elem, x, y, w, h, 0);
	__Dialogs_Object_Restore((void*)f->elem);
	__Dialogs_Panel_MarkMenu(f->elem->panel);
	Texts_OpenReader(&r, Texts_Reader__typ, t, t->len - 1);
	Texts_Read(&r, Texts_Reader__typ, &ch);
	if (ch == '!') {
		Texts_Delete(t, t->len - 1, t->len);
	}
}

static void EnumPtrs(void (*P)(void*))
{
	__ENUMR(&DialogGraphics_w, Texts_Writer__typ, 36, 1, P);
}

__TDESC(DialogGraphics_ItemDesc, 26, 3) = {__TDFLDS("ItemDesc", 124), {0, 104, 112, -16}};
__TDESC(DialogGraphics_FrameDesc, 1, 5) = {__TDFLDS("FrameDesc", 52), {0, 4, 20, 44, 48, -24}};

export void *DialogGraphics__init(void)
{
	__DEFMOD;
	__IMPORT(DialogFrames);
	__IMPORT(DialogTexts);
	__IMPORT(Dialogs);
	__IMPORT(Display);
	__IMPORT(Files);
	__IMPORT(GraphicFrames);
	__IMPORT(Graphics);
	__IMPORT(In);
	__IMPORT(MenuViewers);
	__IMPORT(Oberon);
	__IMPORT(TextFrames);
	__IMPORT(TextPrinter);
	__IMPORT(Texts);
	__IMPORT(Viewers);
	__REGMOD(DialogGraphics, EnumPtrs);
	__REGCMD("Insert", DialogGraphics_Insert);
	__REGCMD("Update", DialogGraphics_Update);
	__INITYP(DialogGraphics_ItemDesc, Dialogs_ObjectDesc, 1);
	__INITBP(DialogGraphics_ItemDesc, DialogGraphics_Item_Copy, 1);
	__INITBP(DialogGraphics_ItemDesc, DialogGraphics_Item_Draw, 2);
	__INITBP(DialogGraphics_ItemDesc, DialogGraphics_Item_Edit, 3);
	__INITBP(DialogGraphics_ItemDesc, DialogGraphics_Item_Handle, 6);
	__INITBP(DialogGraphics_ItemDesc, DialogGraphics_Item_Load, 12);
	__INITBP(DialogGraphics_ItemDesc, DialogGraphics_Item_Open, 24);
	__INITBP(DialogGraphics_ItemDesc, DialogGraphics_Item_Print, 14);
	__INITBP(DialogGraphics_ItemDesc, DialogGraphics_Item_Store, 21);
	__INITYP(DialogGraphics_FrameDesc, GraphicFrames_FrameDesc, 2);
/* BEGIN */
	Texts_OpenWriter(&DialogGraphics_w, Texts_Writer__typ);
	DialogGraphics_dUnit = TextFrames_Unit;
	DialogGraphics_GripW = (int)__DIV(54000, (LONGINT)DialogGraphics_dUnit);
	DialogGraphics_GripH = DialogGraphics_GripW;
	__ENDMOD;
}
