/* Ofront 1.4 -xtspkaG */
#include "SYSTEM.h"
#include "DialogFrames.h"
#include "DialogTexts.h"
#include "Dialogs.h"
#include "Display.h"
#include "Files.h"
#include "In.h"
#include "KeplerFrames.h"
#include "KeplerGraphs.h"
#include "KeplerPorts.h"
#include "MenuViewers.h"
#include "Oberon.h"
#include "TextFrames.h"
#include "TextPrinter.h"
#include "Texts.h"
#include "Viewers.h"

typedef
	struct DialogKeplers_FrameDesc *DialogKeplers_Frame;

typedef
	struct DialogKeplers_ItemDesc *DialogKeplers_Item;

typedef
	struct DialogKeplers_FrameDesc { /* KeplerFrames_FrameDesc */
		Display_Frame dsc, next;
		INTEGER X, Y, W, H;
		Display_Handler handle;
		INTEGER x0, y0, scale;
		KeplerPorts_Port ext;
		KeplerGraphs_Graph G;
		INTEGER col, grid;
		DialogKeplers_Item elem;
	} DialogKeplers_FrameDesc;

typedef
	struct DialogKeplers_ItemDesc { /* Dialogs_ObjectDesc */
		LONGINT _prvt0;
		char _prvt1[16];
		CHAR name[16];
		CHAR cmd[32];
		CHAR par[32];
		BOOLEAN selected, overlapping;
		Dialogs_Panel panel;
		char _prvt2[4];
		KeplerGraphs_Graph graph;
		INTEGER dx, dy, grid;
	} DialogKeplers_ItemDesc;

export void DialogKeplers_Item_Copy (DialogKeplers_Item g, Dialogs_Object *dup);
export void DialogKeplers_Item_Draw (DialogKeplers_Item g, INTEGER x, INTEGER y, Display_Frame f);
export void DialogKeplers_Item_Edit (DialogKeplers_Item g);
export void DialogKeplers_Item_Handle (DialogKeplers_Item g, Display_Frame f, Display_FrameMsg *m, LONGINT *m__typ);
export void DialogKeplers_Item_Load (DialogKeplers_Item g, Files_Rider *r, LONGINT *r__typ);
static void DialogKeplers_Item_Open (DialogKeplers_Item g);
export void DialogKeplers_Item_Print (DialogKeplers_Item g, INTEGER x, INTEGER y);
export void DialogKeplers_Item_Store (DialogKeplers_Item g, Files_Rider *r, LONGINT *r__typ);
#define __DialogKeplers_Item_Open(g) __SEND(__TYPEOF(g), 24, void(*)(DialogKeplers_Item), (g))



export LONGINT *DialogKeplers_ItemDesc__typ;
export LONGINT *DialogKeplers_FrameDesc__typ;

static KeplerGraphs_Graph DialogKeplers_CopyGraph (KeplerGraphs_Graph g);
static void DialogKeplers_HandleFrame (Display_Frame f0, Display_FrameMsg *msg, LONGINT *msg__typ);
export void DialogKeplers_Insert (void);
static INTEGER DialogKeplers_Max (INTEGER x, INTEGER y);
export void DialogKeplers_Update (void);


static INTEGER DialogKeplers_Max (INTEGER x, INTEGER y)
{
	if (x > y) {
		return x;
	} else {
		return y;
	}
	__RETCHK;
}

static KeplerGraphs_Graph DialogKeplers_CopyGraph (KeplerGraphs_Graph g)
{
	Files_File buf = NIL;
	Files_Rider R = {0};
	KeplerGraphs_Object o = NIL;
	buf = Files_New((CHAR*)"", (LONGINT)1);
	Files_Set(&R, Files_Rider__typ, buf, 0);
	KeplerGraphs_Reset();
	KeplerGraphs_WriteObj(&R, Files_Rider__typ, (void*)g);
	Files_Set(&R, Files_Rider__typ, buf, 0);
	KeplerGraphs_Reset();
	KeplerGraphs_ReadObj(&R, Files_Rider__typ, &o);
	return __GUARDP(o, KeplerGraphs_GraphDesc, 1);
}

void DialogKeplers_Item_Draw (DialogKeplers_Item g, INTEGER x, INTEGER y, Display_Frame f)
{
	INTEGER x0 = 0, y0 = 0, w = 0, h = 0, mode = 0, col = 0;
	KeplerFrames_Frame f1 = NIL;
	__Dialogs_Object_GetDim((void*)g, &x0, &y0, &w, &h);
	col = __GUARDP(f, DialogFrames_FrameDesc, 1)->col;
	if (g->selected) {
		mode = 2;
	} else {
		mode = 0;
	}
	if ((((g->graph == NIL || x < f->X) || y < f->Y) || x + w > f->X + f->W) || y + h > f->Y + f->H) {
		Display_ReplPatternC(f, 15, Display_grey1, x, y, w, h, 0, 0, mode);
	} else {
		f1 = KeplerFrames_New(g->graph);
		f = (Display_Frame)f1;
		f1->X = x;
		f1->Y = y;
		f1->W = w;
		f1->H = h;
		f1->x0 = -g->dx;
		f1->y0 = -(__ASHL(f1->H, 2) + g->dy);
		f1->scale = 4;
		if (g->graph->cons != NIL) {
			__KeplerGraphs_Object_Draw((void*)g->graph, (void*)f1);
		} else {
			__KeplerPorts_Port_DrawRect((void*)f1, 0, 0, 99, 99, col, mode);
		}
	}
}

void DialogKeplers_Item_Print (DialogKeplers_Item g, INTEGER x, INTEGER y)
{
	INTEGER x0 = 0, y0 = 0, w = 0, h = 0;
	KeplerPorts_PrinterPort p = NIL;
	__Dialogs_Object_GetPDim((void*)g, &x0, &y0, &w, &h);
	__NEW(p, KeplerPorts_PrinterPortDesc);
	p->X = x;
	p->Y = y;
	p->W = __ASHL(w, 2);
	p->H = __ASHL(h, 2);
	p->x0 = -g->dx;
	p->y0 = -(p->H + g->dy);
	p->scale = 1;
	__KeplerGraphs_Object_Draw((void*)g->graph, (void*)p);
}

void DialogKeplers_Item_Copy (DialogKeplers_Item g, Dialogs_Object *dup)
{
	DialogKeplers_Item x = NIL;
	if (*dup == NIL) {
		__NEW(x, DialogKeplers_ItemDesc);
		*dup = (Dialogs_Object)x;
	} else {
		x = __GUARDP(*dup, DialogKeplers_ItemDesc, 1);
	}
	Dialogs_Object_Copy((void*)g, &*dup);
	x->graph = DialogKeplers_CopyGraph(g->graph);
	x->dx = g->dx;
	x->dy = g->dy;
	x->grid = g->grid;
}

void DialogKeplers_Item_Load (DialogKeplers_Item g, Files_Rider *r, LONGINT *r__typ)
{
	KeplerGraphs_Object o = NIL;
	Dialogs_Object_Load((void*)g, &*r, r__typ);
	Files_ReadInt(&*r, r__typ, &g->dx);
	Files_ReadInt(&*r, r__typ, &g->dy);
	Files_ReadInt(&*r, r__typ, &g->grid);
	KeplerGraphs_Reset();
	KeplerGraphs_ReadObj(&*r, r__typ, &o);
	g->graph = __GUARDP(o, KeplerGraphs_GraphDesc, 1);
}

void DialogKeplers_Item_Store (DialogKeplers_Item g, Files_Rider *r, LONGINT *r__typ)
{
	Dialogs_Object_Store((void*)g, &*r, r__typ);
	Files_WriteInt(&*r, r__typ, g->dx);
	Files_WriteInt(&*r, r__typ, g->dy);
	Files_WriteInt(&*r, r__typ, g->grid);
	KeplerGraphs_Reset();
	KeplerGraphs_WriteObj(&*r, r__typ, (void*)g->graph);
}

void DialogKeplers_Item_Handle (DialogKeplers_Item g, Display_Frame f, Display_FrameMsg *m, LONGINT *m__typ)
{
	Texts_Text t = NIL;
	Dialogs_Object_Handle((void*)g, f, &*m, m__typ);
	if (__ISP(f, DialogFrames_FrameDesc, 1)) {
		if (__IS(m__typ, Oberon_InputMsg, 1)) {
			Oberon_InputMsg *m__ = (void*)m;
			if (((*m__).id == 1 && (((*m__).keys == 0x02 || (*m__).keys == 0x04) || (*m__).keys == 0x01)) && g->cmd[0] != 0x00) {
				(*DialogTexts_GetParText)(g->par, 32, g->panel, &t);
				__Dialogs_Object_CallCmd((void*)g, (void*)(*(DialogFrames_Frame*)&f), Viewers_This((*m__).X, (*m__).Y), t);
			}
		}
	}
}

static void DialogKeplers_HandleFrame (Display_Frame f0, Display_FrameMsg *msg, LONGINT *msg__typ)
{
	DialogKeplers_Frame f = NIL;
	if (__ISP(f0, DialogKeplers_FrameDesc, 4)) {
		if (__IS(msg__typ, Oberon_CopyMsg, 1)) {
			Oberon_CopyMsg *msg__ = (void*)msg;
			__NEW(f, DialogKeplers_FrameDesc);
			(*msg__).F = (Display_Frame)f;
			__GUARDEQP(f, DialogKeplers_FrameDesc) = *(*(DialogKeplers_Frame*)&f0);
		} else {
			KeplerFrames_Handle((void*)(*(DialogKeplers_Frame*)&f0), &*msg, msg__typ);
		}
	} else __WITHCHK;
}

void DialogKeplers_Item_Edit (DialogKeplers_Item g)
{
	Viewers_Viewer v = NIL;
	DialogKeplers_Frame f = NIL;
	INTEGER x = 0, y = 0;
	__NEW(f, DialogKeplers_FrameDesc);
	f->G = DialogKeplers_CopyGraph(g->graph);
	f->G->notify = KeplerFrames_NotifyDisplay;
	f->handle = DialogKeplers_HandleFrame;
	f->grid = g->grid;
	f->scale = 4;
	f->x0 = 0;
	f->y0 = 0;
	f->elem = g;
	Oberon_AllocateSystemViewer(Oberon_Mouse.X, &x, &y);
	v = (Viewers_Viewer)MenuViewers_New((void*)TextFrames_NewMenu((CHAR*)"DialogKeplers.Graph", (LONGINT)20, (CHAR*)"^DialogKepler.Menu.Text", (LONGINT)24), (void*)f, TextFrames_menuH, x, y);
}

static void DialogKeplers_Item_Open (DialogKeplers_Item g)
{
	KeplerPorts_BalloonPort b = NIL;
	INTEGER x = 0, y = 0, w = 0, h = 0;
	__NEW(b, KeplerPorts_BalloonPortDesc);
	KeplerPorts_InitBalloon(b);
	__KeplerGraphs_Object_Draw((void*)g->graph, (void*)b);
	__Dialogs_Object_GetDim((void*)g, &x, &y, &w, &h);
	g->dx = b->X;
	g->dy = b->Y;
	w = DialogKeplers_Max(w, __ASHR(b->W, 2));
	h = DialogKeplers_Max(h, __ASHR(b->H, 2));
	__Dialogs_Object_SetDim((void*)g, x, y, w, h, g->overlapping);
}

void DialogKeplers_Insert (void)
{
	INTEGER w = 0, h = 0, x = 0, y = 0, xh = 0, yh = 0;
	DialogKeplers_Item g = NIL;
	Dialogs_Panel p = NIL;
	CHAR name[64];
	KeplerGraphs_Graph graph = NIL;
	__NEW(g, DialogKeplers_ItemDesc);
	DialogFrames_GetCaretPosition(&p, &x, &y);
	if (p != NIL) {
		__Dialogs_Object_Init((void*)g);
		In_Open();
		In_Name((void*)name, 64);
		if (!In_Done) {
			__COPY("", name, 64);
			In_Open();
		}
		__Dialogs_Object_SetName((void*)g, name, 64);
		graph = KeplerGraphs_Old(name, 64);
		if (graph == NIL) {
			__NEW(graph, KeplerGraphs_GraphDesc);
		}
		g->graph = graph;
		In_Int(&xh);
		In_Int(&yh);
		In_Int(&w);
		In_Int(&h);
		if (In_Done) {
			x = xh;
			y = yh;
		} else {
			w = 100;
			h = 100;
		}
		__Dialogs_Object_SetDim((void*)g, x, y, w, h, 0);
		__DialogKeplers_Item_Open(g);
		__Dialogs_Panel_Insert(p, (void*)g, 0);
	} else {
		Dialogs_res = 5;
	}
	if (Dialogs_res != 0) {
		Dialogs_Error((CHAR*)"DialogKeplers", (LONGINT)14);
	}
}

void DialogKeplers_Update (void)
{
	Viewers_Viewer v = NIL;
	DialogKeplers_Frame f = NIL;
	KeplerGraphs_Graph g = NIL;
	v = Oberon_Par->vwr;
	f = __GUARDP(v->dsc->next, DialogKeplers_FrameDesc, 4);
	g = f->G;
	__KeplerGraphs_Graph_All(g, 0);
	f->elem->graph = DialogKeplers_CopyGraph(f->G);
	f->elem->grid = f->grid;
	__DialogKeplers_Item_Open(f->elem);
	__Dialogs_Object_Restore((void*)f->elem);
	__Dialogs_Panel_MarkMenu(f->elem->panel);
}

__TDESC(DialogKeplers_ItemDesc, 26, 3) = {__TDFLDS("ItemDesc", 124), {0, 104, 112, -16}};
__TDESC(DialogKeplers_FrameDesc, 21, 5) = {__TDFLDS("FrameDesc", 44), {0, 4, 28, 32, 40, -24}};

export void *DialogKeplers__init(void)
{
	__DEFMOD;
	__IMPORT(DialogFrames);
	__IMPORT(DialogTexts);
	__IMPORT(Dialogs);
	__IMPORT(Display);
	__IMPORT(Files);
	__IMPORT(In);
	__IMPORT(KeplerFrames);
	__IMPORT(KeplerGraphs);
	__IMPORT(KeplerPorts);
	__IMPORT(MenuViewers);
	__IMPORT(Oberon);
	__IMPORT(TextFrames);
	__IMPORT(TextPrinter);
	__IMPORT(Texts);
	__IMPORT(Viewers);
	__REGMOD(DialogKeplers, 0);
	__REGCMD("Insert", DialogKeplers_Insert);
	__REGCMD("Update", DialogKeplers_Update);
	__INITYP(DialogKeplers_ItemDesc, Dialogs_ObjectDesc, 1);
	__INITBP(DialogKeplers_ItemDesc, DialogKeplers_Item_Copy, 1);
	__INITBP(DialogKeplers_ItemDesc, DialogKeplers_Item_Draw, 2);
	__INITBP(DialogKeplers_ItemDesc, DialogKeplers_Item_Edit, 3);
	__INITBP(DialogKeplers_ItemDesc, DialogKeplers_Item_Handle, 6);
	__INITBP(DialogKeplers_ItemDesc, DialogKeplers_Item_Load, 12);
	__INITBP(DialogKeplers_ItemDesc, DialogKeplers_Item_Open, 24);
	__INITBP(DialogKeplers_ItemDesc, DialogKeplers_Item_Print, 14);
	__INITBP(DialogKeplers_ItemDesc, DialogKeplers_Item_Store, 21);
	__INITYP(DialogKeplers_FrameDesc, KeplerFrames_FrameDesc, 4);
/* BEGIN */
	__ENDMOD;
}
