/* Ofront 1.4 -xtspkaG */
#include "SYSTEM.h"
#include "Console.h"
#include "Display.h"
#include "Kernel.h"
#include "Unix.h"
#include "X11.h"


static LONGINT Display1_GcC, Display1_Window;
static INTEGER Display1_lastcolC, Display1_lastmodeC, Display1_lastx, Display1_lasty, Display1_lastw, Display1_lasth;
static BOOLEAN Display1_width0;
static LONGINT Display1_scrPat[10];


export void Display1_Circle (Display_Frame F, INTEGER col, INTEGER X, INTEGER Y, INTEGER R, INTEGER mode);
static void Display1_CreateGC (void);
export void Display1_Ellipse (Display_Frame F, INTEGER col, INTEGER X, INTEGER Y, INTEGER A, INTEGER B, INTEGER mode);
export void Display1_GetPatternSize (LONGINT pat, INTEGER *w, INTEGER *h);
static void Display1_Init (void);
export void Display1_Line (Display_Frame F, INTEGER col, INTEGER X0, INTEGER Y0, INTEGER X1, INTEGER Y1, INTEGER mode);
static BOOLEAN Display1_Secondary (Display_Frame F, INTEGER col, INTEGER mode);
export LONGINT Display1_ThisPattern (INTEGER n);


void Display1_GetPatternSize (LONGINT pat, INTEGER *w, INTEGER *h)
{
	X11_PatternPtr p = NIL;
	p = (X11_PatternPtr)pat;
	*w = p->w;
	*h = p->h;
}

LONGINT Display1_ThisPattern (INTEGER n)
{
	if (n >= 10) {
		return 0;
	} else {
		return Display1_scrPat[__X(n, 10)];
	}
	__RETCHK;
}

static BOOLEAN Display1_Secondary (Display_Frame F, INTEGER col, INTEGER mode)
{
	X11_Rectangle cliprect = {0};
	if (X11_lclen > 0) {
		X11_FlushLCache();
	}
	Kernel_Lock();
	if (((F->X != Display1_lastx || F->Y != Display1_lasty) || F->W != Display1_lastw) || F->H != Display1_lasth) {
		Display1_lastx = F->X;
		Display1_lasty = F->Y;
		Display1_lastw = F->W;
		Display1_lasth = F->H;
		cliprect.x = Display1_lastx;
		cliprect.y = (X11_Height - Display1_lasty) - Display1_lasth;
		cliprect.w = Display1_lastw;
		cliprect.h = Display1_lasth;
		if (Display1_lasty >= 0) {
			Display1_Window = X11_primary;
		} else {
			Display1_Window = X11_secondary;
			cliprect.y += X11_UBottom;
		}
		if (((cliprect.x <= 0 && cliprect.y <= 0) && cliprect.w >= X11_Width) && cliprect.h >= X11_Height) {
			X11_SetClipMask(X11_display, Display1_GcC, 0);
		} else {
			X11_SetClipRectangles(X11_display, Display1_GcC, 0, 0, (LONGINT)&cliprect, 1, 3);
		}
	}
	if (mode == 1) {
		mode = 0;
	}
	if (mode != Display1_lastmodeC) {
		X11_SetFunction(X11_display, Display1_GcC, X11_function[__X(mode, 3)]);
		Display1_lastmodeC = mode;
	}
	if (col != Display1_lastcolC) {
		X11_SetForeground(X11_display, Display1_GcC, X11_pixelValues[__X(col, 256)]);
		Display1_lastcolC = col;
	}
	Kernel_Unlock();
	return Display1_Window == X11_secondary;
}

void Display1_Line (Display_Frame F, INTEGER col, INTEGER X0, INTEGER Y0, INTEGER X1, INTEGER Y1, INTEGER mode)
{
	if (Display1_Secondary(F, col, mode)) {
		Y0 -= X11_UBottom;
		Y1 -= X11_UBottom;
	}
	Kernel_Lock();
	if (!Display1_width0) {
		Display1_width0 = 1;
		X11_SetLineAttributes(X11_display, Display1_GcC, 0, 0, 1, 0);
	}
	X11_DrawLine(X11_display, Display1_Window, Display1_GcC, X0, (X11_Height - Y0) - 1, X1, (X11_Height - Y1) - 1);
	Kernel_Unlock();
}

void Display1_Circle (Display_Frame F, INTEGER col, INTEGER X, INTEGER Y, INTEGER R, INTEGER mode)
{
	if (R > 0) {
		if (Display1_Secondary(F, col, mode)) {
			Y -= X11_UBottom;
		}
		Kernel_Lock();
		if (Display1_width0) {
			Display1_width0 = 0;
			X11_SetLineAttributes(X11_display, Display1_GcC, 1, 0, 1, 0);
		}
		X11_DrawArc(X11_display, Display1_Window, Display1_GcC, X - R, (X11_Height - Y) - R, __ASHL(R, 1) - 1, __ASHL(R, 1) - 1, 0, 23040);
		Kernel_Unlock();
	}
}

void Display1_Ellipse (Display_Frame F, INTEGER col, INTEGER X, INTEGER Y, INTEGER A, INTEGER B, INTEGER mode)
{
	if (A > 0 && B > 0) {
		if (Display1_Secondary(F, col, mode)) {
			Y -= X11_UBottom;
		}
		Kernel_Lock();
		if (Display1_width0) {
			Display1_width0 = 0;
			X11_SetLineAttributes(X11_display, Display1_GcC, 1, 0, 1, 0);
		}
		X11_DrawArc(X11_display, Display1_Window, Display1_GcC, X - A, (X11_Height - Y) - B, __ASHL(A, 1) - 1, __ASHL(B, 1) - 1, 0, 23040);
		Kernel_Unlock();
	}
}

static void Display1_CreateGC (void)
{
	Kernel_Lock();
	Display1_GcC = X11_CreateGC(X11_display, X11_primary, 0, 0);
	if (Display1_GcC == 0) {
		Console_String((CHAR*)"Cannot create X graphic context", (LONGINT)32);
		Console_Ln();
		Unix_Exit(1);
	}
	X11_SetPlaneMask(X11_display, Display1_GcC, X11_planesMask);
	X11_SetGraphicsExposures(X11_display, Display1_GcC, 1);
	X11_SetBackground(X11_display, Display1_GcC, X11_background);
	X11_SetLineAttributes(X11_display, Display1_GcC, 0, 0, 1, 0);
	X11_SetArcMode(X11_display, Display1_GcC, 1);
	Display1_lastcolC = -1;
	Display1_lastmodeC = -1;
	Display1_width0 = 1;
	Display1_lastx = -1;
	Display1_lasty = -1;
	Display1_lastw = -1;
	Display1_lasth = -1;
	Display1_Window = X11_secondary;
	Kernel_Unlock();
}

static struct Init__5 {
	SET (*image)[17];
	struct Init__5 *lnk;
} *Init__5_s;

static void Repl__6 (INTEGER step);

static void Repl__6 (INTEGER step)
{
	INTEGER i = 0;
	i = step;
	while (i < 16) {
		(*Init__5_s->image)[__X(i + 1, 17)] = (*Init__5_s->image)[__X((i - step) + 1, 17)];
		i += 1;
	}
}

static void Display1_Init (void)
{
	SET image[17];
	struct Init__5 _s;
	_s.image = (void*)image;
	_s.lnk = Init__5_s;
	Init__5_s = &_s;
	image[1] = 0x0;
	Repl__6(1);
	Display1_scrPat[0] = Display_NewPattern((void*)image, 17, 16, 16);
	image[4] = 0x0101;
	image[3] = 0x0;
	image[2] = 0x1010;
	image[1] = 0x0;
	Repl__6(4);
	Display1_scrPat[1] = Display_NewPattern((void*)image, 17, 16, 16);
	image[2] = 0x1111;
	image[1] = 0x4444;
	Repl__6(2);
	Display1_scrPat[2] = Display_NewPattern((void*)image, 17, 16, 16);
	image[1] = 0x5555;
	image[0] = 0xaaaa;
	Repl__6(2);
	Display1_scrPat[3] = Display_NewPattern((void*)image, 17, 16, 16);
	image[2] = 0xeeee;
	image[1] = 0xbbbb;
	Repl__6(2);
	Display1_scrPat[4] = Display_NewPattern((void*)image, 17, 16, 16);
	image[1] = 0xffff;
	Repl__6(1);
	Display1_scrPat[5] = Display_NewPattern((void*)image, 17, 16, 16);
	image[4] = 0x8888;
	image[3] = 0x4444;
	image[2] = 0x2222;
	image[1] = 0x1111;
	Repl__6(4);
	Display1_scrPat[6] = Display_NewPattern((void*)image, 17, 16, 16);
	image[4] = 0x1111;
	image[3] = 0x2222;
	image[2] = 0x4444;
	image[1] = 0x8888;
	Repl__6(4);
	Display1_scrPat[7] = Display_NewPattern((void*)image, 17, 16, 16);
	image[1] = 0x4444;
	Repl__6(1);
	Display1_scrPat[8] = Display_NewPattern((void*)image, 17, 16, 16);
	image[4] = 0x0;
	image[3] = 0x0;
	image[2] = 0x0;
	image[1] = 0xffff;
	Repl__6(4);
	Display1_scrPat[9] = Display_NewPattern((void*)image, 17, 16, 16);
	Init__5_s = _s.lnk;
}


export void *Display1__init(void)
{
	__DEFMOD;
	__IMPORT(Console);
	__IMPORT(Display);
	__IMPORT(Kernel);
	__IMPORT(Unix);
	__IMPORT(X11);
	__REGMOD(Display1, 0);
/* BEGIN */
	Display1_CreateGC();
	Display1_Init();
	__ENDMOD;
}
