/* Ofront 1.4 -xtspkaG */
#include "SYSTEM.h"
#include "Display.h"
#include "Files.h"
#include "In.h"
#include "KeplerFrames.h"
#include "KeplerGraphs.h"
#include "KeplerPorts.h"
#include "Math.h"
#include "Out.h"

typedef
	struct Kepler8_AttrRectDesc *Kepler8_AttrRect;

typedef
	struct Kepler8_AttrRectDesc { /* KeplerGraphs_ConsDesc */
		INTEGER nofpts;
		KeplerGraphs_Star p[4];
		KeplerGraphs_Constellation next;
		INTEGER texture, lineWidth, shadow, shadowWidth, corner;
	} Kepler8_AttrRectDesc;

export void Kepler8_AttrRect_Draw (Kepler8_AttrRect self, KeplerPorts_Port f);
export void Kepler8_AttrRect_Read (Kepler8_AttrRect self, Files_Rider *r, LONGINT *r__typ);
export void Kepler8_AttrRect_Write (Kepler8_AttrRect self, Files_Rider *r, LONGINT *r__typ);

typedef
	struct Kepler8_CircleIntersectDesc *Kepler8_CircleIntersect;

typedef
	struct Kepler8_CircleIntersectDesc { /* KeplerGraphs_PlanetDesc */
		INTEGER x, y, refcnt;
		char _prvt0[2];
		BOOLEAN sel;
		KeplerGraphs_Star next;
		KeplerGraphs_Constellation c;
	} Kepler8_CircleIntersectDesc;

export void Kepler8_CircleIntersect_Calc (Kepler8_CircleIntersect self);

typedef
	struct Kepler8_EllipIntersectDesc *Kepler8_EllipIntersect;

typedef
	struct Kepler8_EllipIntersectDesc { /* KeplerGraphs_PlanetDesc */
		INTEGER x, y, refcnt;
		char _prvt0[2];
		BOOLEAN sel;
		KeplerGraphs_Star next;
		KeplerGraphs_Constellation c;
	} Kepler8_EllipIntersectDesc;

export void Kepler8_EllipIntersect_Calc (Kepler8_EllipIntersect self);

typedef
	struct Kepler8_FilledCircleDesc *Kepler8_FilledCircle;

typedef
	struct Kepler8_FilledCircleDesc { /* KeplerGraphs_ConsDesc */
		INTEGER nofpts;
		KeplerGraphs_Star p[4];
		KeplerGraphs_Constellation next;
		INTEGER texture;
	} Kepler8_FilledCircleDesc;

export void Kepler8_FilledCircle_Draw (Kepler8_FilledCircle self, KeplerPorts_Port f);
export void Kepler8_FilledCircle_Read (Kepler8_FilledCircle self, Files_Rider *r, LONGINT *r__typ);
export void Kepler8_FilledCircle_Write (Kepler8_FilledCircle self, Files_Rider *r, LONGINT *r__typ);

typedef
	struct Kepler8_RectIntersectDesc *Kepler8_RectIntersect;

typedef
	struct Kepler8_RectIntersectDesc { /* KeplerGraphs_PlanetDesc */
		INTEGER x, y, refcnt;
		char _prvt0[2];
		BOOLEAN sel;
		KeplerGraphs_Star next;
		KeplerGraphs_Constellation c;
	} Kepler8_RectIntersectDesc;

export void Kepler8_RectIntersect_Calc (Kepler8_RectIntersect self);



export LONGINT *Kepler8_RectIntersectDesc__typ;
export LONGINT *Kepler8_CircleIntersectDesc__typ;
export LONGINT *Kepler8_EllipIntersectDesc__typ;
export LONGINT *Kepler8_AttrRectDesc__typ;
export LONGINT *Kepler8_FilledCircleDesc__typ;

export void Kepler8_ChangeAttrRect (void);
export void Kepler8_GetAttrRect (void);
static void Kepler8_MinMax (INTEGER a, INTEGER b, INTEGER *min, INTEGER *max);
export void Kepler8_NewAttrRect (void);
export void Kepler8_NewCircleIntersect (void);
export void Kepler8_NewEllipseIntersect (void);
export void Kepler8_NewFilledCircle (void);
export void Kepler8_NewRectIntersect (void);


static void Kepler8_MinMax (INTEGER a, INTEGER b, INTEGER *min, INTEGER *max)
{
	if (a < b) {
		*min = a;
		*max = b;
	} else {
		*min = b;
		*max = a;
	}
}

void Kepler8_RectIntersect_Calc (Kepler8_RectIntersect self)
{
	INTEGER mx = 0, my = 0, x1 = 0, y1 = 0, x2 = 0, y2 = 0;
	REAL slope = 0;
	mx = __ASHR(self->c->p[0]->x + self->c->p[1]->x, 1);
	my = __ASHR(self->c->p[0]->y + self->c->p[1]->y, 1);
	if (mx == self->c->p[2]->x && my == self->c->p[2]->y) {
		self->x = mx;
		self->y = self->c->p[1]->y;
	} else {
		if (self->c->p[2]->x - mx != 0) {
			slope = (self->c->p[2]->y - my) / (REAL)(self->c->p[2]->x - mx);
			if (self->c->p[1]->x != mx && __ABS(slope) > __ABS((self->c->p[1]->y - my) / (REAL)(self->c->p[1]->x - mx))) {
				if (self->c->p[2]->y < my && self->c->p[0]->y < my || self->c->p[2]->y > my && self->c->p[0]->y > my) {
					self->y = self->c->p[0]->y;
					self->x = mx + (int)__ENTIER((self->c->p[0]->y - my) / slope);
				} else {
					self->y = self->c->p[1]->y;
					self->x = mx + (int)__ENTIER((self->c->p[1]->y - my) / slope);
				}
			} else {
				if (self->c->p[2]->y - my != 0) {
					if (self->c->p[2]->x < mx && self->c->p[0]->x < mx || self->c->p[2]->x > mx && self->c->p[0]->x > mx) {
						self->x = self->c->p[0]->x;
						self->y = my + (int)__ENTIER((self->c->p[0]->x - mx) * slope);
					} else {
						self->x = self->c->p[1]->x;
						self->y = my + (int)__ENTIER((self->c->p[1]->x - mx) * slope);
					}
				} else {
					self->y = my;
					if (self->c->p[2]->x < mx && self->c->p[0]->x < mx || self->c->p[2]->x > mx && self->c->p[0]->x > mx) {
						self->x = self->c->p[0]->x;
					} else {
						self->x = self->c->p[1]->x;
					}
				}
			}
		} else {
			self->x = mx;
			if (self->c->p[2]->y < my && self->c->p[0]->y < my || self->c->p[2]->y > my && self->c->p[0]->y > my) {
				self->y = self->c->p[0]->y;
			} else {
				self->y = self->c->p[1]->y;
			}
		}
	}
}

void Kepler8_NewRectIntersect (void)
{
	Kepler8_RectIntersect new = NIL;
	if (KeplerFrames_nofpts >= 3) {
		__NEW(new, Kepler8_RectIntersectDesc);
		__NEW(new->c, KeplerGraphs_ConsDesc);
		new->c->nofpts = 3;
		KeplerFrames_ConsumePoint(&new->c->p[0]);
		KeplerFrames_ConsumePoint(&new->c->p[1]);
		KeplerFrames_ConsumePoint(&new->c->p[2]);
		__KeplerGraphs_Planet_Calc((void*)new);
		__KeplerGraphs_Graph_Append(KeplerFrames_Focus, (void*)new);
		__KeplerGraphs_Graph_FlipSelection(KeplerFrames_Focus, (void*)new);
	}
}

void Kepler8_CircleIntersect_Calc (Kepler8_CircleIntersect self)
{
	REAL factor = 0;
	LONGINT x0 = 0, y0 = 0, x1 = 0, y1 = 0, x2 = 0, y2 = 0;
	x0 = self->c->p[0]->x;
	y0 = self->c->p[0]->y;
	x1 = self->c->p[1]->x;
	y1 = self->c->p[1]->y;
	x2 = self->c->p[2]->x;
	y2 = self->c->p[2]->y;
	if (x0 == x2 && y0 == y2) {
		self->x = (int)x1;
		self->y = (int)y1;
	} else {
		factor = Math_sqrt(((x1 - x0) * (x1 - x0) + (y1 - y0) * (y1 - y0)) / (REAL)((x2 - x0) * (x2 - x0) + (y2 - y0) * (y2 - y0)));
		self->x = (int)x0 + (int)__ENTIER(factor * (x2 - x0));
		self->y = (int)y0 + (int)__ENTIER(factor * (y2 - y0));
	}
}

void Kepler8_NewCircleIntersect (void)
{
	Kepler8_CircleIntersect new = NIL;
	if (KeplerFrames_nofpts >= 3) {
		__NEW(new, Kepler8_CircleIntersectDesc);
		__NEW(new->c, KeplerGraphs_ConsDesc);
		new->c->nofpts = 3;
		KeplerFrames_ConsumePoint(&new->c->p[0]);
		KeplerFrames_ConsumePoint(&new->c->p[1]);
		KeplerFrames_ConsumePoint(&new->c->p[2]);
		__KeplerGraphs_Planet_Calc((void*)new);
		__KeplerGraphs_Graph_Append(KeplerFrames_Focus, (void*)new);
		__KeplerGraphs_Graph_FlipSelection(KeplerFrames_Focus, (void*)new);
	}
}

void Kepler8_EllipIntersect_Calc (Kepler8_EllipIntersect self)
{
	LONGINT a2 = 0, b2 = 0;
	REAL slope2 = 0, temp = 0;
	INTEGER xsign = 0, ysign = 0, t = 0;
	if (self->c->p[3]->x > self->c->p[0]->x) {
		xsign = 1;
	} else {
		xsign = -1;
	}
	if (self->c->p[3]->y > self->c->p[0]->y) {
		ysign = 1;
	} else {
		ysign = -1;
	}
	if (self->c->p[3]->x != self->c->p[0]->x) {
		if (self->c->p[3]->y != self->c->p[0]->y) {
			a2 = self->c->p[1]->x - self->c->p[0]->x;
			a2 = a2 * a2;
			b2 = self->c->p[2]->y - self->c->p[0]->y;
			b2 = b2 * b2;
			t = self->c->p[3]->y - self->c->p[0]->y;
			slope2 = t / (REAL)(self->c->p[3]->x - self->c->p[0]->x);
			slope2 = slope2 * slope2;
			temp = (a2 / (REAL)(b2 + a2 * slope2)) * b2;
			self->x = xsign * (int)__ENTIER(Math_sqrt(temp)) + self->c->p[0]->x;
			self->y = ysign * (int)__ENTIER(Math_sqrt(slope2 * temp)) + self->c->p[0]->y;
		} else {
			t = self->c->p[1]->x - self->c->p[0]->x;
			self->x = self->c->p[0]->x + xsign * t;
			self->y = self->c->p[0]->y;
		}
	} else {
		self->x = self->c->p[0]->x;
		t = self->c->p[2]->y - self->c->p[0]->y;
		self->y = self->c->p[0]->y + ysign * t;
	}
}

void Kepler8_NewEllipseIntersect (void)
{
	Kepler8_EllipIntersect new = NIL;
	if (KeplerFrames_nofpts >= 4) {
		__NEW(new, Kepler8_EllipIntersectDesc);
		__NEW(new->c, KeplerGraphs_ConsDesc);
		new->c->nofpts = 4;
		KeplerFrames_ConsumePoint(&new->c->p[0]);
		KeplerFrames_ConsumePoint(&new->c->p[1]);
		KeplerFrames_ConsumePoint(&new->c->p[2]);
		KeplerFrames_ConsumePoint(&new->c->p[3]);
		__KeplerGraphs_Planet_Calc((void*)new);
		__KeplerGraphs_Graph_Append(KeplerFrames_Focus, (void*)new);
		__KeplerGraphs_Graph_FlipSelection(KeplerFrames_Focus, (void*)new);
	}
}

void Kepler8_AttrRect_Read (Kepler8_AttrRect self, Files_Rider *r, LONGINT *r__typ)
{
	Files_ReadInt(&*r, r__typ, &self->texture);
	Files_ReadInt(&*r, r__typ, &self->lineWidth);
	Files_ReadInt(&*r, r__typ, &self->shadow);
	Files_ReadInt(&*r, r__typ, &self->shadowWidth);
	Files_ReadInt(&*r, r__typ, &self->corner);
	KeplerGraphs_Constellation_Read((void*)self, &*r, r__typ);
}

void Kepler8_AttrRect_Write (Kepler8_AttrRect self, Files_Rider *r, LONGINT *r__typ)
{
	Files_WriteInt(&*r, r__typ, self->texture);
	Files_WriteInt(&*r, r__typ, self->lineWidth);
	Files_WriteInt(&*r, r__typ, self->shadow);
	Files_WriteInt(&*r, r__typ, self->shadowWidth);
	Files_WriteInt(&*r, r__typ, self->corner);
	KeplerGraphs_Constellation_Write((void*)self, &*r, r__typ);
}

void Kepler8_AttrRect_Draw (Kepler8_AttrRect self, KeplerPorts_Port f)
{
	INTEGER x1 = 0, y1 = 0, x2 = 0, y2 = 0;
	Kepler8_MinMax(self->p[0]->x, self->p[1]->x, &x1, &x2);
	Kepler8_MinMax(self->p[0]->y, self->p[1]->y, &y1, &y2);
	if (self->corner > 1) {
		if (self->shadow > 0 && self->shadowWidth > 0) {
			__KeplerPorts_Port_FillCircle(f, (x2 + self->shadowWidth) - self->corner, (y2 - self->shadowWidth) - self->corner, self->corner, 15, self->shadow, 0);
			__KeplerPorts_Port_FillCircle(f, (x1 + self->shadowWidth) + self->corner, (y1 - self->shadowWidth) + self->corner, self->corner, 15, self->shadow, 0);
			__KeplerPorts_Port_FillCircle(f, (x2 + self->shadowWidth) - self->corner, (y1 - self->shadowWidth) + self->corner, self->corner, 15, self->shadow, 0);
			if (self->shadowWidth > self->corner) {
				__KeplerPorts_Port_FillRect(f, x2, (y2 - self->shadowWidth) - self->corner, self->shadowWidth - self->corner, self->corner + f->scale, 15, self->shadow, 0);
				__KeplerPorts_Port_FillRect(f, x1 + self->shadowWidth, (y1 - self->shadowWidth) + self->corner, self->corner, self->shadowWidth - self->corner, 15, self->shadow, 0);
				__KeplerPorts_Port_FillRect(f, (x2 - self->corner) + f->scale, y1 - f->scale, self->corner, self->corner, 15, self->shadow, 0);
			}
			__KeplerPorts_Port_FillRect(f, x2 + f->scale, (y1 - self->shadowWidth) + self->corner, self->shadowWidth, (y2 - y1) - __ASHL(self->corner, 1), 15, self->shadow, 0);
			__KeplerPorts_Port_FillRect(f, (x1 + self->shadowWidth) + self->corner, (y1 - self->shadowWidth) - f->scale, (x2 - x1) - __ASHL(self->corner, 1), self->shadowWidth, 15, self->shadow, 0);
		}
		__KeplerPorts_Port_FillCircle(f, x1 + self->corner, y1 + self->corner, self->corner, 15, 5, 0);
		__KeplerPorts_Port_FillCircle(f, x2 - self->corner, y1 + self->corner, self->corner, 15, 5, 0);
		__KeplerPorts_Port_FillCircle(f, x2 - self->corner, y2 - self->corner, self->corner, 15, 5, 0);
		__KeplerPorts_Port_FillCircle(f, x1 + self->corner, y2 - self->corner, self->corner, 15, 5, 0);
		if (self->corner > self->lineWidth) {
			__KeplerPorts_Port_FillCircle(f, x1 + self->corner, y1 + self->corner, self->corner - self->lineWidth, 15, self->texture, 0);
			__KeplerPorts_Port_FillCircle(f, x2 - self->corner, y1 + self->corner, self->corner - self->lineWidth, 15, self->texture, 0);
			__KeplerPorts_Port_FillCircle(f, x2 - self->corner, y2 - self->corner, self->corner - self->lineWidth, 15, self->texture, 0);
			__KeplerPorts_Port_FillCircle(f, x1 + self->corner, y2 - self->corner, self->corner - self->lineWidth, 15, self->texture, 0);
		}
		__KeplerPorts_Port_FillRect(f, (x1 + self->lineWidth) - f->scale, y1 + self->corner, ((x2 - x1) - __ASHL(self->lineWidth, 1)) + __ASHL(f->scale, 1), (y2 - y1) - __ASHL(self->corner, 1), 15, self->texture, 0);
		__KeplerPorts_Port_FillRect(f, x1 + self->corner, (y1 + self->lineWidth) - f->scale, (x2 - x1) - __ASHL(self->corner, 1), ((y2 - y1) - __ASHL(self->lineWidth, 1)) + __ASHL(f->scale, 1), 15, self->texture, 0);
		__KeplerPorts_Port_FillRect(f, x1 + self->corner, y1 - f->scale, (x2 - x1) - __ASHL(self->corner, 1), (self->lineWidth + f->scale) - 1, 15, 5, 0);
		__KeplerPorts_Port_FillRect(f, x1 + self->corner, (y2 - self->lineWidth) + f->scale, (x2 - x1) - __ASHL(self->corner, 1), (self->lineWidth + f->scale) - 1, 15, 5, 0);
		__KeplerPorts_Port_FillRect(f, x1 - f->scale, y1 + self->corner, (self->lineWidth + f->scale) - 1, (y2 - y1) - __ASHL(self->corner, 1), 15, 5, 0);
		__KeplerPorts_Port_FillRect(f, (x2 - self->lineWidth) + f->scale, y1 + self->corner, (self->lineWidth + f->scale) - 1, (y2 - y1) - __ASHL(self->corner, 1), 15, 5, 0);
	} else {
		__KeplerPorts_Port_FillRect(f, x2, y1 - self->shadowWidth, self->shadowWidth, y2 - y1, 15, self->shadow, 0);
		__KeplerPorts_Port_FillRect(f, x1 + self->shadowWidth, y1 - self->shadowWidth, x2 - x1, self->shadowWidth, 15, self->shadow, 0);
		__KeplerPorts_Port_FillRect(f, x1 + self->lineWidth, y1 + self->lineWidth, (x2 - x1) - __ASHL(self->lineWidth, 1), (y2 - y1) - __ASHL(self->lineWidth, 1), 15, self->texture, 0);
		__KeplerPorts_Port_FillRect(f, x1, y1, x2 - x1, self->lineWidth, 15, 5, 0);
		__KeplerPorts_Port_FillRect(f, x1, y2 - self->lineWidth, x2 - x1, self->lineWidth, 15, 5, 0);
		__KeplerPorts_Port_FillRect(f, x1, y1, self->lineWidth, y2 - y1, 15, 5, 0);
		__KeplerPorts_Port_FillRect(f, x2 - self->lineWidth, y1, self->lineWidth, y2 - y1, 15, 5, 0);
	}
}

void Kepler8_NewAttrRect (void)
{
	Kepler8_AttrRect new = NIL;
	INTEGER texture = 0, lineWidth = 0, shadow = 0, shadowWidth = 0, corner = 0;
	if (KeplerFrames_nofpts >= 2) {
		__NEW(new, Kepler8_AttrRectDesc);
		new->nofpts = 2;
		In_Open();
		In_Int(&texture);
		if (texture < 0) {
			new->texture = 0;
		} else {
			new->texture = texture;
		}
		In_Int(&lineWidth);
		if (lineWidth < 0) {
			new->lineWidth = 0;
		} else {
			new->lineWidth = lineWidth;
		}
		In_Int(&shadow);
		if (shadow < 0) {
			new->shadow = 0;
		} else {
			new->shadow = shadow;
		}
		In_Int(&shadowWidth);
		if (shadowWidth < 0) {
			new->shadowWidth = 0;
		} else {
			new->shadowWidth = shadowWidth;
		}
		In_Int(&corner);
		if (corner <= 1) {
			new->corner = 0;
		} else {
			new->corner = corner;
		}
		if (In_Done) {
			KeplerFrames_ConsumePoint(&new->p[0]);
			KeplerFrames_ConsumePoint(&new->p[1]);
			__KeplerGraphs_Graph_Append(KeplerFrames_Focus, (void*)new);
		}
	}
}

void Kepler8_ChangeAttrRect (void)
{
	KeplerGraphs_Graph G = NIL;
	KeplerGraphs_Constellation c = NIL;
	INTEGER i1 = 0, i2 = 0, i3 = 0, i4 = 0, i5 = 0;
	KeplerPorts_BalloonPort F = NIL;
	In_Open();
	In_Int(&i1);
	In_Int(&i2);
	In_Int(&i3);
	In_Int(&i4);
	In_Int(&i5);
	KeplerFrames_GetSelection(&G);
	if (G != NIL && In_Done) {
		__NEW(F, KeplerPorts_BalloonPortDesc);
		KeplerPorts_InitBalloon(F);
		c = G->cons;
		while (c != NIL) {
			if (__ISP(c, Kepler8_AttrRectDesc, 2)) {
				if (__KeplerGraphs_Constellation_State((void*)(*(Kepler8_AttrRect*)&c)) == 2) {
					__KeplerGraphs_Object_Draw((void*)(*(Kepler8_AttrRect*)&c), (void*)F);
					(*(Kepler8_AttrRect*)&c)->texture = i1;
					(*(Kepler8_AttrRect*)&c)->lineWidth = i2;
					(*(Kepler8_AttrRect*)&c)->shadow = i3;
					(*(Kepler8_AttrRect*)&c)->shadowWidth = i4;
					(*(Kepler8_AttrRect*)&c)->corner = i5;
					__KeplerGraphs_Object_Draw((void*)(*(Kepler8_AttrRect*)&c), (void*)F);
				}
			}
			c = c->next;
		}
		(*G->notify)(1, G, NIL, (void*)F);
	}
}

void Kepler8_GetAttrRect (void)
{
	KeplerGraphs_Graph G = NIL;
	KeplerGraphs_Constellation c = NIL;
	KeplerFrames_GetSelection(&G);
	if (G != NIL) {
		c = G->cons;
		while (c != NIL) {
			if (__ISP(c, Kepler8_AttrRectDesc, 2)) {
				if (__KeplerGraphs_Constellation_State((void*)(*(Kepler8_AttrRect*)&c)) == 2) {
					Out_String((CHAR*)"Kepler8.ChangeAttrRect ", (LONGINT)24);
					Out_Int((*(Kepler8_AttrRect*)&c)->texture, 5);
					Out_Int((*(Kepler8_AttrRect*)&c)->lineWidth, 5);
					Out_Int((*(Kepler8_AttrRect*)&c)->shadow, 5);
					Out_Int((*(Kepler8_AttrRect*)&c)->shadowWidth, 5);
					Out_Int((*(Kepler8_AttrRect*)&c)->corner, 5);
					Out_Ln();
				}
			}
			c = c->next;
		}
	}
}

void Kepler8_FilledCircle_Read (Kepler8_FilledCircle self, Files_Rider *r, LONGINT *r__typ)
{
	Files_ReadInt(&*r, r__typ, &self->texture);
	KeplerGraphs_Constellation_Read((void*)self, &*r, r__typ);
}

void Kepler8_FilledCircle_Write (Kepler8_FilledCircle self, Files_Rider *r, LONGINT *r__typ)
{
	Files_WriteInt(&*r, r__typ, self->texture);
	KeplerGraphs_Constellation_Write((void*)self, &*r, r__typ);
}

void Kepler8_FilledCircle_Draw (Kepler8_FilledCircle self, KeplerPorts_Port f)
{
	LONGINT rx = 0, ry = 0;
	INTEGER r = 0;
	rx = self->p[1]->x - self->p[0]->x;
	ry = self->p[1]->y - self->p[0]->y;
	r = (int)__ENTIER(Math_sqrt(rx * rx + ry * ry));
	__KeplerPorts_Port_FillCircle(f, self->p[0]->x, self->p[0]->y, r, 15, self->texture, 0);
}

void Kepler8_NewFilledCircle (void)
{
	Kepler8_FilledCircle new = NIL;
	INTEGER texture = 0;
	In_Open();
	In_Int(&texture);
	if (((KeplerFrames_nofpts >= 2 && In_Done) && texture >= 0) && texture < 10) {
		__NEW(new, Kepler8_FilledCircleDesc);
		new->nofpts = 2;
		new->texture = texture;
		KeplerFrames_ConsumePoint(&new->p[0]);
		KeplerFrames_ConsumePoint(&new->p[1]);
		__KeplerGraphs_Graph_Append(KeplerFrames_Focus, (void*)new);
	}
}

__TDESC(Kepler8_RectIntersectDesc, 5, 2) = {__TDFLDS("RectIntersectDesc", 20), {12, 16, -12}};
__TDESC(Kepler8_CircleIntersectDesc, 5, 2) = {__TDFLDS("CircleIntersectDesc", 20), {12, 16, -12}};
__TDESC(Kepler8_EllipIntersectDesc, 5, 2) = {__TDFLDS("EllipIntersectDesc", 20), {12, 16, -12}};
__TDESC(Kepler8_AttrRectDesc, 5, 5) = {__TDFLDS("AttrRectDesc", 36), {4, 8, 12, 16, 20, -24}};
__TDESC(Kepler8_FilledCircleDesc, 5, 5) = {__TDFLDS("FilledCircleDesc", 28), {4, 8, 12, 16, 20, -24}};

export void *Kepler8__init(void)
{
	__DEFMOD;
	__IMPORT(Display);
	__IMPORT(Files);
	__IMPORT(In);
	__IMPORT(KeplerFrames);
	__IMPORT(KeplerGraphs);
	__IMPORT(KeplerPorts);
	__IMPORT(Math);
	__IMPORT(Out);
	__REGMOD(Kepler8, 0);
	__REGCMD("ChangeAttrRect", Kepler8_ChangeAttrRect);
	__REGCMD("GetAttrRect", Kepler8_GetAttrRect);
	__REGCMD("NewAttrRect", Kepler8_NewAttrRect);
	__REGCMD("NewCircleIntersect", Kepler8_NewCircleIntersect);
	__REGCMD("NewEllipseIntersect", Kepler8_NewEllipseIntersect);
	__REGCMD("NewFilledCircle", Kepler8_NewFilledCircle);
	__REGCMD("NewRectIntersect", Kepler8_NewRectIntersect);
	__INITYP(Kepler8_RectIntersectDesc, KeplerGraphs_PlanetDesc, 3);
	__INITBP(Kepler8_RectIntersectDesc, Kepler8_RectIntersect_Calc, 3);
	__INITYP(Kepler8_CircleIntersectDesc, KeplerGraphs_PlanetDesc, 3);
	__INITBP(Kepler8_CircleIntersectDesc, Kepler8_CircleIntersect_Calc, 3);
	__INITYP(Kepler8_EllipIntersectDesc, KeplerGraphs_PlanetDesc, 3);
	__INITBP(Kepler8_EllipIntersectDesc, Kepler8_EllipIntersect_Calc, 3);
	__INITYP(Kepler8_AttrRectDesc, KeplerGraphs_ConsDesc, 2);
	__INITBP(Kepler8_AttrRectDesc, Kepler8_AttrRect_Draw, 0);
	__INITBP(Kepler8_AttrRectDesc, Kepler8_AttrRect_Read, 1);
	__INITBP(Kepler8_AttrRectDesc, Kepler8_AttrRect_Write, 2);
	__INITYP(Kepler8_FilledCircleDesc, KeplerGraphs_ConsDesc, 2);
	__INITBP(Kepler8_FilledCircleDesc, Kepler8_FilledCircle_Draw, 0);
	__INITBP(Kepler8_FilledCircleDesc, Kepler8_FilledCircle_Read, 1);
	__INITBP(Kepler8_FilledCircleDesc, Kepler8_FilledCircle_Write, 2);
/* BEGIN */
	__ENDMOD;
}
