/* Ofront 1.4 -xtspkaG */
#include "SYSTEM.h"
#include "Display.h"
#include "Input.h"
#include "Oberon.h"
#include "Viewers.h"

typedef
	struct MenuViewers_ModifyMsg { /* Display_FrameMsg */
		INTEGER id, dY, Y, H;
	} MenuViewers_ModifyMsg;

typedef
	struct MenuViewers_ViewerDesc *MenuViewers_Viewer;

typedef
	struct MenuViewers_ViewerDesc { /* Viewers_ViewerDesc */
		Display_Frame dsc, next;
		INTEGER X, Y, W, H;
		Display_Handler handle;
		INTEGER state;
		INTEGER menuH;
	} MenuViewers_ViewerDesc;


export MenuViewers_Viewer MenuViewers_Ancestor;

export LONGINT *MenuViewers_ViewerDesc__typ;
export LONGINT *MenuViewers_ModifyMsg__typ;

static void MenuViewers_Adjust (Display_Frame F, INTEGER id, INTEGER dY, INTEGER Y, INTEGER H);
static void MenuViewers_Change (MenuViewers_Viewer V, INTEGER X, INTEGER Y, SET Keys);
static void MenuViewers_Copy (MenuViewers_Viewer V, MenuViewers_Viewer *V1);
static void MenuViewers_Draw (Viewers_Viewer V);
static void MenuViewers_Extend (MenuViewers_Viewer V, INTEGER newY);
static void MenuViewers_Grow (MenuViewers_Viewer V, INTEGER oldH);
export void MenuViewers_Handle (Display_Frame V, Display_FrameMsg *M, LONGINT *M__typ);
static void MenuViewers_Modify (MenuViewers_Viewer V, INTEGER Y, INTEGER H);
export MenuViewers_Viewer MenuViewers_New (Display_Frame Menu, Display_Frame Main, INTEGER menuH, INTEGER X, INTEGER Y);
static void MenuViewers_Reduce (MenuViewers_Viewer V, INTEGER newY);
static void MenuViewers_Restore (MenuViewers_Viewer V);
static void MenuViewers_Shrink (MenuViewers_Viewer V, INTEGER newH);
static void MenuViewers_Suspend (MenuViewers_Viewer V);


static void MenuViewers_Copy (MenuViewers_Viewer V, MenuViewers_Viewer *V1)
{
	Display_Frame Menu = NIL, Main = NIL;
	Oberon_CopyMsg M = {0};
	Menu = V->dsc;
	Main = V->dsc->next;
	__NEW(*V1, MenuViewers_ViewerDesc);
	__GUARDEQP(*V1, MenuViewers_ViewerDesc) = *V;
	(*V1)->state = 0;
	M.F = NIL;
	(*Menu->handle)(Menu, (void*)&M, Oberon_CopyMsg__typ);
	(*V1)->dsc = M.F;
	M.F = NIL;
	(*Main->handle)(Main, (void*)&M, Oberon_CopyMsg__typ);
	(*V1)->dsc->next = M.F;
}

static void MenuViewers_Draw (Viewers_Viewer V)
{
	Display_ReplConst(15, V->X, V->Y, 1, V->H, 0);
	Display_ReplConst(15, (V->X + V->W) - 1, V->Y, 1, V->H, 0);
	Display_ReplConst(15, V->X + 1, V->Y, V->W - 2, 1, 0);
	Display_ReplConst(15, V->X + 1, (V->Y + V->H) - 1, V->W - 2, 1, 0);
}

static void MenuViewers_Extend (MenuViewers_Viewer V, INTEGER newY)
{
	INTEGER dH = 0;
	dH = V->Y - newY;
	if (dH > 0) {
		Display_ReplConst(0, V->X + 1, newY + 1, V->W - 2, dH, 0);
		Display_ReplConst(15, V->X, newY, 1, dH, 0);
		Display_ReplConst(15, (V->X + V->W) - 1, newY, 1, dH, 0);
		Display_ReplConst(15, V->X + 1, newY, V->W - 2, 1, 0);
	}
}

static void MenuViewers_Reduce (MenuViewers_Viewer V, INTEGER newY)
{
	Display_ReplConst(15, V->X + 1, newY, V->W - 2, 1, 0);
}

static void MenuViewers_Grow (MenuViewers_Viewer V, INTEGER oldH)
{
	INTEGER dH = 0;
	dH = V->H - oldH;
	if (dH > 0) {
		Display_ReplConst(15, V->X, V->Y + oldH, 1, dH, 0);
		Display_ReplConst(15, (V->X + V->W) - 1, V->Y + oldH, 1, dH, 0);
		Display_ReplConst(15, V->X + 1, (V->Y + V->H) - 1, V->W - 2, 1, 0);
	}
}

static void MenuViewers_Shrink (MenuViewers_Viewer V, INTEGER newH)
{
	Display_ReplConst(15, V->X + 1, (V->Y + newH) - 1, V->W - 2, 1, 0);
}

static void MenuViewers_Adjust (Display_Frame F, INTEGER id, INTEGER dY, INTEGER Y, INTEGER H)
{
	MenuViewers_ModifyMsg M = {0};
	M.id = id;
	M.dY = dY;
	M.Y = Y;
	M.H = H;
	(*F->handle)(F, (void*)&M, MenuViewers_ModifyMsg__typ);
	F->Y = Y;
	F->H = H;
}

static void MenuViewers_Restore (MenuViewers_Viewer V)
{
	Display_Frame Menu = NIL, Main = NIL;
	Menu = V->dsc;
	Main = V->dsc->next;
	Oberon_RemoveMarks(V->X, V->Y, V->W, V->H);
	MenuViewers_Draw((void*)V);
	Menu->X = V->X + 1;
	Menu->Y = (V->Y + V->H) - 1;
	Menu->W = V->W - 2;
	Menu->H = 0;
	Main->X = V->X + 1;
	Main->Y = (V->Y + V->H) - V->menuH;
	Main->W = V->W - 2;
	Main->H = 0;
	if (V->H > V->menuH + 1) {
		MenuViewers_Adjust(Menu, 0, 0, (V->Y + V->H) - V->menuH, V->menuH - 1);
		MenuViewers_Adjust(Main, 0, 0, V->Y + 1, (V->H - V->menuH) - 1);
	} else {
		MenuViewers_Adjust(Menu, 0, 0, V->Y + 1, V->H - 2);
	}
}

static void MenuViewers_Modify (MenuViewers_Viewer V, INTEGER Y, INTEGER H)
{
	Display_Frame Menu = NIL, Main = NIL;
	Menu = V->dsc;
	Main = V->dsc->next;
	if (Y < V->Y) {
		Oberon_RemoveMarks(V->X, Y, V->W, V->Y - Y);
		MenuViewers_Extend(V, Y);
		if (H > V->menuH + 1) {
			MenuViewers_Adjust(Menu, 0, 0, (Y + H) - V->menuH, V->menuH - 1);
			MenuViewers_Adjust(Main, 0, 0, Y + 1, (H - V->menuH) - 1);
		} else {
			MenuViewers_Adjust(Menu, 0, 0, Y + 1, H - 2);
		}
	} else if (Y > V->Y) {
		Oberon_RemoveMarks(V->X, V->Y, V->W, V->H);
		if (H > V->menuH + 1) {
			MenuViewers_Adjust(Main, 1, 0, Y + 1, (H - V->menuH) - 1);
			MenuViewers_Adjust(Menu, 1, 0, (Y + H) - V->menuH, V->menuH - 1);
		} else {
			MenuViewers_Adjust(Main, 1, 0, (Y + H) - V->menuH, 0);
			MenuViewers_Adjust(Menu, 1, 0, Y + 1, H - 2);
		}
		MenuViewers_Reduce(V, Y);
	}
}

static void MenuViewers_Change (MenuViewers_Viewer V, INTEGER X, INTEGER Y, SET Keys)
{
	Display_Frame Menu = NIL, Main = NIL;
	Viewers_Viewer V1 = NIL;
	SET keysum = 0;
	INTEGER Y0 = 0, dY = 0, H = 0;
	Menu = V->dsc;
	Main = V->dsc->next;
	Oberon_DrawCursor(&Oberon_Mouse, Oberon_Cursor__typ, &Oberon_Arrow, Oberon_Marker__typ, X, Y);
	Display_ReplConst(15, V->X + 1, (V->Y + V->H) - V->dsc->H, V->W - 2, V->dsc->H - 1, 2);
	Y0 = Y;
	keysum = Keys;
	for (;;) {
		Input_Mouse(&Keys, &X, &Y);
		if (Keys == 0x0) {
			break;
		}
		keysum = keysum | Keys;
		Oberon_DrawCursor(&Oberon_Mouse, Oberon_Cursor__typ, &Oberon_Arrow, Oberon_Marker__typ, X, Y);
	}
	Display_ReplConst(15, V->X + 1, (V->Y + V->H) - V->dsc->H, V->W - 2, V->dsc->H - 1, 2);
	if (!__IN(0, keysum)) {
		if (__IN(1, keysum)) {
			V1 = Viewers_This(X, Y);
			if (__ISP(V1, MenuViewers_ViewerDesc, 2) && Y > ((V1->Y + V1->H) - __GUARDP(V1, MenuViewers_ViewerDesc, 2)->menuH) - 2) {
				Y = V1->Y + V1->H;
			}
			if (Y < (V1->Y + V->menuH) + 2) {
				Y = (V1->Y + V->menuH) + 2;
			}
			Viewers_Close((void*)V);
			Viewers_Open((void*)V, X, Y);
			MenuViewers_Restore(V);
		} else {
			if (Y > Y0) {
				dY = Y - Y0;
				V1 = Viewers_Next((void*)V);
				if (V1->state > 1) {
					if (__ISP(V1, MenuViewers_ViewerDesc, 2)) {
						if (V1->H < __GUARDP(V1, MenuViewers_ViewerDesc, 2)->menuH + 2) {
							dY = 0;
						} else if (V1->H < (__GUARDP(V1, MenuViewers_ViewerDesc, 2)->menuH + 2) + dY) {
							dY = (V1->H - __GUARDP(V1, MenuViewers_ViewerDesc, 2)->menuH) - 2;
						}
					} else if (V1->H < 1 + dY) {
						dY = V1->H - 1;
					}
				} else if (V1->H < dY) {
					dY = V1->H;
				}
				Viewers_Change((void*)V, (V->Y + V->H) + dY);
				Oberon_RemoveMarks(V->X, V->Y, V->W, V->H);
				MenuViewers_Grow(V, V->H - dY);
				if (V->H > V->menuH + 1) {
					MenuViewers_Adjust(Menu, 0, dY, (V->Y + V->H) - V->menuH, V->menuH - 1);
					MenuViewers_Adjust(Main, 0, dY, V->Y + 1, (V->H - V->menuH) - 1);
				} else {
					MenuViewers_Adjust(Menu, 0, dY, V->Y + 1, V->H - 2);
					MenuViewers_Adjust(Main, 0, dY, (V->Y + V->H) - V->menuH, 0);
				}
			} else if (Y < Y0) {
				dY = Y0 - Y;
				if (V->H >= V->menuH + 2) {
					if (V->H < (V->menuH + 2) + dY) {
						dY = (V->H - V->menuH) - 2;
					}
					Oberon_RemoveMarks(V->X, V->Y, V->W, V->H);
					H = V->H - dY;
					MenuViewers_Adjust(Main, 1, dY, V->Y + 1, (H - V->menuH) - 1);
					MenuViewers_Adjust(Menu, 1, dY, (V->Y + H) - V->menuH, V->menuH - 1);
					MenuViewers_Shrink(V, H);
					Viewers_Change((void*)V, V->Y + H);
				}
			}
		}
	}
}

static void MenuViewers_Suspend (MenuViewers_Viewer V)
{
	Display_Frame Menu = NIL, Main = NIL;
	Menu = V->dsc;
	Main = V->dsc->next;
	MenuViewers_Adjust(Main, 1, 0, (V->Y + V->H) - V->menuH, 0);
	MenuViewers_Adjust(Menu, 1, 0, (V->Y + V->H) - 1, 0);
}

void MenuViewers_Handle (Display_Frame V, Display_FrameMsg *M, LONGINT *M__typ)
{
	Display_Frame Menu = NIL, Main = NIL;
	MenuViewers_Viewer V1 = NIL;
	if (__ISP(V, MenuViewers_ViewerDesc, 2)) {
		MenuViewers_Ancestor = (*(MenuViewers_Viewer*)&V);
		Menu = (*(MenuViewers_Viewer*)&V)->dsc;
		Main = (*(MenuViewers_Viewer*)&V)->dsc->next;
		if (__IS(M__typ, Oberon_InputMsg, 1)) {
			if (__IS(M__typ, Oberon_InputMsg, 1)) {
				Oberon_InputMsg *M__ = (void*)M;
				if ((*M__).id == 1) {
					if ((*M__).Y < (*(MenuViewers_Viewer*)&V)->Y + 1) {
						Oberon_DrawCursor(&Oberon_Mouse, Oberon_Cursor__typ, &Oberon_Arrow, Oberon_Marker__typ, (*M__).X, (*M__).Y);
					} else if ((*M__).Y < ((*(MenuViewers_Viewer*)&V)->Y + (*(MenuViewers_Viewer*)&V)->H) - (*(MenuViewers_Viewer*)&V)->menuH) {
						(*Main->handle)(Main, (void*)&*M__, M__typ);
					} else if ((*M__).Y < (((*(MenuViewers_Viewer*)&V)->Y + (*(MenuViewers_Viewer*)&V)->H) - (*(MenuViewers_Viewer*)&V)->menuH) + 2) {
						(*Menu->handle)(Menu, (void*)&*M__, M__typ);
					} else if ((*M__).Y < ((*(MenuViewers_Viewer*)&V)->Y + (*(MenuViewers_Viewer*)&V)->H) - 1) {
						if (__IN(2, (*M__).keys)) {
							MenuViewers_Change((*(MenuViewers_Viewer*)&V), (*M__).X, (*M__).Y, (*M__).keys);
						} else {
							(*Menu->handle)(Menu, (void*)&*M__, M__typ);
						}
					} else {
						Oberon_DrawCursor(&Oberon_Mouse, Oberon_Cursor__typ, &Oberon_Arrow, Oberon_Marker__typ, (*M__).X, (*M__).Y);
					}
				} else {
					(*Menu->handle)(Menu, (void*)&*M__, M__typ);
					(*Main->handle)(Main, (void*)&*M__, M__typ);
				}
			} else __WITHCHK;
		} else if (__IS(M__typ, Oberon_ControlMsg, 1)) {
			if (__IS(M__typ, Oberon_ControlMsg, 1)) {
				Oberon_ControlMsg *M__ = (void*)M;
				if ((*M__).id == 2) {
					Oberon_DrawCursor(&Oberon_Mouse, Oberon_Cursor__typ, &Oberon_Arrow, Oberon_Marker__typ, (*M__).X, (*M__).Y);
					Oberon_DrawCursor(&Oberon_Pointer, Oberon_Cursor__typ, &Oberon_Star, Oberon_Marker__typ, (*M__).X, (*M__).Y);
				} else {
					(*Menu->handle)(Menu, (void*)&*M__, M__typ);
					(*Main->handle)(Main, (void*)&*M__, M__typ);
				}
			} else __WITHCHK;
		} else if (__IS(M__typ, Oberon_CopyMsg, 1)) {
			if (__IS(M__typ, Oberon_CopyMsg, 1)) {
				Oberon_CopyMsg *M__ = (void*)M;
				MenuViewers_Copy((*(MenuViewers_Viewer*)&V), &V1);
				(*M__).F = (Display_Frame)V1;
			} else __WITHCHK;
		} else if (__IS(M__typ, Viewers_ViewerMsg, 1)) {
			if (__IS(M__typ, Viewers_ViewerMsg, 1)) {
				Viewers_ViewerMsg *M__ = (void*)M;
				if ((*M__).id == 0) {
					MenuViewers_Restore((*(MenuViewers_Viewer*)&V));
				} else if ((*M__).id == 1) {
					MenuViewers_Modify((*(MenuViewers_Viewer*)&V), (*M__).Y, (*M__).H);
				} else if ((*M__).id == 2) {
					MenuViewers_Suspend((*(MenuViewers_Viewer*)&V));
				}
			} else __WITHCHK;
		} else {
			(*Menu->handle)(Menu, &*M, M__typ);
			(*Main->handle)(Main, &*M, M__typ);
		}
	} else __WITHCHK;
}

MenuViewers_Viewer MenuViewers_New (Display_Frame Menu, Display_Frame Main, INTEGER menuH, INTEGER X, INTEGER Y)
{
	MenuViewers_Viewer V = NIL;
	__NEW(V, MenuViewers_ViewerDesc);
	V->handle = MenuViewers_Handle;
	V->dsc = Menu;
	V->dsc->next = Main;
	V->menuH = menuH;
	Viewers_Open((void*)V, X, Y);
	MenuViewers_Restore(V);
	return V;
}

static void EnumPtrs(void (*P)(void*))
{
	P(MenuViewers_Ancestor);
}

__TDESC(MenuViewers_ViewerDesc, 1, 2) = {__TDFLDS("ViewerDesc", 24), {0, 4, -12}};
__TDESC(MenuViewers_ModifyMsg, 1, 0) = {__TDFLDS("ModifyMsg", 8), {-4}};

export void *MenuViewers__init(void)
{
	__DEFMOD;
	__IMPORT(Display);
	__IMPORT(Input);
	__IMPORT(Oberon);
	__IMPORT(Viewers);
	__REGMOD(MenuViewers, EnumPtrs);
	__INITYP(MenuViewers_ViewerDesc, Viewers_ViewerDesc, 2);
	__INITYP(MenuViewers_ModifyMsg, Display_FrameMsg, 1);
/* BEGIN */
	__ENDMOD;
}
