/* Ofront 1.4 -xtspkaG */
#include "SYSTEM.h"
#include "Display.h"
#include "Elems.h"
#include "Files.h"
#include "Fonts.h"
#include "GUtils.h"
#include "Input.h"
#include "MenuViewers.h"
#include "Oberon.h"
#include "Out.h"
#include "PanelElems.h"
#include "Strings.h"
#include "TextFrames.h"
#include "Texts.h"
#include "Viewers.h"

typedef
	struct Panel_FrameDesc *Panel_Frame;

typedef
	struct Panel_FrameDesc { /* Display_FrameDesc */
		Display_Frame dsc, next;
		INTEGER X, Y, W, H;
		Display_Handler handle;
		PanelElems_Panel p;
		Texts_Text txt;
		INTEGER px, py;
		BOOLEAN hasCar, hasSel;
		LONGINT selTime;
	} Panel_FrameDesc;


static Texts_Writer Panel_W;

export LONGINT *Panel_FrameDesc__typ;

static void Panel_CopyInto (Panel_Frame f, Texts_Text txt, LONGINT beg);
static void Panel_Draw (Panel_Frame f);
static void Panel_DrawPanel (Panel_Frame f);
static void Panel_DrawPanelArea (Panel_Frame f, INTEGER mode);
static void Panel_Extend (Panel_Frame f, INTEGER newY);
static void Panel_Forward (Panel_Frame f, Display_FrameMsg *msg, LONGINT *msg__typ);
static void Panel_GetMenu (PanelElems_Panel p, CHAR *mnu, LONGINT mnu__len);
static void Panel_Handle (Display_Frame f, Display_FrameMsg *msg, LONGINT *msg__typ);
static void Panel_HandleInput (Panel_Frame f, Oberon_InputMsg *msg, LONGINT *msg__typ);
static BOOLEAN Panel_InPanel (Panel_Frame f, INTEGER x, INTEGER y);
static void Panel_InvertCaret (Panel_Frame f);
export void Panel_LoadPanel (CHAR *name, LONGINT name__len, PanelElems_Panel *p, Fonts_Font *fnt, SHORTINT *col);
static INTEGER Panel_Min (INTEGER a, INTEGER b);
static void Panel_Modify (Panel_Frame f, INTEGER id, INTEGER dY, INTEGER Y, INTEGER H);
export void Panel_Open (void);
export void Panel_OpenPanel (PanelElems_Panel p, CHAR *name, LONGINT name__len, Fonts_Font fnt, SHORTINT col);
static void Panel_Reduce (Panel_Frame f, INTEGER newY);
static void Panel_RemoveCaret (Panel_Frame f);
static void Panel_SetCaret (Panel_Frame f);
static void Panel_SetPanel (Panel_Frame f);
export void Panel_Store (void);


static void Panel_GetMenu (PanelElems_Panel p, CHAR *mnu, LONGINT mnu__len)
{
	Files_File f = NIL;
	if (p != NIL) {
		Elems_GetString((void*)p, (CHAR*)"Menu", (LONGINT)5, (void*)mnu, mnu__len);
	}
	if (p != NIL && Elems_Done) {
		f = Files_Old(mnu, mnu__len);
		if (f != NIL) {
			Strings_Insert((CHAR*)"^", (LONGINT)2, 0, (void*)mnu, mnu__len);
		}
	}
	if ((p == NIL || !Elems_Done) || f == NIL) {
		f = Files_Old((CHAR*)"Panel.Menu.Text", (LONGINT)16);
		if (f == NIL) {
			__COPY("System.Close  System.Copy  System.Grow  Panel.Store ", mnu, mnu__len);
		} else {
			__COPY("^Panel.Menu.Text", mnu, mnu__len);
		}
	}
}

static INTEGER Panel_Min (INTEGER a, INTEGER b)
{
	if (a < b) {
		return a;
	} else {
		return b;
	}
	__RETCHK;
}

static void Panel_InvertCaret (Panel_Frame f)
{
	GUtils_Frame((void*)f, 0, f->X, f->Y, f->W, f->H, 3, 2);
}

static void Panel_SetCaret (Panel_Frame f)
{
	Oberon_PassFocus(Viewers_This(f->X, f->Y));
	f->hasCar = 1;
	Panel_InvertCaret(f);
}

static void Panel_RemoveCaret (Panel_Frame f)
{
	if (f->hasCar) {
		Panel_InvertCaret(f);
		f->hasCar = 0;
	}
}

static void Panel_DrawPanel (Panel_Frame f)
{
	TextFrames_DisplayMsg disp = {0};
	Elems_GetLook((void*)f->p, &disp.fnt, &disp.col);
	disp.pos = 0;
	disp.frame = (Display_Frame)f;
	disp.X0 = f->X + f->px;
	disp.Y0 = f->py;
	disp.indent = 0;
	disp.elemFrame = NIL;
	disp.prepare = 1;
	(*f->p->handle)((void*)f->p, (void*)&disp, TextFrames_DisplayMsg__typ);
	disp.prepare = 0;
	(*f->p->handle)((void*)f->p, (void*)&disp, TextFrames_DisplayMsg__typ);
	if (disp.elemFrame == NIL) {
		__HALT(99);
	}
	f->dsc = disp.elemFrame;
}

static void Panel_DrawPanelArea (Panel_Frame f, INTEGER mode)
{
	GUtils_ReplConst((void*)f, 12, f->X + f->px, f->py, (int)__DIV(f->p->W, (LONGINT)TextFrames_Unit), (int)__DIV(f->p->H, (LONGINT)TextFrames_Unit), mode);
}

static void Panel_SetPanel (Panel_Frame f)
{
	Texts_Reader r = {0};
	if (f->p == NIL) {
		Texts_OpenReader(&r, Texts_Reader__typ, f->txt, 0);
		Texts_ReadElem(&r, Texts_Reader__typ);
		if (r.elem != NIL) {
			f->p = __GUARDP(r.elem, PanelElems_PanelDesc, 3);
		}
	}
}

static void Panel_Forward (Panel_Frame f, Display_FrameMsg *msg, LONGINT *msg__typ)
{
	if (f->dsc != NIL && !f->hasSel) {
		(*f->dsc->handle)(f->dsc, &*msg, msg__typ);
	}
}

static void Panel_Draw (Panel_Frame f)
{
	Oberon_ControlMsg neutralize = {0};
	GUtils_SetDevice(2);
	GUtils_ReplConst((void*)f, 12, f->X, f->Y, f->W, f->H, 0);
	if (f->p != NIL && f->H > 0) {
		f->px = 0;
		f->py = (f->Y + f->H) - (int)__DIV(f->p->H, (LONGINT)TextFrames_Unit);
		Panel_DrawPanel(f);
	} else if (f->H == 0) {
		neutralize.id = 1;
		Panel_Forward(f, (void*)&neutralize, Oberon_ControlMsg__typ);
	}
}

static void Panel_Reduce (Panel_Frame f, INTEGER newY)
{
	f->H = (f->H + f->Y) - newY;
	f->Y = newY;
}

static void Panel_Extend (Panel_Frame f, INTEGER newY)
{
	INTEGER dY = 0, newH = 0;
	dY = f->Y - newY;
	f->H = (f->H + f->Y) - newY;
	f->Y = newY;
	newH = f->H;
	f->H = dY;
	f->H = newH;
}

static void Panel_Modify (Panel_Frame f, INTEGER id, INTEGER dY, INTEGER Y, INTEGER H)
{
	Oberon_RemoveMarks(f->X, f->Y, f->W, f->H);
	f->hasCar = 0;
	f->hasSel = 0;
	if (id == 0) {
		if (dY > 0) {
			f->Y = f->Y + dY;
		}
		Panel_Extend(f, Y);
	} else if (id == 1) {
		Panel_Reduce(f, Y + dY);
		if (dY > 0) {
			f->Y = Y;
		}
	}
}

static BOOLEAN Panel_InPanel (Panel_Frame f, INTEGER x, INTEGER y)
{
	INTEGER w = 0, h = 0;
	w = (int)__DIV(f->p->W, (LONGINT)TextFrames_Unit);
	h = (int)__DIV(f->p->H, (LONGINT)TextFrames_Unit);
	return ((x >= f->X + f->px && x <= (f->X + f->px) + w) && y >= f->py) && y <= f->py + h;
}

static void Panel_CopyInto (Panel_Frame f, Texts_Text txt, LONGINT beg)
{
	Texts_Reader r = {0};
	Texts_CopyMsg copy = {0};
	Texts_OpenReader(&r, Texts_Reader__typ, txt, beg);
	Texts_ReadElem(&r, Texts_Reader__typ);
	if (r.elem != NIL && __ISP(r.elem, PanelElems_PanelDesc, 3)) {
		copy.e = NIL;
		(*r.elem->handle)(r.elem, (void*)&copy, Texts_CopyMsg__typ);
		Texts_SetFont(&Panel_W, Texts_Writer__typ, r.fnt);
		Texts_SetColor(&Panel_W, Texts_Writer__typ, r.col);
		Texts_WriteElem(&Panel_W, Texts_Writer__typ, copy.e);
		Texts_Append(f->txt, Panel_W.buf);
	}
}

static void Panel_HandleInput (Panel_Frame f, Oberon_InputMsg *msg, LONGINT *msg__typ)
{
	SET keysum = 0;
	Oberon_CopyOverMsg copyOver = {0};
	Texts_Text txt = NIL;
	LONGINT beg = 0, end = 0, time = 0;
	keysum = (*msg).keys;
	if ((*msg).id == 1) {
		if ((*msg).keys == 0x04) {
			if (f->p == NIL) {
				Panel_SetCaret(f);
				do {
					Input_Mouse(&(*msg).keys, &(*msg).X, &(*msg).Y);
					keysum = keysum | (*msg).keys;
					Oberon_DrawCursor(&Oberon_Mouse, Oberon_Cursor__typ, &Oberon_Arrow, Oberon_Marker__typ, (*msg).X, (*msg).Y);
				} while (!((*msg).keys == 0x0));
				if (keysum == 0x06) {
					Oberon_GetSelection(&txt, &beg, &end, &time);
					if (txt != NIL) {
						Panel_CopyInto(f, txt, beg);
					}
				}
			} else if (Panel_InPanel(f, (*msg).X, (*msg).Y)) {
				Panel_Forward(f, (void*)&*msg, msg__typ);
			}
		} else if ((*msg).keys == 0x02) {
			if (f->p != NIL && Panel_InPanel(f, (*msg).X, (*msg).Y)) {
				Panel_Forward(f, (void*)&*msg, msg__typ);
			}
		} else if ((*msg).keys == 0x01) {
			if (f->p != NIL && !Panel_InPanel(f, (*msg).X, (*msg).Y)) {
				f->selTime = Oberon_Time();
				f->hasSel = !f->hasSel;
				Panel_DrawPanelArea(f, 2);
				do {
					Input_Mouse(&(*msg).keys, &(*msg).X, &(*msg).Y);
					keysum = keysum | (*msg).keys;
					Oberon_DrawCursor(&Oberon_Mouse, Oberon_Cursor__typ, &Oberon_Arrow, Oberon_Marker__typ, (*msg).X, (*msg).Y);
				} while (!((*msg).keys == 0x0));
				if (keysum == 0x05 && f->hasSel) {
					Texts_Delete(f->txt, 0, f->txt->len);
				} else if (keysum == 0x03 && f->hasSel) {
					copyOver.text = f->txt;
					copyOver.beg = 0;
					copyOver.end = f->txt->len;
					Viewers_Broadcast((void*)&copyOver, Oberon_CopyOverMsg__typ);
				}
			} else {
				Panel_Forward(f, (void*)&*msg, msg__typ);
			}
		} else {
			Panel_Forward(f, (void*)&*msg, msg__typ);
		}
	} else {
		Panel_Forward(f, (void*)&*msg, msg__typ);
	}
}

static void Panel_Handle (Display_Frame f, Display_FrameMsg *msg, LONGINT *msg__typ)
{
	Panel_Frame F = NIL;
	if (__ISP(f, Panel_FrameDesc, 1)) {
		if (__IS(msg__typ, Oberon_InputMsg, 1)) {
			Oberon_InputMsg *msg__ = (void*)msg;
			Panel_HandleInput((*(Panel_Frame*)&f), &*msg__, msg__typ);
		} else if (__IS(msg__typ, Oberon_ControlMsg, 1)) {
			Oberon_ControlMsg *msg__ = (void*)msg;
			if ((*(Panel_Frame*)&f)->hasCar && __IN((*msg__).id, 0x03)) {
				Panel_RemoveCaret((*(Panel_Frame*)&f));
			}
			if ((*(Panel_Frame*)&f)->hasSel && (*msg__).id == 1) {
				Panel_DrawPanelArea((*(Panel_Frame*)&f), 2);
				(*(Panel_Frame*)&f)->hasSel = 0;
			}
			Panel_Forward((*(Panel_Frame*)&f), (void*)&*msg__, msg__typ);
		} else if (__IS(msg__typ, Oberon_CopyOverMsg, 1)) {
			Oberon_CopyOverMsg *msg__ = (void*)msg;
			if ((*(Panel_Frame*)&f)->hasCar) {
				Panel_CopyInto((*(Panel_Frame*)&f), (*msg__).text, (*msg__).beg);
			} else {
				Panel_Forward((*(Panel_Frame*)&f), (void*)&*msg__, msg__typ);
			}
		} else if (__IS(msg__typ, Oberon_SelectionMsg, 1)) {
			Oberon_SelectionMsg *msg__ = (void*)msg;
			Panel_Forward((*(Panel_Frame*)&f), (void*)&*msg__, msg__typ);
			if ((*(Panel_Frame*)&f)->hasSel && (*(Panel_Frame*)&f)->selTime > (*msg__).time) {
				(*msg__).text = (*(Panel_Frame*)&f)->txt;
				(*msg__).beg = 0;
				(*msg__).end = (*(Panel_Frame*)&f)->txt->len;
				(*msg__).time = (*(Panel_Frame*)&f)->selTime;
			}
		} else if (__IS(msg__typ, MenuViewers_ModifyMsg, 1)) {
			MenuViewers_ModifyMsg *msg__ = (void*)msg;
			Panel_Modify((*(Panel_Frame*)&f), (*msg__).id, (*msg__).dY, (*msg__).Y, (*msg__).H);
			Panel_Draw((*(Panel_Frame*)&f));
		} else if (__IS(msg__typ, Oberon_CopyMsg, 1)) {
			Oberon_CopyMsg *msg__ = (void*)msg;
			__NEW(F, Panel_FrameDesc);
			(*msg__).F = (Display_Frame)F;
			__GUARDEQP(F, Panel_FrameDesc) = *(*(Panel_Frame*)&f);
		} else if (__IS(msg__typ, TextFrames_UpdateMsg, 1)) {
			TextFrames_UpdateMsg *msg__ = (void*)msg;
			if ((*msg__).text == (*(Panel_Frame*)&f)->txt) {
				if ((*msg__).id == 2) {
					(*(Panel_Frame*)&f)->p = NIL;
					(*(Panel_Frame*)&f)->dsc = NIL;
				}
				(*(Panel_Frame*)&f)->hasSel = 0;
				(*(Panel_Frame*)&f)->hasCar = 0;
				Panel_SetPanel((*(Panel_Frame*)&f));
				Panel_Draw((*(Panel_Frame*)&f));
			} else {
				Panel_Forward((*(Panel_Frame*)&f), (void*)&*msg__, msg__typ);
			}
		} else if (__IS(msg__typ, TextFrames_InsertElemMsg, 1)) {
			TextFrames_InsertElemMsg *msg__ = (void*)msg;
			if ((*(Panel_Frame*)&f)->hasCar && __ISP((*msg__).e, PanelElems_PanelDesc, 3)) {
				Texts_WriteElem(&Panel_W, Texts_Writer__typ, (*msg__).e);
				Texts_Append((*(Panel_Frame*)&f)->txt, Panel_W.buf);
			} else {
				Panel_Forward((*(Panel_Frame*)&f), (void*)&*msg__, msg__typ);
			}
		} else {
			Panel_Forward((*(Panel_Frame*)&f), &*msg, msg__typ);
		}
	} else __WITHCHK;
}

void Panel_OpenPanel (PanelElems_Panel p, CHAR *name, LONGINT name__len, Fonts_Font fnt, SHORTINT col)
{
	Panel_Frame f = NIL;
	Viewers_Viewer V = NIL, v = NIL, vmax = NIL;
	INTEGER X = 0, Y = 0, h = 0;
	CHAR mnu[64];
	__DUP(name, name__len, CHAR);
	__NEW(f, Panel_FrameDesc);
	f->handle = Panel_Handle;
	f->hasCar = 0;
	f->hasSel = 0;
	f->txt = TextFrames_Text((CHAR*)"", (LONGINT)1);
	if (p == NIL) {
		__NEW(p, PanelElems_PanelDesc);
		PanelElems_Init(p);
		fnt = Fonts_Default;
		col = 13;
	}
	Texts_SetFont(&Panel_W, Texts_Writer__typ, fnt);
	Texts_SetColor(&Panel_W, Texts_Writer__typ, col);
	Texts_WriteElem(&Panel_W, Texts_Writer__typ, (void*)p);
	Texts_Append(f->txt, Panel_W.buf);
	Panel_SetPanel(f);
	Oberon_AllocateSystemViewer(Oberon_Mouse.X, &X, &Y);
	v = Viewers_This(X, 0);
	vmax = NIL;
	h = 0;
	while (v->state > 1) {
		if (v->H > h) {
			vmax = v;
			h = v->H;
		}
		v = Viewers_Next(v);
	}
	if (vmax != NIL) {
		Y = Panel_Min(((vmax->Y + (int)__DIV(f->p->H, Display_Unit)) + TextFrames_menuH) + 3, ((vmax->Y + vmax->H) - TextFrames_menuH) - 2);
	}
	Panel_GetMenu(f->p, (void*)mnu, 64);
	V = (Viewers_Viewer)MenuViewers_New((void*)TextFrames_NewMenu(name, name__len, mnu, 64), (void*)f, TextFrames_menuH, X, Y);
	__DEL(name);
}

void Panel_LoadPanel (CHAR *name, LONGINT name__len, PanelElems_Panel *p, Fonts_Font *fnt, SHORTINT *col)
{
	Texts_Reader r = {0};
	Texts_Text txt = NIL;
	__DUP(name, name__len, CHAR);
	*p = NIL;
	if (name[0] != 0x00 && Files_Old(name, name__len) != NIL) {
		txt = TextFrames_Text(name, name__len);
		Texts_OpenReader(&r, Texts_Reader__typ, txt, 0);
		Texts_ReadElem(&r, Texts_Reader__typ);
		if (r.elem != NIL && __ISP(r.elem, PanelElems_PanelDesc, 3)) {
			*p = __GUARDP(r.elem, PanelElems_PanelDesc, 3);
			*fnt = r.fnt;
			*col = r.col;
			Texts_Delete(txt, 0, txt->len);
		}
	}
	__DEL(name);
}

void Panel_Open (void)
{
	Texts_Scanner s = {0};
	Texts_Text t = NIL;
	LONGINT beg = 0, end = 0, time = 0;
	SHORTINT col = 0;
	Fonts_Font fnt = NIL;
	PanelElems_Panel p = NIL;
	Texts_OpenScanner(&s, Texts_Scanner__typ, Oberon_Par->text, Oberon_Par->pos);
	Texts_Scan(&s, Texts_Scanner__typ);
	if (s.class == 6 && s.c == '^') {
		Oberon_GetSelection(&t, &beg, &end, &time);
		if (t != NIL) {
			Texts_OpenScanner(&s, Texts_Scanner__typ, t, beg);
			Texts_Scan(&s, Texts_Scanner__typ);
		} else {
			s.class = 0;
		}
	}
	if (__IN(s.class, 0x06)) {
		Panel_LoadPanel(s.s, 256, &p, &fnt, &col);
		Panel_OpenPanel(p, s.s, 256, fnt, col);
	} else {
		Panel_OpenPanel(NIL, (CHAR*)"", (LONGINT)1, NIL, 0);
	}
}

void Panel_Store (void)
{
	Viewers_Viewer v = NIL;
	Texts_Scanner s = {0};
	CHAR name[64];
	Texts_Text t = NIL;
	Texts_Reader R = {0};
	CHAR ch = 0;
	v = Oberon_Par->vwr;
	name[0] = 0x00;
	if (((v->dsc != Oberon_Par->frame || v->dsc == NIL) || v->dsc->next == NIL) || !__ISP(v->dsc->next, Panel_FrameDesc, 1)) {
		v = Oberon_MarkedViewer();
		Texts_OpenScanner(&s, Texts_Scanner__typ, Oberon_Par->text, Oberon_Par->pos);
		Texts_Scan(&s, Texts_Scanner__typ);
		if (s.line == 0 && __IN(s.class, 0x06)) {
			__COPY(s.s, name, 64);
		}
	}
	if ((v->dsc != NIL && v->dsc->next != NIL) && __ISP(v->dsc->next, Panel_FrameDesc, 1)) {
		if (name[0] == 0x00 && __ISP(v->dsc, TextFrames_FrameDesc, 1)) {
			Texts_OpenScanner(&s, Texts_Scanner__typ, __GUARDP(v->dsc, TextFrames_FrameDesc, 1)->text, 0);
			Texts_Scan(&s, Texts_Scanner__typ);
			if (s.line == 0 && __IN(s.class, 0x06)) {
				__COPY(s.s, name, 64);
			}
		}
		if (name[0] != 0x00) {
			Texts_SetColor(&Panel_W, Texts_Writer__typ, 15);
			Texts_SetFont(&Panel_W, Texts_Writer__typ, Fonts_Default);
			Texts_WriteString(&Panel_W, Texts_Writer__typ, (CHAR*)"Panel.Store ", (LONGINT)13);
			Texts_WriteString(&Panel_W, Texts_Writer__typ, name, 64);
			Texts_Close(__GUARDP(v->dsc->next, Panel_FrameDesc, 1)->txt, name, 64);
			Texts_WriteString(&Panel_W, Texts_Writer__typ, (CHAR*)" done", (LONGINT)6);
			Texts_WriteLn(&Panel_W, Texts_Writer__typ);
			Texts_Append(Oberon_Log, Panel_W.buf);
			t = __GUARDP(v->dsc, TextFrames_FrameDesc, 1)->text;
			Texts_OpenReader(&R, Texts_Reader__typ, t, t->len - 1);
			Texts_Read(&R, Texts_Reader__typ, &ch);
			if (ch == '!') {
				Texts_Delete(t, t->len - 1, t->len);
			}
		}
	}
}

static void EnumPtrs(void (*P)(void*))
{
	__ENUMR(&Panel_W, Texts_Writer__typ, 36, 1, P);
}

__TDESC(Panel_FrameDesc, 1, 4) = {__TDFLDS("FrameDesc", 40), {0, 4, 20, 24, -20}};

export void *Panel__init(void)
{
	__DEFMOD;
	__IMPORT(Display);
	__IMPORT(Elems);
	__IMPORT(Files);
	__IMPORT(Fonts);
	__IMPORT(GUtils);
	__IMPORT(Input);
	__IMPORT(MenuViewers);
	__IMPORT(Oberon);
	__IMPORT(Out);
	__IMPORT(PanelElems);
	__IMPORT(Strings);
	__IMPORT(TextFrames);
	__IMPORT(Texts);
	__IMPORT(Viewers);
	__REGMOD(Panel, EnumPtrs);
	__REGCMD("Open", Panel_Open);
	__REGCMD("Store", Panel_Store);
	__INITYP(Panel_FrameDesc, Display_FrameDesc, 1);
/* BEGIN */
	Texts_OpenWriter(&Panel_W, Texts_Writer__typ);
	__ENDMOD;
}
