/* Ofront 1.4 -xtspkaG */
#include "SYSTEM.h"
#include "Display.h"
#include "Elems.h"
#include "Fonts.h"
#include "GUtils.h"
#include "Input.h"
#include "MenuViewers.h"
#include "Oberon.h"
#include "PanelElems.h"
#include "TextFrames.h"
#include "Texts.h"
#include "Viewers.h"

typedef
	struct PanelFrames_FrameDesc *PanelFrames_Frame;

typedef
	struct PanelFrames_FrameDesc { /* Display_FrameDesc */
		Display_Frame dsc, next;
		INTEGER X, Y, W, H;
		Display_Handler handle;
		PanelElems_Panel p;
		Texts_Text txt;
		INTEGER px, py;
		BOOLEAN hasCar, hasSel;
		LONGINT selTime;
	} PanelFrames_FrameDesc;


static Texts_Writer PanelFrames_W;

export LONGINT *PanelFrames_FrameDesc__typ;

static void PanelFrames_CopyInto (PanelFrames_Frame f, Texts_Text txt, LONGINT beg);
static void PanelFrames_Draw (PanelFrames_Frame f);
static void PanelFrames_DrawPanel (PanelFrames_Frame f);
static void PanelFrames_DrawPanelArea (PanelFrames_Frame f, INTEGER mode);
static void PanelFrames_Extend (PanelFrames_Frame f, INTEGER newY);
static void PanelFrames_Forward (PanelFrames_Frame f, Display_FrameMsg *msg, LONGINT *msg__typ);
export void PanelFrames_Handle (Display_Frame f, Display_FrameMsg *msg, LONGINT *msg__typ);
static void PanelFrames_HandleInput (PanelFrames_Frame f, Oberon_InputMsg *msg, LONGINT *msg__typ);
static BOOLEAN PanelFrames_InPanel (PanelFrames_Frame f, INTEGER x, INTEGER y);
static void PanelFrames_InvertCaret (PanelFrames_Frame f);
static void PanelFrames_Modify (PanelFrames_Frame f, INTEGER id, INTEGER dY, INTEGER Y, INTEGER H);
export void PanelFrames_Open (PanelFrames_Frame f, PanelElems_Panel p, Fonts_Font fnt, SHORTINT col, Display_Handler handle);
static void PanelFrames_Reduce (PanelFrames_Frame f, INTEGER newY);
static void PanelFrames_RemoveCaret (PanelFrames_Frame f);
static void PanelFrames_SetCaret (PanelFrames_Frame f);
static void PanelFrames_SetPanel (PanelFrames_Frame f);


static void PanelFrames_InvertCaret (PanelFrames_Frame f)
{
	GUtils_Frame((void*)f, 0, f->X, f->Y, f->W, f->H, 3, 2);
}

static void PanelFrames_SetCaret (PanelFrames_Frame f)
{
	Oberon_PassFocus(Viewers_This(f->X, f->Y));
	f->hasCar = 1;
	PanelFrames_InvertCaret(f);
}

static void PanelFrames_RemoveCaret (PanelFrames_Frame f)
{
	if (f->hasCar) {
		PanelFrames_InvertCaret(f);
		f->hasCar = 0;
	}
}

static void PanelFrames_DrawPanel (PanelFrames_Frame f)
{
	TextFrames_DisplayMsg disp = {0};
	Elems_GetLook((void*)f->p, &disp.fnt, &disp.col);
	disp.pos = 0;
	disp.frame = (Display_Frame)f;
	disp.X0 = f->X + f->px;
	disp.Y0 = f->py;
	disp.indent = 0;
	disp.elemFrame = NIL;
	disp.prepare = 1;
	(*f->p->handle)((void*)f->p, (void*)&disp, TextFrames_DisplayMsg__typ);
	disp.prepare = 0;
	(*f->p->handle)((void*)f->p, (void*)&disp, TextFrames_DisplayMsg__typ);
	if (disp.elemFrame == NIL) {
		__HALT(99);
	}
	f->dsc = disp.elemFrame;
}

static void PanelFrames_DrawPanelArea (PanelFrames_Frame f, INTEGER mode)
{
	GUtils_ReplConst((void*)f, 12, f->X + f->px, f->py, (int)__DIV(f->p->W, (LONGINT)TextFrames_Unit), (int)__DIV(f->p->H, (LONGINT)TextFrames_Unit), mode);
}

static void PanelFrames_SetPanel (PanelFrames_Frame f)
{
	Texts_Reader r = {0};
	if (f->p == NIL) {
		Texts_OpenReader(&r, Texts_Reader__typ, f->txt, 0);
		Texts_ReadElem(&r, Texts_Reader__typ);
		if (r.elem != NIL) {
			f->p = __GUARDP(r.elem, PanelElems_PanelDesc, 3);
		}
	}
}

static void PanelFrames_Forward (PanelFrames_Frame f, Display_FrameMsg *msg, LONGINT *msg__typ)
{
	if (f->dsc != NIL && !f->hasSel) {
		(*f->dsc->handle)(f->dsc, &*msg, msg__typ);
	}
}

static void PanelFrames_Draw (PanelFrames_Frame f)
{
	Oberon_ControlMsg neutralize = {0};
	GUtils_SetDevice(2);
	GUtils_ReplConst((void*)f, 12, f->X, f->Y, f->W, f->H, 0);
	if (f->p != NIL && f->H > 0) {
		f->px = 0;
		f->py = (f->Y + f->H) - (int)__DIV(f->p->H, (LONGINT)TextFrames_Unit);
		PanelFrames_DrawPanel(f);
	} else if (f->H == 0) {
		neutralize.id = 1;
		PanelFrames_Forward(f, (void*)&neutralize, Oberon_ControlMsg__typ);
	}
}

static void PanelFrames_Reduce (PanelFrames_Frame f, INTEGER newY)
{
	f->H = (f->H + f->Y) - newY;
	f->Y = newY;
}

static void PanelFrames_Extend (PanelFrames_Frame f, INTEGER newY)
{
	INTEGER dY = 0, newH = 0;
	dY = f->Y - newY;
	f->H = (f->H + f->Y) - newY;
	f->Y = newY;
	newH = f->H;
	f->H = dY;
	f->H = newH;
}

static void PanelFrames_Modify (PanelFrames_Frame f, INTEGER id, INTEGER dY, INTEGER Y, INTEGER H)
{
	Oberon_RemoveMarks(f->X, f->Y, f->W, f->H);
	f->hasCar = 0;
	f->hasSel = 0;
	if (id == 0) {
		if (dY > 0) {
			f->Y += dY;
		}
		PanelFrames_Extend(f, Y);
	} else if (id == 1) {
		PanelFrames_Reduce(f, Y + dY);
		if (dY > 0) {
			f->Y = Y;
		}
	}
}

static BOOLEAN PanelFrames_InPanel (PanelFrames_Frame f, INTEGER x, INTEGER y)
{
	INTEGER w = 0, h = 0;
	w = (int)__DIV(f->p->W, (LONGINT)TextFrames_Unit);
	h = (int)__DIV(f->p->H, (LONGINT)TextFrames_Unit);
	return ((x >= f->X + f->px && x <= (f->X + f->px) + w) && y >= f->py) && y <= f->py + h;
}

static void PanelFrames_CopyInto (PanelFrames_Frame f, Texts_Text txt, LONGINT beg)
{
	Texts_Reader r = {0};
	Texts_CopyMsg copy = {0};
	Texts_OpenReader(&r, Texts_Reader__typ, txt, beg);
	Texts_ReadElem(&r, Texts_Reader__typ);
	if (r.elem != NIL && __ISP(r.elem, PanelElems_PanelDesc, 3)) {
		copy.e = NIL;
		(*r.elem->handle)(r.elem, (void*)&copy, Texts_CopyMsg__typ);
		Texts_SetFont(&PanelFrames_W, Texts_Writer__typ, r.fnt);
		Texts_SetColor(&PanelFrames_W, Texts_Writer__typ, r.col);
		Texts_WriteElem(&PanelFrames_W, Texts_Writer__typ, copy.e);
		Texts_Append(f->txt, PanelFrames_W.buf);
	}
}

static void PanelFrames_HandleInput (PanelFrames_Frame f, Oberon_InputMsg *msg, LONGINT *msg__typ)
{
	SET keysum = 0;
	Oberon_CopyOverMsg copyOver = {0};
	Texts_Text txt = NIL;
	LONGINT beg = 0, end = 0, time = 0;
	keysum = (*msg).keys;
	if ((*msg).id == 1) {
		if ((*msg).keys == 0x04) {
			if (f->p == NIL) {
				PanelFrames_SetCaret(f);
				do {
					Input_Mouse(&(*msg).keys, &(*msg).X, &(*msg).Y);
					keysum = keysum | (*msg).keys;
					Oberon_DrawCursor(&Oberon_Mouse, Oberon_Cursor__typ, &Oberon_Arrow, Oberon_Marker__typ, (*msg).X, (*msg).Y);
				} while (!((*msg).keys == 0x0));
				if (keysum == 0x06) {
					Oberon_GetSelection(&txt, &beg, &end, &time);
					if (txt != NIL) {
						PanelFrames_CopyInto(f, txt, beg);
					}
				}
			} else if (PanelFrames_InPanel(f, (*msg).X, (*msg).Y)) {
				PanelFrames_Forward(f, (void*)&*msg, msg__typ);
			}
		} else if ((*msg).keys == 0x02) {
			if (f->p != NIL && PanelFrames_InPanel(f, (*msg).X, (*msg).Y)) {
				PanelFrames_Forward(f, (void*)&*msg, msg__typ);
			}
		} else if ((*msg).keys == 0x01) {
			if (f->p != NIL && !PanelFrames_InPanel(f, (*msg).X, (*msg).Y)) {
				f->selTime = Oberon_Time();
				f->hasSel = !f->hasSel;
				PanelFrames_DrawPanelArea(f, 2);
				do {
					Input_Mouse(&(*msg).keys, &(*msg).X, &(*msg).Y);
					keysum = keysum | (*msg).keys;
					Oberon_DrawCursor(&Oberon_Mouse, Oberon_Cursor__typ, &Oberon_Arrow, Oberon_Marker__typ, (*msg).X, (*msg).Y);
				} while (!((*msg).keys == 0x0));
				if (keysum == 0x05 && f->hasSel) {
					Texts_Delete(f->txt, 0, f->txt->len);
				} else if (keysum == 0x03 && f->hasSel) {
					copyOver.text = f->txt;
					copyOver.beg = 0;
					copyOver.end = f->txt->len;
					Viewers_Broadcast((void*)&copyOver, Oberon_CopyOverMsg__typ);
				}
			} else {
				PanelFrames_Forward(f, (void*)&*msg, msg__typ);
			}
		}
	} else {
		PanelFrames_Forward(f, (void*)&*msg, msg__typ);
	}
}

void PanelFrames_Handle (Display_Frame f, Display_FrameMsg *msg, LONGINT *msg__typ)
{
	PanelFrames_Frame F = NIL;
	if (__ISP(f, PanelFrames_FrameDesc, 1)) {
		if (__IS(msg__typ, Oberon_InputMsg, 1)) {
			Oberon_InputMsg *msg__ = (void*)msg;
			PanelFrames_HandleInput((*(PanelFrames_Frame*)&f), &*msg__, msg__typ);
		} else if (__IS(msg__typ, Oberon_ControlMsg, 1)) {
			Oberon_ControlMsg *msg__ = (void*)msg;
			if ((*(PanelFrames_Frame*)&f)->hasCar && __IN((*msg__).id, 0x03)) {
				PanelFrames_RemoveCaret((*(PanelFrames_Frame*)&f));
			}
			if ((*(PanelFrames_Frame*)&f)->hasSel && (*msg__).id == 1) {
				PanelFrames_DrawPanelArea((*(PanelFrames_Frame*)&f), 2);
				(*(PanelFrames_Frame*)&f)->hasSel = 0;
			}
			PanelFrames_Forward((*(PanelFrames_Frame*)&f), (void*)&*msg__, msg__typ);
		} else if (__IS(msg__typ, Oberon_CopyOverMsg, 1)) {
			Oberon_CopyOverMsg *msg__ = (void*)msg;
			if ((*(PanelFrames_Frame*)&f)->hasCar) {
				PanelFrames_CopyInto((*(PanelFrames_Frame*)&f), (*msg__).text, (*msg__).beg);
			} else {
				PanelFrames_Forward((*(PanelFrames_Frame*)&f), (void*)&*msg__, msg__typ);
			}
		} else if (__IS(msg__typ, Oberon_SelectionMsg, 1)) {
			Oberon_SelectionMsg *msg__ = (void*)msg;
			PanelFrames_Forward((*(PanelFrames_Frame*)&f), (void*)&*msg__, msg__typ);
			if ((*(PanelFrames_Frame*)&f)->hasSel && (*(PanelFrames_Frame*)&f)->selTime > (*msg__).time) {
				(*msg__).text = (*(PanelFrames_Frame*)&f)->txt;
				(*msg__).beg = 0;
				(*msg__).end = (*(PanelFrames_Frame*)&f)->txt->len;
				(*msg__).time = (*(PanelFrames_Frame*)&f)->selTime;
			}
		} else if (__IS(msg__typ, MenuViewers_ModifyMsg, 1)) {
			MenuViewers_ModifyMsg *msg__ = (void*)msg;
			PanelFrames_Modify((*(PanelFrames_Frame*)&f), (*msg__).id, (*msg__).dY, (*msg__).Y, (*msg__).H);
			PanelFrames_Draw((*(PanelFrames_Frame*)&f));
		} else if (__IS(msg__typ, Oberon_CopyMsg, 1)) {
			Oberon_CopyMsg *msg__ = (void*)msg;
			__NEW(F, PanelFrames_FrameDesc);
			(*msg__).F = (Display_Frame)F;
			__GUARDEQP(F, PanelFrames_FrameDesc) = *(*(PanelFrames_Frame*)&f);
		} else if (__IS(msg__typ, TextFrames_UpdateMsg, 1)) {
			TextFrames_UpdateMsg *msg__ = (void*)msg;
			if ((*msg__).text == (*(PanelFrames_Frame*)&f)->txt) {
				if ((*msg__).id == 2) {
					(*(PanelFrames_Frame*)&f)->p = NIL;
					(*(PanelFrames_Frame*)&f)->dsc = NIL;
				}
				(*(PanelFrames_Frame*)&f)->hasSel = 0;
				(*(PanelFrames_Frame*)&f)->hasCar = 0;
				PanelFrames_SetPanel((*(PanelFrames_Frame*)&f));
				PanelFrames_Draw((*(PanelFrames_Frame*)&f));
			} else {
				PanelFrames_Forward((*(PanelFrames_Frame*)&f), (void*)&*msg__, msg__typ);
			}
		} else if (__IS(msg__typ, TextFrames_InsertElemMsg, 1)) {
			TextFrames_InsertElemMsg *msg__ = (void*)msg;
			if ((*(PanelFrames_Frame*)&f)->hasCar && __ISP((*msg__).e, PanelElems_PanelDesc, 3)) {
				Texts_WriteElem(&PanelFrames_W, Texts_Writer__typ, (*msg__).e);
				Texts_Append((*(PanelFrames_Frame*)&f)->txt, PanelFrames_W.buf);
			} else {
				PanelFrames_Forward((*(PanelFrames_Frame*)&f), (void*)&*msg__, msg__typ);
			}
		} else {
			PanelFrames_Forward((*(PanelFrames_Frame*)&f), &*msg, msg__typ);
		}
	} else __WITHCHK;
}

void PanelFrames_Open (PanelFrames_Frame f, PanelElems_Panel p, Fonts_Font fnt, SHORTINT col, Display_Handler handle)
{
	f->txt = TextFrames_Text((CHAR*)"", (LONGINT)1);
	f->hasCar = 0;
	f->hasSel = 0;
	f->p = p;
	Texts_SetFont(&PanelFrames_W, Texts_Writer__typ, fnt);
	Texts_SetColor(&PanelFrames_W, Texts_Writer__typ, col);
	Texts_WriteElem(&PanelFrames_W, Texts_Writer__typ, (void*)p);
	Texts_Append(f->txt, PanelFrames_W.buf);
	f->handle = handle;
}

static void EnumPtrs(void (*P)(void*))
{
	__ENUMR(&PanelFrames_W, Texts_Writer__typ, 36, 1, P);
}

__TDESC(PanelFrames_FrameDesc, 1, 4) = {__TDFLDS("FrameDesc", 40), {0, 4, 20, 24, -20}};

export void *PanelFrames__init(void)
{
	__DEFMOD;
	__IMPORT(Display);
	__IMPORT(Elems);
	__IMPORT(Fonts);
	__IMPORT(GUtils);
	__IMPORT(Input);
	__IMPORT(MenuViewers);
	__IMPORT(Oberon);
	__IMPORT(PanelElems);
	__IMPORT(TextFrames);
	__IMPORT(Texts);
	__IMPORT(Viewers);
	__REGMOD(PanelFrames, EnumPtrs);
	__INITYP(PanelFrames_FrameDesc, Display_FrameDesc, 1);
/* BEGIN */
	Texts_OpenWriter(&PanelFrames_W, Texts_Writer__typ);
	__ENDMOD;
}
