/* 
*	The body prefix file of the Ofront runtime system, Version 1.0
*
*	Copyright (c) Software Templ, 1994-2016
*
*	Module SYSTEM is subject to change any time without prior notification.
*	Software Templ disclaims all warranties with regard to module SYSTEM,
*	in particular shall Software Templ not be liable for any damage resulting
*	from inappropriate use or modification of module SYSTEM.
*
*	Version 1.1  jt, 24.11.95  fixes for correct pointer arithmetic on Cray computers
*	jt 31.1.2007  ANSI prototypes for malloc and exit in order to avoid cc warnings
*	jt 22.09.2016  module loading/unloading integrated with shared objects
*
*/

#include "SYSTEM.h"
#ifdef __STDC__
#include "stdarg.h"
#else
#include "varargs.h"
#endif

extern void *malloc(long size);
extern void exit(int status);

void (*SYSTEM_Halt)();
LONGINT SYSTEM_halt;	/* x in HALT(x) */
LONGINT SYSTEM_assert;	/* x in ASSERT(cond, x) */
LONGINT SYSTEM_argc;
LONGINT SYSTEM_argv;
LONGINT SYSTEM_lock;
BOOLEAN SYSTEM_interrupted;
static LONGINT SYSTEM_mainfrm;	/* adr of main proc stack frame, used for stack collection */

#define Lock	SYSTEM_lock++
#define Unlock	SYSTEM_lock--; if (SYSTEM_interrupted && (SYSTEM_lock == 0)) __HALT(-9)


static void SYSTEM_InitHeap();
void *SYSTEM__init();

void SYSTEM_INIT(argc, argvadr)
	int argc; long argvadr;
{
	SYSTEM_mainfrm = argvadr;
	SYSTEM_argc = argc;
	SYSTEM_argv = *(long*)argvadr;
	SYSTEM_InitHeap();
	SYSTEM_halt = -128;
	SYSTEM__init();
}

void SYSTEM_FINI()
{
	SYSTEM_FINALL();
}

long SYSTEM_XCHK(i, ub) long i, ub; {return __X(i, ub);}
long SYSTEM_RCHK(i, ub) long i, ub; {return __R(i, ub);}
long SYSTEM_ASH(i, n) long i, n; {return __ASH(i, n);}
long SYSTEM_ABS(i) long i; {return __ABS(i);}
double SYSTEM_ABSD(i) double i; {return __ABS(i);}

void SYSTEM_INHERIT(t, t0)
	long *t, *t0;
{
	t -= __TPROC0OFF;
	t0 -= __TPROC0OFF;
	while (*t0 != __EOM) {*t = *t0; t--; t0--;}
}

void SYSTEM_ENUMP(adr, n, P)
	long *adr;
	long n;
	void (*P)();
{
	while (n > 0) {P(*adr); adr++; n--;}
}

void SYSTEM_ENUMR(adr, typ, size, n, P)
	char *adr;
	long *typ, size, n;
	void (*P)();
{
	long *t, off;
	if (typ == 0) return; /* not yet initialized */
	typ++;
	while (n > 0) {
		t = typ;
		off = *t;
		while (off >= 0) {P(*(long*)(adr+off)); t++; off = *t;}
		adr += size; n--;
	}
}

long SYSTEM_DIV(x, y)
	unsigned long x, y;
{  if ((long) x >= 0) return (x / y);
	else return -((y - 1 - x) / y);
}

long SYSTEM_MOD(x, y)
	unsigned long x, y;
{ unsigned long m;
	if ((long) x >= 0) return (x % y);
	else { m = (-x) % y;
		if (m != 0) return (y - m); else return 0;
	}
}

long SYSTEM_ENTIER(x)
	double x;
{
	long y;
	if (x >= 0)
		return (long)x;
	else {
		y = (long)x;
		if (y <= x) return y; else return y - 1;
	}
}

void SYSTEM_HALT(n)
	int n;
{
	SYSTEM_halt = n;
	if (SYSTEM_Halt!=0) SYSTEM_Halt(n);
	exit(n);
}

#ifdef __STDC__
SYSTEM_PTR SYSTEM_NEWARR(long *typ, long elemsz, int elemalgn, int nofdim, int nofdyn, ...)
#else
SYSTEM_PTR SYSTEM_NEWARR(typ, elemsz, elemalgn, nofdim, nofdyn, va_alist)
	long *typ, elemsz;
	int elemalgn, nofdim, nofdyn;
	va_dcl
#endif
{
	long nofelems, size, dataoff, n, nptr, *x, *p, nofptrs, i, *ptab, off;
	va_list ap;
#ifdef __STDC__
	va_start(ap, nofdyn);
#else
	va_start(ap);
#endif
	nofelems = 1;
	while (nofdim > 0) {
		nofelems = nofelems * va_arg(ap, long); nofdim--;
		if (nofelems <= 0) __HALT(-20);
	}
	va_end(ap);
	dataoff = nofdyn * sizeof(long);
	if (elemalgn > sizeof(long)) {
		n = dataoff % elemalgn;
		if (n != 0) dataoff += elemalgn - n;
	}
	size = dataoff + nofelems * elemsz;
	Lock;
	if (typ == NIL) {
		/* element typ does not contain pointers */
		x = SYSTEM_NEWBLK(size);
	}
	else if (typ == POINTER__typ) {
		/* element type is a pointer */
		x = SYSTEM_NEWBLK(size + nofelems * sizeof(long));
		p = (long*)x[-1];
		p[-nofelems] = *p;	/* build new type desc in situ: 1. copy block size; 2. setup ptr tab; 3. set sentinel; 4. patch tag */
		p -= nofelems - 1; n = 1;	/* n =1 for skipping the size field */
		while (n <= nofelems) {*p = n*sizeof(long); p++; n++;}
		*p = - (nofelems + 1) * sizeof(long);	/* sentinel */
		x[-1] -= nofelems * sizeof(long);
	}
	else {
		/* element type is a record that contains pointers */
		ptab = typ + 1; nofptrs = 0;
		while (ptab[nofptrs] >= 0) {nofptrs++;}	/* number of pointers per element */
		nptr = nofelems * nofptrs;	/* total number of pointers */
		x = SYSTEM_NEWBLK(size + nptr * sizeof(long));
		p = (long*)x[- 1];
		p[-nptr] = *p;	/* build new type desc in situ; 1. copy block size; 2. setup ptr tab; 3. set sentinel; 4. patch tag */
		p -= nptr - 1; n = 0; off = dataoff;
		while (n < nofelems) {i = 0;
			while (i < nofptrs) {*p = off + ptab[i]; p++; i++;}
			off += elemsz; n++;
		}
		*p = - (nptr + 1) * sizeof(long);	/* sentinel */
		x[-1] -= nptr * sizeof(long);
	}
	if (nofdyn != 0) {
		/* setup len vector for index checks */
#ifdef __STDC__
		va_start(ap, nofdyn);
#else
		va_start(ap);
#endif
		p = x;
		while (nofdyn > 0) {*p = va_arg(ap, long); p++, nofdyn--;}
		va_end(ap);
	}
	Unlock;
	return x;
}

/* registers must be stored on the stack before doing garbage collection on the stack.
Pretty portable version: uses jump buffer for saving processor state incl. registers on the stack.
*/

#include <setjmp.h>
extern void SYSTEM_MarkStack (void);

static void SYSTEM_callMarkStack(void) {
	jmp_buf env;
	setjmp(env);
	SYSTEM_MarkStack();
}

/* ----------- end of SYSTEM.co ------------- */

/* Ofront 1.4 -xtspkaG */
#include "SYSTEM.h"

struct SYSTEM__1 {
	CHAR ch;
	SYSTEM_PTR p;
};

typedef
	struct SYSTEM_CmdDesc *SYSTEM_Cmd;

typedef
	CHAR SYSTEM_CmdName[24];

typedef
	void (*SYSTEM_Command)(void);

typedef
	struct SYSTEM_CmdDesc {
		SYSTEM_Cmd next;
		SYSTEM_CmdName name;
		SYSTEM_Command cmd;
	} SYSTEM_CmdDesc;

typedef
	void (*SYSTEM_EnumProc)(void(*)(SYSTEM_PTR));

typedef
	struct SYSTEM_FinDesc *SYSTEM_FinNode;

typedef
	void (*SYSTEM_Finalizer)(SYSTEM_PTR);

typedef
	struct SYSTEM_FinDesc {
		SYSTEM_FinNode next;
		LONGINT obj;
		BOOLEAN marked;
		SYSTEM_Finalizer finalize;
	} SYSTEM_FinDesc;

typedef
	struct SYSTEM_ModuleDesc *SYSTEM_Module;

typedef
	CHAR SYSTEM_ModuleName[20];

typedef
	struct SYSTEM_ModuleDesc {
		SYSTEM_Module next;
		SYSTEM_ModuleName name;
		LONGINT refcnt;
		SYSTEM_Cmd cmds;
		LONGINT types;
		SYSTEM_EnumProc enumPtrs;
		LONGINT reserved1, reserved2;
	} SYSTEM_ModuleDesc;


export SYSTEM_PTR SYSTEM_modules;
static LONGINT SYSTEM_freeList[10];
static LONGINT SYSTEM_bigBlocks;
export LONGINT SYSTEM_allocated;
static BOOLEAN SYSTEM_firstTry;
static LONGINT SYSTEM_heap, SYSTEM_heapMin, SYSTEM_heapMax;
export LONGINT SYSTEM_heapsize;
static LONGINT SYSTEM_heapMinExpand;
static SYSTEM_FinNode SYSTEM_fin;
export SHORTINT SYSTEM_gclock;
export LONGINT (*SYSTEM_refLibHandle)(LONGINT);

export LONGINT *SYSTEM_ModuleDesc__typ;
export LONGINT *SYSTEM_CmdDesc__typ;
export LONGINT *SYSTEM_FinDesc__typ;
export LONGINT *SYSTEM__1__typ;

static void SYSTEM_CheckFin (void);
static void SYSTEM_ExtendHeap (LONGINT blksz);
export void SYSTEM_FINALL (void);
static void SYSTEM_Finalize (void);
export void SYSTEM_GC (BOOLEAN markStack);
static void SYSTEM_HeapSort (LONGINT n, LONGINT *a, LONGINT a__len);
export void SYSTEM_INCREF (SYSTEM_Module m, LONGINT *imps);
static void SYSTEM_InitHeap (void);
static void SYSTEM_Mark (LONGINT q);
static void SYSTEM_MarkCandidates (LONGINT n, LONGINT *cand, LONGINT cand__len);
static void SYSTEM_MarkP (SYSTEM_PTR p);
export void SYSTEM_MarkStack (void);
export void SYSTEM_MarkStack0 (LONGINT sp);
export SYSTEM_PTR SYSTEM_NEWBLK (LONGINT size);
export SYSTEM_PTR SYSTEM_NEWREC (LONGINT tag);
static LONGINT SYSTEM_NewChunk (LONGINT blksz);
export void SYSTEM_REGCMD (SYSTEM_Module m, SYSTEM_CmdName name, SYSTEM_Command cmd);
export void SYSTEM_REGFIN (SYSTEM_PTR obj, SYSTEM_Finalizer finalize);
export SYSTEM_PTR SYSTEM_REGMOD (SYSTEM_ModuleName name, SYSTEM_EnumProc enumPtrs, LONGINT body, LONGINT imps);
export void SYSTEM_REGTYP (SYSTEM_Module m, LONGINT typ);
static void SYSTEM_Scan (void);
static void SYSTEM_Sift (LONGINT l, LONGINT r, LONGINT *a, LONGINT a__len);

#define SYSTEM_CallMarkStack()	SYSTEM_callMarkStack()
#define SYSTEM_Lock()	Lock
#define SYSTEM_Mainfrm()	SYSTEM_mainfrm
#define SYSTEM_Unlock()	Unlock
#define SYSTEM_free(ptr)	free((void*)ptr)
#define SYSTEM_malloc(size)	(LONGINT)malloc(size)
#define SYSTEM_uLE(x, y)	((unsigned long)x <= (unsigned long)y)
#define SYSTEM_uLT(x, y)	((unsigned long)x < (unsigned long)y)

SYSTEM_PTR SYSTEM_REGMOD (SYSTEM_ModuleName name, SYSTEM_EnumProc enumPtrs, LONGINT body, LONGINT imps)
{
	SYSTEM_Module m;
	if (__STRCMP(name, "SYSTEM") == 0) {
		__SYSNEW(m, 48);
		m->cmds = NIL;
	} else {
		__NEW(m, SYSTEM_ModuleDesc);
	}
	__COPY(name, m->name, 20);
	m->refcnt = 0;
	m->enumPtrs = enumPtrs;
	if (SYSTEM_refLibHandle != NIL) {
		m->reserved1 = (*SYSTEM_refLibHandle)(body);
	} else {
		m->reserved1 = 0;
	}
	m->reserved2 = imps;
	m->next = (SYSTEM_Module)SYSTEM_modules;
	SYSTEM_modules = (SYSTEM_PTR)m;
	return (void*)m;
}

void SYSTEM_REGCMD (SYSTEM_Module m, SYSTEM_CmdName name, SYSTEM_Command cmd)
{
	SYSTEM_Cmd c;
	__NEW(c, SYSTEM_CmdDesc);
	__COPY(name, c->name, 24);
	c->cmd = cmd;
	c->next = m->cmds;
	m->cmds = c;
}

void SYSTEM_REGTYP (SYSTEM_Module m, LONGINT typ)
{
	__PUT(typ, m->types, LONGINT);
	m->types = typ;
}

void SYSTEM_INCREF (SYSTEM_Module m, LONGINT *imps)
{
	LONGINT imp;
	m->refcnt += 1;
	imp = SYSTEM_malloc(8);
	__PUT(imp + 4, m, SYSTEM_Module);
	__PUT(imp, *imps, LONGINT);
	*imps = imp;
}

static LONGINT SYSTEM_NewChunk (LONGINT blksz)
{
	LONGINT chnk, blk, end;
	chnk = SYSTEM_malloc(blksz + 12);
	if (chnk != 0) {
		blk = chnk + 12;
		end = blk + blksz;
		__PUT(chnk + 4, end, LONGINT);
		__PUT(blk, blk + 4, LONGINT);
		__PUT(blk + 4, blksz, LONGINT);
		__PUT(blk + 8, -4, LONGINT);
		__PUT(blk + 12, SYSTEM_bigBlocks, LONGINT);
		SYSTEM_bigBlocks = blk;
		SYSTEM_heapsize += blksz;
		if (SYSTEM_uLT(blk + 4, SYSTEM_heapMin)) {
			SYSTEM_heapMin = blk + 4;
		}
		if (SYSTEM_uLT(SYSTEM_heapMax, end)) {
			SYSTEM_heapMax = end;
		}
	}
	return chnk;
}

static void SYSTEM_ExtendHeap (LONGINT blksz)
{
	LONGINT size, chnk, j, next;
		if (SYSTEM_uLT(SYSTEM_heapMinExpand, blksz)) {
		size = blksz;
	} else {
		size = SYSTEM_heapMinExpand;
	}
	chnk = SYSTEM_NewChunk(size);
	if (chnk != 0) {
		if (SYSTEM_uLT(chnk, SYSTEM_heap)) {
			__PUT(chnk, SYSTEM_heap, LONGINT);
			SYSTEM_heap = chnk;
		} else {
			j = SYSTEM_heap;
			__GET(j, next, LONGINT);
			while (next != 0 && SYSTEM_uLT(next, chnk)) {
				j = next;
				__GET(j, next, LONGINT);
			}
			__PUT(chnk, next, LONGINT);
			__PUT(j, chnk, LONGINT);
		}
	} else if (!SYSTEM_firstTry) {
		SYSTEM_heapMinExpand = 16;
	}
}

SYSTEM_PTR SYSTEM_NEWREC (LONGINT tag)
{
	LONGINT i, i0, di, blksz, restsize, t, adr, end, next, prev;
	SYSTEM_PTR new;
	SYSTEM_Lock();
	__GET(tag, blksz, LONGINT);
		i0 = __LSHR(blksz, 4, LONGINT);
	i = i0;
	if (i < 9) {
		adr = SYSTEM_freeList[i];
		while (adr == 0) {
			i += 1;
			adr = SYSTEM_freeList[i];
		}
	}
	if (i < 9) {
		__GET(adr + 12, next, LONGINT);
		SYSTEM_freeList[i] = next;
		if (i != i0) {
			di = i - i0;
			restsize = __ASHL(di, 4);
			end = adr + restsize;
			__PUT(end + 4, blksz, LONGINT);
			__PUT(end + 8, -4, LONGINT);
			__PUT(end, end + 4, LONGINT);
			__PUT(adr + 4, restsize, LONGINT);
			__PUT(adr + 12, SYSTEM_freeList[di], LONGINT);
			SYSTEM_freeList[di] = adr;
			adr += restsize;
		}
	} else {
		adr = SYSTEM_bigBlocks;
		prev = 0;
		for (;;) {
			if (adr == 0) {
				if (SYSTEM_firstTry) {
					SYSTEM_GC(1);
					blksz += 16;
					t = __LSHR(SYSTEM_allocated + blksz, 6, LONGINT) * 80;
					if (SYSTEM_uLT(SYSTEM_heapsize, t)) {
						SYSTEM_ExtendHeap(t - SYSTEM_heapsize);
					}
					SYSTEM_firstTry = 0;
					new = SYSTEM_NEWREC(tag);
					if (new == NIL) {
						SYSTEM_ExtendHeap(blksz);
						new = SYSTEM_NEWREC(tag);
					}
					SYSTEM_firstTry = 1;
					SYSTEM_Unlock();
					return new;
				} else {
					SYSTEM_Unlock();
					return NIL;
				}
			}
			__GET(adr + 4, t, LONGINT);
			if (SYSTEM_uLE(blksz, t)) {
				break;
			}
			prev = adr;
			__GET(adr + 12, adr, LONGINT);
		}
		restsize = t - blksz;
		end = adr + restsize;
		__PUT(end + 4, blksz, LONGINT);
		__PUT(end + 8, -4, LONGINT);
		__PUT(end, end + 4, LONGINT);
		if (SYSTEM_uLT(144, restsize)) {
			__PUT(adr + 4, restsize, LONGINT);
		} else {
			__GET(adr + 12, next, LONGINT);
			if (prev == 0) {
				SYSTEM_bigBlocks = next;
			} else {
				__PUT(prev + 12, next, LONGINT);
			}
			if (restsize != 0) {
				di = __ASHR(restsize, 4);
				__PUT(adr + 4, restsize, LONGINT);
				__PUT(adr + 12, SYSTEM_freeList[di], LONGINT);
				SYSTEM_freeList[di] = adr;
			}
		}
		adr += restsize;
	}
	i = adr + 16;
	end = adr + blksz;
	while (SYSTEM_uLT(i, end)) {
		__PUT(i, 0, LONGINT);
		__PUT(i + 4, 0, LONGINT);
		__PUT(i + 8, 0, LONGINT);
		__PUT(i + 12, 0, LONGINT);
		i += 16;
	}
	__PUT(adr + 12, 0, LONGINT);
	__PUT(adr, tag, LONGINT);
	__PUT(adr + 4, 0, LONGINT);
	__PUT(adr + 8, 0, LONGINT);
	SYSTEM_allocated += blksz;
	SYSTEM_Unlock();
	return (SYSTEM_PTR)(adr + 4);
}

SYSTEM_PTR SYSTEM_NEWBLK (LONGINT size)
{
	LONGINT blksz, tag;
	SYSTEM_PTR new;
	SYSTEM_Lock();
	blksz = __ASHL(__ASHR(size + 31, 4), 4);
	new = SYSTEM_NEWREC((LONGINT)&blksz);
	tag = ((LONGINT)new + blksz) - 12;
	__PUT(tag - 4, 0, LONGINT);
	__PUT(tag, blksz, LONGINT);
	__PUT(tag + 4, -4, LONGINT);
	__PUT((LONGINT)new - 4, tag, LONGINT);
	SYSTEM_Unlock();
	return new;
}

static void SYSTEM_Mark (LONGINT q)
{
	LONGINT p, tag, fld, n, offset, tagbits;
	if (q != 0) {
		__GET(q - 4, tagbits, LONGINT);
		if (!__ODD(tagbits)) {
			__PUT(q - 4, tagbits + 1, LONGINT);
			p = 0;
			tag = tagbits + 4;
			for (;;) {
				__GET(tag, offset, LONGINT);
				if (offset < 0) {
					__PUT(q - 4, (tag + offset) + 1, LONGINT);
					if (p == 0) {
						break;
					}
					n = q;
					q = p;
					__GET(q - 4, tag, LONGINT);
					tag -= 1;
					__GET(tag, offset, LONGINT);
					fld = q + offset;
					__GET(fld, p, LONGINT);
					__PUT(fld, n, LONGINT);
				} else {
					fld = q + offset;
					__GET(fld, n, LONGINT);
					if (n != 0) {
						__GET(n - 4, tagbits, LONGINT);
						if (!__ODD(tagbits)) {
							__PUT(n - 4, tagbits + 1, LONGINT);
							__PUT(q - 4, tag + 1, LONGINT);
							__PUT(fld, p, LONGINT);
							p = q;
							q = n;
							tag = tagbits;
						}
					}
				}
				tag += 4;
			}
		}
	}
}

static void SYSTEM_MarkP (SYSTEM_PTR p)
{
	SYSTEM_Mark((LONGINT)p);
}

static void SYSTEM_Scan (void)
{
	LONGINT chnk, prevChnk, nextChnk, adr, end, start, tag, i, size, freesize;
	SYSTEM_bigBlocks = 0;
	i = 1;
	while (i < 9) {
		SYSTEM_freeList[i] = 0;
		i += 1;
	}
	freesize = 0;
	SYSTEM_allocated = 0;
	chnk = SYSTEM_heap;
	prevChnk = 0;
	while (chnk != 0) {
		adr = chnk + 12;
		__GET(chnk + 4, end, LONGINT);
		while (SYSTEM_uLT(adr, end)) {
			__GET(adr, tag, LONGINT);
			if (__ODD(tag)) {
				if (freesize != 0) {
					start = adr - freesize;
					__PUT(start, start + 4, LONGINT);
					__PUT(start + 4, freesize, LONGINT);
					__PUT(start + 8, -4, LONGINT);
					i = __LSHR(freesize, 4, LONGINT);
					freesize = 0;
					if (SYSTEM_uLT(i, 9)) {
						__PUT(start + 12, SYSTEM_freeList[i], LONGINT);
						SYSTEM_freeList[i] = start;
					} else {
						__PUT(start + 12, SYSTEM_bigBlocks, LONGINT);
						SYSTEM_bigBlocks = start;
					}
				}
				tag -= 1;
				__PUT(adr, tag, LONGINT);
				__GET(tag, size, LONGINT);
				SYSTEM_allocated += size;
				adr += size;
			} else {
				__GET(tag, size, LONGINT);
				freesize += size;
				adr += size;
			}
		}
		__GET(chnk, nextChnk, LONGINT);
		if (freesize != 0) {
			start = adr - freesize;
			if (start == chnk + 12 && prevChnk != nextChnk) {
				SYSTEM_heapsize -= freesize;
				if (prevChnk == 0) {
					SYSTEM_heap = nextChnk;
					SYSTEM_heapMin = (nextChnk + 12) + 4;
				} else if (nextChnk != 0) {
					__PUT(prevChnk, nextChnk, LONGINT);
				} else {
					__PUT(prevChnk, nextChnk, LONGINT);
					SYSTEM_heapMax = end;
				}
				SYSTEM_free(chnk);
			} else {
				__PUT(start, start + 4, LONGINT);
				__PUT(start + 4, freesize, LONGINT);
				__PUT(start + 8, -4, LONGINT);
				i = __LSHR(freesize, 4, LONGINT);
				if (SYSTEM_uLT(i, 9)) {
					__PUT(start + 12, SYSTEM_freeList[i], LONGINT);
					SYSTEM_freeList[i] = start;
				} else {
					__PUT(start + 12, SYSTEM_bigBlocks, LONGINT);
					SYSTEM_bigBlocks = start;
				}
				prevChnk = chnk;
			}
			freesize = 0;
		} else {
			prevChnk = chnk;
		}
		chnk = nextChnk;
	}
}

static void SYSTEM_Sift (LONGINT l, LONGINT r, LONGINT *a, LONGINT a__len)
{
	LONGINT i, j, x;
	j = l;
	x = a[j];
	for (;;) {
		i = j;
		j = __ASHL(j, 1) + 1;
		if (j < r && SYSTEM_uLT(a[j], a[j + 1])) {
			j += 1;
		}
		if (j > r || SYSTEM_uLE(a[j], x)) {
			break;
		}
		a[i] = a[j];
	}
	a[i] = x;
}

static void SYSTEM_HeapSort (LONGINT n, LONGINT *a, LONGINT a__len)
{
	LONGINT l, r, x;
	l = __ASHR(n, 1);
	r = n - 1;
	while (l > 0) {
		l -= 1;
		SYSTEM_Sift(l, r, (void*)a, a__len);
	}
	while (r > 0) {
		x = a[0];
		a[0] = a[r];
		a[r] = x;
		r -= 1;
		SYSTEM_Sift(l, r, (void*)a, a__len);
	}
}

static void SYSTEM_MarkCandidates (LONGINT n, LONGINT *cand, LONGINT cand__len)
{
	LONGINT chnk, end, adr, tag, next, i, ptr, size;
		chnk = SYSTEM_heap;
	i = 0;
	while (chnk != 0) {
		__GET(chnk + 4, end, LONGINT);
		adr = chnk + 12;
		while (SYSTEM_uLT(adr, end)) {
			__GET(adr, tag, LONGINT);
			if (__ODD(tag)) {
				__GET(tag - 1, size, LONGINT);
				adr += size;
				ptr = adr + 4;
				while (SYSTEM_uLT(cand[i], ptr)) {
					i += 1;
					if (i == n) {
						return;
					}
				}
			} else {
				__GET(tag, size, LONGINT);
				ptr = adr + 4;
				adr += size;
				while (SYSTEM_uLT(cand[i], ptr)) {
					i += 1;
					if (i == n) {
						return;
					}
				}
				if (SYSTEM_uLT(cand[i], adr)) {
					SYSTEM_Mark(ptr);
				}
			}
			if (SYSTEM_uLE(end, cand[i])) {
				adr = end;
			}
		}
		__GET(chnk, chnk, LONGINT);
	}
}

static void SYSTEM_CheckFin (void)
{
	SYSTEM_FinNode n;
	LONGINT tag;
	n = SYSTEM_fin;
	while (n != NIL) {
		__GET(n->obj - 4, tag, LONGINT);
		if (!__ODD(tag)) {
			n->marked = 0;
			SYSTEM_Mark(n->obj);
		} else {
			n->marked = 1;
		}
		n = n->next;
	}
}

static void SYSTEM_Finalize (void)
{
	SYSTEM_FinNode n, prev;
	n = SYSTEM_fin;
	prev = NIL;
	while (n != NIL) {
		if (!n->marked) {
			if (n == SYSTEM_fin) {
				SYSTEM_fin = SYSTEM_fin->next;
			} else {
				prev->next = n->next;
			}
			(*n->finalize)((SYSTEM_PTR)n->obj);
			if (prev == NIL) {
				n = SYSTEM_fin;
			} else {
				n = n->next;
			}
		} else {
			prev = n;
			n = n->next;
		}
	}
}

void SYSTEM_FINALL (void)
{
	SYSTEM_FinNode n;
	while (SYSTEM_fin != NIL) {
		n = SYSTEM_fin;
		SYSTEM_fin = SYSTEM_fin->next;
		(*n->finalize)((SYSTEM_PTR)n->obj);
	}
}

void SYSTEM_MarkStack0 (LONGINT sp)
{
	LONGINT inc, nofcand, p, stack0, ptr;
	struct SYSTEM__1 align;
	LONGINT cand[10000];
	nofcand = 0;
	stack0 = SYSTEM_Mainfrm();
	inc = (LONGINT)&align.p - (LONGINT)&align;
	if (SYSTEM_uLT(stack0, sp)) {
		inc = -inc;
	}
	while (sp != stack0) {
		__GET(sp, p, LONGINT);
		if (SYSTEM_uLE(SYSTEM_heapMin, p) && SYSTEM_uLT(p, SYSTEM_heapMax)) {
			if (nofcand == 10000) {
				SYSTEM_HeapSort(nofcand, (void*)cand, 10000);
				SYSTEM_MarkCandidates(nofcand, (void*)cand, 10000);
				nofcand = 0;
			}
			cand[nofcand] = p;
			nofcand += 1;
		}
		sp += inc;
	}
	if (nofcand > 0) {
		SYSTEM_HeapSort(nofcand, (void*)cand, 10000);
		SYSTEM_MarkCandidates(nofcand, (void*)cand, 10000);
	}
}

void SYSTEM_MarkStack (void)
{
	LONGINT x;
	SYSTEM_MarkStack0((LONGINT)&x);
}

void SYSTEM_GC (BOOLEAN markStack)
{
	SYSTEM_Module m;
	if (SYSTEM_gclock == 0 || SYSTEM_gclock == 1 && !markStack) {
		SYSTEM_Lock();
		m = (SYSTEM_Module)SYSTEM_modules;
		while (m != NIL) {
			if (m->enumPtrs != NIL) {
				(*m->enumPtrs)(SYSTEM_MarkP);
			}
			m = m->next;
		}
		if (markStack) {
			SYSTEM_CallMarkStack();
		}
		SYSTEM_CheckFin();
		SYSTEM_Scan();
		SYSTEM_Finalize();
		SYSTEM_Unlock();
	}
}

void SYSTEM_REGFIN (SYSTEM_PTR obj, SYSTEM_Finalizer finalize)
{
	SYSTEM_FinNode f;
	__NEW(f, SYSTEM_FinDesc);
	f->obj = (LONGINT)obj;
	f->finalize = finalize;
	f->marked = 1;
	f->next = SYSTEM_fin;
	SYSTEM_fin = f;
}

static void SYSTEM_InitHeap (void)
{
	SYSTEM_heap = 0;
	SYSTEM_heapsize = 0;
	SYSTEM_allocated = 0;
	SYSTEM_gclock = 0;
	SYSTEM_heapMin = -1;
	SYSTEM_heapMax = 0;
	SYSTEM_heap = SYSTEM_NewChunk(128000);
	__PUT(SYSTEM_heap, 0, LONGINT);
	SYSTEM_firstTry = 1;
	SYSTEM_freeList[9] = 1;
}

static void EnumPtrs(void (*P)(void*))
{
	P(SYSTEM_modules);
	P(SYSTEM_fin);
}

__TDESC(SYSTEM_ModuleDesc, 1, 2) = {__TDFLDS("ModuleDesc", 48), {0, 28, -12}};
__TDESC(SYSTEM_CmdDesc, 1, 1) = {__TDFLDS("CmdDesc", 32), {0, -8}};
__TDESC(SYSTEM_FinDesc, 1, 1) = {__TDFLDS("FinDesc", 16), {0, -8}};
__TDESC(SYSTEM__1, 1, 1) = {__TDFLDS("", 8), {4, -8}};

export void *SYSTEM__init(void)
{
	__DEFMOD;
	__REGMOD(SYSTEM, EnumPtrs);
	__INITYP(SYSTEM_ModuleDesc, SYSTEM_ModuleDesc, 0);
	__INITYP(SYSTEM_CmdDesc, SYSTEM_CmdDesc, 0);
	__INITYP(SYSTEM_FinDesc, SYSTEM_FinDesc, 0);
	__INITYP(SYSTEM__1, SYSTEM__1, 0);
/* BEGIN */
	SYSTEM_heapMinExpand = 128000;
	__ENDMOD;
}
