/* Ofront 1.4 -xtspkaG */
#include "SYSTEM.h"
#include "Args.h"
#include "Console.h"
#include "Display.h"
#include "Files.h"
#include "Fonts.h"
#include "Input.h"
#include "Kernel.h"
#include "MenuViewers.h"
#include "Modules.h"
#include "Oberon.h"
#include "Out.h"
#include "TextFrames.h"
#include "Texts.h"
#include "Unix.h"
#include "Viewers.h"


static Texts_Writer System_W;
static INTEGER System_trapLevel;


export void System_ChangeDirectory (void);
export void System_ClearLog (void);
export void System_Close (void);
export void System_CloseTrack (void);
export void System_Collect (void);
export void System_Copy (void);
static void System_CopyFile (CHAR *name, LONGINT name__len, Texts_Scanner *S, LONGINT *S__typ);
export void System_CopyFiles (void);
static void System_DeleteFile (CHAR *name, LONGINT name__len);
export void System_DeleteFiles (void);
export void System_Directory (void);
static void System_EndLine (void);
export void System_Execute (void);
export void System_Free (void);
static void System_FreeMod (Texts_Scanner *S, LONGINT *S__typ);
static void System_GetArg (Texts_Scanner *S, LONGINT *S__typ);
export void System_Grow (void);
static LONGINT System_Max (LONGINT i, LONGINT j);
export void System_Open (void);
export void System_OpenLog (void);
export void System_Quit (void);
export void System_Recall (void);
static void System_RenameFile (CHAR *name, LONGINT name__len, Texts_Scanner *S, LONGINT *S__typ);
export void System_RenameFiles (void);
static void System_Scan (INTEGER *pos, CHAR *O, LONGINT O__len, CHAR *path, LONGINT path__len);
export void System_SetColor (void);
export void System_SetFont (void);
export void System_SetOffset (void);
export void System_SetUser (void);
export void System_ShowCommands (void);
export void System_ShowModules (void);
static INTEGER System_SlashPos (CHAR *name, LONGINT name__len);
export void System_Time (void);
export void System_Trap (LONGINT sig, LONGINT siginfo, LONGINT sigctx);
static void System_Trap0 (LONGINT sig, LONGINT siginfo, LONGINT sigctx);
export void System_Watch (void);

#define System_FinalizeAll()	SYSTEM_FINALL()
#define System_Locked()	(SYSTEM_lock > 0)
#define System_assert()	SYSTEM_assert
#define System_halt()	SYSTEM_halt
#define System_resetHalt()	SYSTEM_halt = -128
#define System_setIntd(v)	SYSTEM_interrupted = v

static void System_Trap0 (LONGINT sig, LONGINT siginfo, LONGINT sigctx)
{
	LONGINT addr = 0;
	if (System_Locked()) {
		switch (sig) {
			case 2: case 18: 
				break;
			case 3: case 4: case 7: case 8: case 10: 
			case 11: case 13: 
				Console_String((CHAR*)"received a signal while in critical section (GC or NEW)", (LONGINT)56);
				break;
			default: 
				return;
				break;
		}
	}
	if (System_trapLevel == 0) {
		System_trapLevel = 1;
		switch (sig) {
			case 2: 
				Console_String((CHAR*)"INTERRUPT", (LONGINT)10);
				break;
			case 3: 
				System_FinalizeAll();
				Unix_Exit(0);
				break;
			case 4: 
				switch (System_halt()) {
					case 0: 
						System_resetHalt();
						System_trapLevel = 0;
						Kernel_siglongjmp(&Kernel_trapEnv, Unix_JmpBuf__typ, 1);
						break;
					case -1: 
						Console_String((CHAR*)"ASSERT(", (LONGINT)8);
						Console_Int(System_assert(), 1);
						Console_String((CHAR*)") FAILED", (LONGINT)9);
						break;
					case -2: 
						Console_String((CHAR*)"INDEX OUT OF RANGE", (LONGINT)19);
						break;
					case -3: 
						Console_String((CHAR*)"FUNCTION WITHOUT RETURN", (LONGINT)24);
						break;
					case -4: 
						Console_String((CHAR*)"INVALID CASE", (LONGINT)13);
						break;
					case -5: 
						Console_String((CHAR*)"TYPE GUARD FAILED", (LONGINT)18);
						break;
					case -6: 
						Console_String((CHAR*)"IMPLICIT TYPE GUARD FAILED", (LONGINT)27);
						break;
					case -7: 
						Console_String((CHAR*)"WITH GUARD FAILED", (LONGINT)18);
						break;
					case -8: 
						Console_String((CHAR*)"VALUE OUT OF RANGE", (LONGINT)19);
						break;
					case -9: 
						System_setIntd(0);
						Console_String((CHAR*)"DELAYED INTERRUPT", (LONGINT)18);
						break;
					default: 
						if (System_halt() > 0 && System_halt() < 256) {
							Console_String((CHAR*)"HALT(", (LONGINT)6);
							Console_Int(System_halt(), 1);
							Console_Char(')');
						} else {
							Console_String((CHAR*)"ILLEGAL INSTRUCTION", (LONGINT)20);
						}
						break;
				}
				System_resetHalt();
				break;
			case 7: case 10: 
				Console_String((CHAR*)"BUS ERROR", (LONGINT)10);
				break;
			case 8: 
				Console_String((CHAR*)"ARITHMETIC EXCEPTION", (LONGINT)21);
				break;
			case 11: 
				Console_String((CHAR*)"SEGMENTATION VIOLATION", (LONGINT)23);
				break;
			case 13: 
				Console_String((CHAR*)"UNCONNECTED PIPE", (LONGINT)17);
				break;
			case 18: 
				return;
				break;
			default: 
				Console_String((CHAR*)"SIGNAL ", (LONGINT)8);
				Console_Int(sig, 0);
				break;
		}
	} else {
		Console_String((CHAR*)"received a signal while handing a signal: sig=", (LONGINT)47);
		Console_Int(sig, 0);
	}
	Console_Ln();
	Unix_Exit(0);
}

void System_Trap (LONGINT sig, LONGINT siginfo, LONGINT sigctx)
{
	LONGINT addr = 0;
	if (!Oberon_LoopStarted) {
		System_Trap0(sig, siginfo, sigctx);
	}
	if (System_Locked()) {
		switch (sig) {
			case 2: case 13: case 18: 
				break;
			case 3: case 4: case 7: case 8: case 10: 
			case 11: 
				System_Trap0(sig, siginfo, sigctx);
				break;
			default: 
				return;
				break;
		}
	}
	if (System_trapLevel == 0) {
		System_trapLevel = 1;
		switch (sig) {
			case 2: 
				if (System_Locked()) {
					System_setIntd(1);
					System_trapLevel = 0;
					return;
				} else {
					Out_String((CHAR*)"INTERRUPT", (LONGINT)10);
				}
				break;
			case 3: 
				System_FinalizeAll();
				Unix_Exit(0);
				break;
			case 4: 
				switch (System_halt()) {
					case 0: 
						System_resetHalt();
						System_trapLevel = 0;
						Kernel_siglongjmp(&Kernel_trapEnv, Unix_JmpBuf__typ, 1);
						break;
					case -1: 
						Out_String((CHAR*)"ASSERT(", (LONGINT)8);
						Out_Int(System_assert(), 1);
						Out_String((CHAR*)") FAILED", (LONGINT)9);
						break;
					case -2: 
						Out_String((CHAR*)"INDEX OUT OF RANGE", (LONGINT)19);
						break;
					case -3: 
						Out_String((CHAR*)"FUNCTION WITHOUT RETURN", (LONGINT)24);
						break;
					case -4: 
						Out_String((CHAR*)"INVALID CASE", (LONGINT)13);
						break;
					case -5: 
						Out_String((CHAR*)"TYPE GUARD FAILED", (LONGINT)18);
						break;
					case -6: 
						Out_String((CHAR*)"IMPLICIT TYPE GUARD FAILED", (LONGINT)27);
						break;
					case -7: 
						Out_String((CHAR*)"WITH GUARD FAILED", (LONGINT)18);
						break;
					case -8: 
						Out_String((CHAR*)"VALUE OUT OF RANGE", (LONGINT)19);
						break;
					case -9: 
						System_setIntd(0);
						Out_String((CHAR*)"DELAYED INTERRUPT", (LONGINT)18);
						break;
					default: 
						if (System_halt() > 0 && System_halt() < 256) {
							Out_String((CHAR*)"HALT(", (LONGINT)6);
							Out_Int(System_halt(), 1);
							Out_Char(')');
						} else {
							Out_String((CHAR*)"ILLEGAL INSTRUCTION", (LONGINT)20);
						}
						break;
				}
				System_resetHalt();
				break;
			case 7: case 10: 
				Out_String((CHAR*)"BUS ERROR", (LONGINT)10);
				break;
			case 8: 
				Out_String((CHAR*)"ARITHMETIC EXCEPTION", (LONGINT)21);
				break;
			case 11: 
				addr = Unix_SiginfoAddr(siginfo);
				if (addr != 0) {
					Viewers_CheckUnload((Display_Handler)addr);
				}
				Out_String((CHAR*)"SEGMENTATION VIOLATION", (LONGINT)23);
				break;
			case 13: 
				if (System_Locked()) {
					System_FinalizeAll();
					Unix_Exit(0);
				} else {
					Out_String((CHAR*)"UNCONNECTED PIPE", (LONGINT)17);
				}
				break;
			case 18: 
				return;
				break;
			default: 
				Out_String((CHAR*)"SIGNAL ", (LONGINT)8);
				Out_Int(sig, 0);
				break;
		}
		Out_Ln();
	} else {
		Console_String((CHAR*)"received a signal while handing a signal: sig=", (LONGINT)47);
		Console_Int(sig, 0);
		Console_Ln();
	}
	System_trapLevel = 0;
	Kernel_siglongjmp(&Kernel_trapEnv, Unix_JmpBuf__typ, 1);
}

static LONGINT System_Max (LONGINT i, LONGINT j)
{
	if (i >= j) {
		return i;
	} else {
		return j;
	}
	__RETCHK;
}

void System_Open (void)
{
	Oberon_ParList par = NIL;
	Texts_Text T = NIL;
	Texts_Scanner S = {0};
	Viewers_Viewer V = NIL;
	INTEGER X = 0, Y = 0;
	LONGINT beg = 0, end = 0, time = 0;
	par = Oberon_Par;
	Texts_OpenScanner(&S, Texts_Scanner__typ, par->text, par->pos);
	Texts_Scan(&S, Texts_Scanner__typ);
	if (S.class == 6 && S.c == '^' || S.line != 0) {
		Oberon_GetSelection(&T, &beg, &end, &time);
		if (time >= 0) {
			Texts_OpenScanner(&S, Texts_Scanner__typ, T, beg);
			Texts_Scan(&S, Texts_Scanner__typ);
		}
	}
	if (__IN(S.class, 0x06)) {
		Oberon_AllocateSystemViewer(par->vwr->X, &X, &Y);
		V = (Viewers_Viewer)MenuViewers_New((void*)TextFrames_NewMenu(S.s, 256, (CHAR*)"^System.Menu.Text", (LONGINT)18), (void*)TextFrames_NewText(TextFrames_Text(S.s, 256), 0), TextFrames_menuH, X, Y);
	}
}

void System_OpenLog (void)
{
	Viewers_Viewer logV = NIL;
	INTEGER X = 0, Y = 0;
	Oberon_AllocateSystemViewer(Oberon_Par->vwr->X, &X, &Y);
	logV = (Viewers_Viewer)MenuViewers_New((void*)TextFrames_NewMenu((CHAR*)"System.Log", (LONGINT)11, (CHAR*)"^Log.Menu.Text", (LONGINT)15), (void*)TextFrames_NewText(Oberon_Log, System_Max(0, Oberon_Log->len - 200)), TextFrames_menuH, X, Y);
}

void System_ClearLog (void)
{
	Texts_Delete(Oberon_Log, 0, Oberon_Log->len);
	Texts_OpenWriter(&System_W, Texts_Writer__typ);
}

void System_Close (void)
{
	Oberon_ParList par = NIL;
	Viewers_Viewer V = NIL;
	Viewers_ViewerMsg M = {0};
	par = Oberon_Par;
	if (par->frame == par->vwr->dsc) {
		V = par->vwr;
	} else {
		V = Oberon_MarkedViewer();
	}
	Viewers_Close(V);
}

void System_CloseTrack (void)
{
	Viewers_Viewer V = NIL;
	V = Oberon_MarkedViewer();
	Viewers_CloseTrack(V->X);
}

void System_Recall (void)
{
	Viewers_Viewer V = NIL;
	Viewers_ViewerMsg M = {0};
	Viewers_Recall(&V);
	if (V != NIL && V->state == 0) {
		Viewers_Open(V, V->X, V->Y + V->H);
		M.id = 0;
		(*V->handle)((void*)V, (void*)&M, Viewers_ViewerMsg__typ);
	}
}

void System_Copy (void)
{
	Viewers_Viewer V = NIL, V1 = NIL;
	Oberon_CopyMsg M = {0};
	Viewers_ViewerMsg N = {0};
	V = Oberon_Par->vwr;
	(*V->handle)((void*)V, (void*)&M, Oberon_CopyMsg__typ);
	V1 = __GUARDP(M.F, Viewers_ViewerDesc, 1);
	Viewers_Open(V1, V->X, V->Y + __ASHR(V->H, 1));
	N.id = 0;
	(*V1->handle)((void*)V1, (void*)&N, Viewers_ViewerMsg__typ);
}

void System_Grow (void)
{
	Viewers_Viewer V = NIL, V1 = NIL;
	Oberon_CopyMsg M = {0};
	Viewers_ViewerMsg N = {0};
	INTEGER DW = 0, DH = 0;
	V = Oberon_Par->vwr;
	DW = Oberon_DisplayWidth(V->X);
	DH = Oberon_DisplayHeight(V->X);
	if (V->H < DH - Viewers_minH) {
		Oberon_OpenTrack(V->X, V->W);
	} else if (V->W < DW) {
		Oberon_OpenTrack(Oberon_UserTrack(V->X), DW);
	}
	if (V->H < DH - Viewers_minH || V->W < DW) {
		(*V->handle)((void*)V, (void*)&M, Oberon_CopyMsg__typ);
		V1 = __GUARDP(M.F, Viewers_ViewerDesc, 1);
		Viewers_Open(V1, V->X, DH);
		N.id = 0;
		(*V1->handle)((void*)V1, (void*)&N, Viewers_ViewerMsg__typ);
	}
}

static void System_GetArg (Texts_Scanner *S, LONGINT *S__typ)
{
	Texts_Text T = NIL;
	LONGINT beg = 0, end = 0, time = 0;
	Texts_OpenScanner(&*S, S__typ, Oberon_Par->text, Oberon_Par->pos);
	Texts_Scan(&*S, S__typ);
	if ((*S).class == 6 && (*S).c == '^') {
		Oberon_GetSelection(&T, &beg, &end, &time);
		if (time >= 0) {
			Texts_OpenScanner(&*S, S__typ, T, beg);
			Texts_Scan(&*S, S__typ);
		}
	}
}

static void System_EndLine (void)
{
	Texts_WriteLn(&System_W, Texts_Writer__typ);
	Texts_Append(Oberon_Log, System_W.buf);
}

void System_SetFont (void)
{
	Texts_Scanner S = {0};
	System_GetArg(&S, Texts_Scanner__typ);
	if (__IN(S.class, 0x06)) {
		Oberon_SetFont(Fonts_This(S.s, 256));
	}
}

void System_SetColor (void)
{
	Texts_Scanner S = {0};
	System_GetArg(&S, Texts_Scanner__typ);
	if (S.class == 3) {
		Oberon_SetColor((int)S.i);
	}
}

void System_SetOffset (void)
{
	Texts_Scanner S = {0};
	System_GetArg(&S, Texts_Scanner__typ);
	if (S.class == 3) {
		Oberon_SetOffset((int)S.i);
	}
}

void System_Time (void)
{
	LONGINT t = 0, d = 0;
	Texts_WriteString(&System_W, Texts_Writer__typ, (CHAR*)"System.Time", (LONGINT)12);
	Oberon_GetClock(&t, &d);
	Texts_WriteDate(&System_W, Texts_Writer__typ, t, d);
	Texts_WriteLn(&System_W, Texts_Writer__typ);
	Texts_Append(Oberon_Log, System_W.buf);
}

void System_Watch (void)
{
	Texts_WriteString(&System_W, Texts_Writer__typ, (CHAR*)"System.Watch", (LONGINT)13);
	Texts_WriteLn(&System_W, Texts_Writer__typ);
	Texts_WriteInt(&System_W, Texts_Writer__typ, Kernel_allocated(), 0);
	Texts_WriteString(&System_W, Texts_Writer__typ, (CHAR*)" bytes allocated from ", (LONGINT)23);
	Texts_WriteInt(&System_W, Texts_Writer__typ, Kernel_heapsize(), 0);
	Texts_WriteLn(&System_W, Texts_Writer__typ);
	Texts_WriteInt(&System_W, Texts_Writer__typ, Kernel_nofiles, 0);
	Texts_WriteString(&System_W, Texts_Writer__typ, (CHAR*)" file(s) open", (LONGINT)14);
	Texts_WriteLn(&System_W, Texts_Writer__typ);
	Texts_Append(Oberon_Log, System_W.buf);
}

void System_Collect (void)
{
	Oberon_Collect(0);
}

void System_SetUser (void)
{
	INTEGER i = 0;
	CHAR ch = 0;
	CHAR user[8];
	CHAR password[16];
	i = 0;
	Input_Read(&ch);
	while (ch != '/' && i < 7) {
		user[__X(i, 8)] = ch;
		i += 1;
		Input_Read(&ch);
	}
	user[__X(i, 8)] = 0x00;
	i = 0;
	Input_Read(&ch);
	while (ch > ' ' && i < 15) {
		password[__X(i, 16)] = ch;
		i += 1;
		Input_Read(&ch);
	}
	password[__X(i, 16)] = 0x00;
	Oberon_SetUser((void*)user, 8, (void*)password, 16);
}

static void System_FreeMod (Texts_Scanner *S, LONGINT *S__typ)
{
	Out_String((*S).s, 256);
	Out_String((CHAR*)" unloading", (LONGINT)11);
	if ((*S).nextCh != '*') {
		Modules_Free((*S).s, 256, 0);
	} else {
		Modules_Free((*S).s, 256, 1);
		Texts_Scan(&*S, S__typ);
		Out_String((CHAR*)" all", (LONGINT)5);
	}
	if (Modules_res != 0) {
		Out_String((CHAR*)" failed: ", (LONGINT)10);
		Out_String(Modules_resMsg, 256);
		Modules_res = 0;
	}
	Out_Ln();
}

void System_Free (void)
{
	Oberon_ParList par = NIL;
	Texts_Text T = NIL;
	Texts_Scanner S = {0};
	Viewers_Viewer V = NIL;
	LONGINT beg = 0, end = 0, time = 0;
	par = Oberon_Par;
	Out_String((CHAR*)"System.Free", (LONGINT)12);
	Out_Ln();
	Texts_OpenScanner(&S, Texts_Scanner__typ, par->text, par->pos);
	Texts_Scan(&S, Texts_Scanner__typ);
	while (S.class == 1) {
		System_FreeMod(&S, Texts_Scanner__typ);
		Texts_Scan(&S, Texts_Scanner__typ);
	}
	if (S.class == 6 && S.c == '^') {
		Oberon_GetSelection(&T, &beg, &end, &time);
		if (time >= 0) {
			Texts_OpenScanner(&S, Texts_Scanner__typ, T, beg);
			Texts_Scan(&S, Texts_Scanner__typ);
			if (S.class == 1) {
				System_FreeMod(&S, Texts_Scanner__typ);
			}
		}
	}
}

void System_ShowModules (void)
{
	Texts_Text T = NIL;
	Viewers_Viewer V = NIL;
	Modules_Module M = NIL;
	INTEGER X = 0, Y = 0;
	T = TextFrames_Text((CHAR*)"", (LONGINT)1);
	Oberon_AllocateSystemViewer(Oberon_Par->vwr->X, &X, &Y);
	V = (Viewers_Viewer)MenuViewers_New((void*)TextFrames_NewMenu((CHAR*)"System.ShowModules", (LONGINT)19, (CHAR*)"^System.Menu.Text", (LONGINT)18), (void*)TextFrames_NewText(T, 0), TextFrames_menuH, X, Y);
	M = Modules_modules();
	while (M != NIL) {
		Texts_WriteString(&System_W, Texts_Writer__typ, M->name, 20);
		Texts_WriteInt(&System_W, Texts_Writer__typ, M->refcnt, 5);
		Texts_WriteLn(&System_W, Texts_Writer__typ);
		M = M->next;
	}
	Texts_Append(T, System_W.buf);
}

void System_ShowCommands (void)
{
	Modules_Module M = NIL;
	Texts_Scanner S = {0};
	LONGINT beg = 0, end = 0, time = 0, i = 0;
	Texts_Text T = NIL;
	Viewers_Viewer V = NIL;
	INTEGER X = 0, Y = 0;
	Modules_Cmd cmd = NIL;
	Texts_OpenScanner(&S, Texts_Scanner__typ, Oberon_Par->text, Oberon_Par->pos);
	Texts_Scan(&S, Texts_Scanner__typ);
	if (S.class == 6 && S.c == '^') {
		Oberon_GetSelection(&T, &beg, &end, &time);
		if (time >= 0) {
			Texts_OpenScanner(&S, Texts_Scanner__typ, T, beg);
			Texts_Scan(&S, Texts_Scanner__typ);
		}
	}
	if (__IN(S.class, 0x06)) {
		i = 0;
		while (S.s[__X(i, 256)] >= '0') {
			i += 1;
		}
		S.s[__X(i, 256)] = 0x00;
		M = Modules_ThisMod(S.s, 256);
		if (M != NIL) {
			Oberon_AllocateSystemViewer(Oberon_Par->vwr->X, &X, &Y);
			T = TextFrames_Text((CHAR*)"", (LONGINT)1);
			V = (Viewers_Viewer)MenuViewers_New((void*)TextFrames_NewMenu((CHAR*)"System.ShowCommands", (LONGINT)20, (CHAR*)"^System.Menu.Text", (LONGINT)18), (void*)TextFrames_NewText(T, 0), TextFrames_menuH, X, Y);
			cmd = M->cmds;
			while (cmd != NIL) {
				Texts_WriteString(&System_W, Texts_Writer__typ, M->name, 20);
				Texts_Write(&System_W, Texts_Writer__typ, '.');
				Texts_WriteString(&System_W, Texts_Writer__typ, cmd->name, 24);
				Texts_WriteLn(&System_W, Texts_Writer__typ);
				cmd = cmd->next;
			}
			Texts_Append(T, System_W.buf);
		}
	}
}

void System_ChangeDirectory (void)
{
	Oberon_ParList par = NIL;
	Texts_Text T = NIL;
	Texts_Scanner S = {0};
	INTEGER res = 0;
	LONGINT beg = 0, end = 0, time = 0;
	Texts_OpenScanner(&S, Texts_Scanner__typ, Oberon_Par->text, Oberon_Par->pos);
	Texts_Scan(&S, Texts_Scanner__typ);
	if (S.class == 6 && S.c == '^' || S.line != 0) {
		Oberon_GetSelection(&T, &beg, &end, &time);
		if (time >= 0) {
			Texts_OpenScanner(&S, Texts_Scanner__typ, T, beg);
			Texts_Scan(&S, Texts_Scanner__typ);
		}
	}
	if (__IN(S.class, 0x06) && S.line == 0) {
		Texts_WriteString(&System_W, Texts_Writer__typ, (CHAR*)"System.ChangeDirectory ", (LONGINT)24);
		Texts_WriteString(&System_W, Texts_Writer__typ, S.s, 256);
		Files_ChangeDirectory(S.s, 256, &res);
		if (res != 0) {
			Texts_WriteString(&System_W, Texts_Writer__typ, (CHAR*)"  -- failed", (LONGINT)12);
		}
		Texts_WriteLn(&System_W, Texts_Writer__typ);
		Texts_Append(Oberon_Log, System_W.buf);
	}
}

static void System_CopyFile (CHAR *name, LONGINT name__len, Texts_Scanner *S, LONGINT *S__typ)
{
	Files_File f = NIL, g = NIL;
	Files_Rider Rf = {0}, Rg = {0};
	CHAR ch = 0;
	__DUP(name, name__len, CHAR);
	Texts_Scan(&*S, S__typ);
	if ((*S).class == 6 && (*S).c == '=') {
		Texts_Scan(&*S, S__typ);
		if ((*S).class == 6 && (*S).c == '>') {
			Texts_Scan(&*S, S__typ);
			if (__IN((*S).class, 0x06)) {
				Texts_WriteString(&System_W, Texts_Writer__typ, name, name__len);
				Texts_WriteString(&System_W, Texts_Writer__typ, (CHAR*)" => ", (LONGINT)5);
				Texts_WriteString(&System_W, Texts_Writer__typ, (*S).s, 256);
				Texts_WriteString(&System_W, Texts_Writer__typ, (CHAR*)" copying", (LONGINT)9);
				Texts_Append(Oberon_Log, System_W.buf);
				f = Files_Old(name, name__len);
				if (f != NIL) {
					g = Files_New((*S).s, 256);
					Files_Set(&Rf, Files_Rider__typ, f, 0);
					Files_Set(&Rg, Files_Rider__typ, g, 0);
					Files_Read(&Rf, Files_Rider__typ, (void*)&ch);
					while (!Rf.eof) {
						Files_Write(&Rg, Files_Rider__typ, ch);
						Files_Read(&Rf, Files_Rider__typ, (void*)&ch);
					}
					Files_Register(g);
				} else {
					Texts_WriteString(&System_W, Texts_Writer__typ, (CHAR*)" failed", (LONGINT)8);
				}
				System_EndLine();
			}
		}
	}
	__DEL(name);
}

void System_CopyFiles (void)
{
	Texts_Scanner S = {0};
	System_GetArg(&S, Texts_Scanner__typ);
	Texts_WriteString(&System_W, Texts_Writer__typ, (CHAR*)"System.CopyFiles", (LONGINT)17);
	System_EndLine();
	while (__IN(S.class, 0x06)) {
		System_CopyFile(S.s, 256, &S, Texts_Scanner__typ);
		Texts_Scan(&S, Texts_Scanner__typ);
	}
}

static void System_RenameFile (CHAR *name, LONGINT name__len, Texts_Scanner *S, LONGINT *S__typ)
{
	INTEGER res = 0;
	__DUP(name, name__len, CHAR);
	Texts_Scan(&*S, S__typ);
	if ((*S).class == 6 && (*S).c == '=') {
		Texts_Scan(&*S, S__typ);
		if ((*S).class == 6 && (*S).c == '>') {
			Texts_Scan(&*S, S__typ);
			if (__IN((*S).class, 0x06)) {
				Texts_WriteString(&System_W, Texts_Writer__typ, name, name__len);
				Texts_WriteString(&System_W, Texts_Writer__typ, (CHAR*)" => ", (LONGINT)5);
				Texts_WriteString(&System_W, Texts_Writer__typ, (*S).s, 256);
				Texts_WriteString(&System_W, Texts_Writer__typ, (CHAR*)" renaming", (LONGINT)10);
				Files_Rename(name, name__len, (*S).s, 256, &res);
				if (res > 1) {
					Texts_WriteString(&System_W, Texts_Writer__typ, (CHAR*)" failed", (LONGINT)8);
				}
				System_EndLine();
			}
		}
	}
	__DEL(name);
}

void System_RenameFiles (void)
{
	Texts_Scanner S = {0};
	System_GetArg(&S, Texts_Scanner__typ);
	Texts_WriteString(&System_W, Texts_Writer__typ, (CHAR*)"System.RenameFiles", (LONGINT)19);
	System_EndLine();
	while (__IN(S.class, 0x06)) {
		System_RenameFile(S.s, 256, &S, Texts_Scanner__typ);
		Texts_Scan(&S, Texts_Scanner__typ);
	}
}

static void System_DeleteFile (CHAR *name, LONGINT name__len)
{
	INTEGER res = 0;
	Texts_WriteString(&System_W, Texts_Writer__typ, name, name__len);
	Texts_WriteString(&System_W, Texts_Writer__typ, (CHAR*)" deleting", (LONGINT)10);
	Files_Delete(name, name__len, &res);
	if (res != 0) {
		Texts_WriteString(&System_W, Texts_Writer__typ, (CHAR*)" failed", (LONGINT)8);
	}
	System_EndLine();
}

void System_DeleteFiles (void)
{
	Texts_Scanner S = {0};
	System_GetArg(&S, Texts_Scanner__typ);
	Texts_WriteString(&System_W, Texts_Writer__typ, (CHAR*)"System.DeleteFiles", (LONGINT)19);
	System_EndLine();
	while (__IN(S.class, 0x06)) {
		System_DeleteFile((void*)S.s, 256);
		Texts_Scan(&S, Texts_Scanner__typ);
	}
}

void System_Quit (void)
{
	LONGINT res = 0;
	res = Unix_Kill(Unix_Getpid(), 3);
}

void System_Execute (void)
{
	Texts_Text T = NIL;
	Texts_Reader R = {0};
	Viewers_Viewer V = NIL;
	Texts_Buffer B = NIL;
	LONGINT i = 0, beg = 0, end = 0, time = 0, status = 0;
	CHAR cmd[4096];
	INTEGER X = 0, Y = 0;
	CHAR ch = 0;
	Oberon_AllocateSystemViewer(Oberon_Par->vwr->X, &X, &Y);
	Texts_OpenReader(&R, Texts_Reader__typ, Oberon_Par->text, Oberon_Par->pos);
	i = 0;
	cmd[0] = 0x00;
	Texts_Read(&R, Texts_Reader__typ, &ch);
	while (ch == ' ') {
		Texts_Read(&R, Texts_Reader__typ, &ch);
	}
	while (ch >= ' ' && ch != '^') {
		cmd[__X(i, 4096)] = ch;
		i += 1;
		Texts_Read(&R, Texts_Reader__typ, &ch);
	}
	if (i == 0 || ch == '^') {
		Oberon_GetSelection(&T, &beg, &end, &time);
		if (time >= 0) {
			Texts_OpenReader(&R, Texts_Reader__typ, T, beg);
			Texts_Read(&R, Texts_Reader__typ, &ch);
			while (Texts_Pos(&R, Texts_Reader__typ) <= end) {
				if (ch == 0x0d) {
					ch = ' ';
				}
				cmd[__X(i, 4096)] = ch;
				i += 1;
				Texts_Read(&R, Texts_Reader__typ, &ch);
			}
		}
	}
	cmd[__X(i, 4096)] = 0x00;
	Oberon_ExecShell(cmd, 4096, &status, &B);
	if (B->len > 0) {
		T = TextFrames_Text((CHAR*)"", (LONGINT)1);
		Texts_Append(T, B);
		V = (Viewers_Viewer)MenuViewers_New((void*)TextFrames_NewMenu((CHAR*)"System.Execute", (LONGINT)15, (CHAR*)"^System.Menu.Text", (LONGINT)18), (void*)TextFrames_NewText(T, 0), TextFrames_menuH, X, Y);
	}
}

static void System_Scan (INTEGER *pos, CHAR *O, LONGINT O__len, CHAR *path, LONGINT path__len)
{
	INTEGER i = 0;
	i = 0;
	while (O[__X(*pos, O__len)] == ' ' || O[__X(*pos, O__len)] == ':') {
		*pos += 1;
	}
	while (O[__X(*pos, O__len)] > ' ' && O[__X(*pos, O__len)] != ':') {
		path[__X(i, path__len)] = O[__X(*pos, O__len)];
		i += 1;
		*pos += 1;
	}
	while (O[__X(*pos, O__len)] == ' ' || O[__X(*pos, O__len)] == ':') {
		*pos += 1;
	}
	path[__X(i, path__len)] = 0x00;
}

static INTEGER System_SlashPos (CHAR *name, LONGINT name__len)
{
	INTEGER i = 0;
	i = 0;
	while (name[__X(i, name__len)] != 0x00) {
		i += 1;
	}
	while (i >= 0 && name[__X(i, name__len)] != '/') {
		i -= 1;
	}
	return i;
}

void System_Directory (void)
{
	Texts_Text T = NIL;
	CHAR pat[1024], OBERON[1024], path[1024];
	INTEGER i = 0, opos = 0, slash = 0;
	CHAR ch = 0;
	Texts_Reader R = {0};
	LONGINT t = 0, beg = 0, end = 0;
	i = 0;
	Texts_OpenReader(&R, Texts_Reader__typ, Oberon_Par->text, Oberon_Par->pos);
	Texts_Read(&R, Texts_Reader__typ, &ch);
	while (ch == ' ' || ch == 0x09) {
		Texts_Read(&R, Texts_Reader__typ, &ch);
	}
	while (ch > ' ') {
		pat[__X(i, 1024)] = ch;
		i += 1;
		Texts_Read(&R, Texts_Reader__typ, &ch);
	}
	pat[__X(i, 1024)] = 0x00;
	if (pat[0] == '^') {
		Oberon_GetSelection(&T, &beg, &end, &t);
		if (t >= 0) {
			Texts_OpenReader(&R, Texts_Reader__typ, T, beg);
			Texts_Read(&R, Texts_Reader__typ, &ch);
			while (ch == ' ' || ch == 0x09) {
				Texts_Read(&R, Texts_Reader__typ, &ch);
			}
			i = 0;
			while (ch > ' ') {
				pat[__X(i, 1024)] = ch;
				i += 1;
				Texts_Read(&R, Texts_Reader__typ, &ch);
			}
			pat[__X(i, 1024)] = 0x00;
		}
	}
	if (pat[0] != 0x00) {
		T = TextFrames_Text((CHAR*)"", (LONGINT)1);
		slash = System_SlashPos((void*)pat, 1024);
		if (slash >= 0) {
			Texts_WriteString(&System_W, Texts_Writer__typ, (CHAR*)"ls -d ", (LONGINT)7);
			if (slash == 1 && pat[0] == '.') {
				pat[0] = ' ';
				pat[1] = ' ';
			}
			Texts_WriteString(&System_W, Texts_Writer__typ, pat, 1024);
		} else {
			OBERON[0] = 0x00;
			opos = 0;
			Args_GetEnv((CHAR*)"OBERON", (LONGINT)7, (void*)OBERON, 1024);
			while (OBERON[__X(opos, 1024)] != 0x00) {
				Texts_WriteString(&System_W, Texts_Writer__typ, (CHAR*)"ls -d ", (LONGINT)7);
				System_Scan(&opos, (void*)OBERON, 1024, (void*)path, 1024);
				if (__STRCMP(path, ".") != 0) {
					Texts_WriteString(&System_W, Texts_Writer__typ, path, 1024);
					Texts_Write(&System_W, Texts_Writer__typ, '/');
				}
				Texts_WriteString(&System_W, Texts_Writer__typ, pat, 1024);
				Texts_WriteString(&System_W, Texts_Writer__typ, (CHAR*)" 2>/dev/null;", (LONGINT)14);
			}
		}
		Texts_Append(T, System_W.buf);
		Oberon_Par->text = T;
		Oberon_Par->pos = 0;
		System_Execute();
	}
}

static void EnumPtrs(void (*P)(void*))
{
	__ENUMR(&System_W, Texts_Writer__typ, 36, 1, P);
}


export void *System__init(void)
{
	__DEFMOD;
	__IMPORT(Args);
	__IMPORT(Console);
	__IMPORT(Display);
	__IMPORT(Files);
	__IMPORT(Fonts);
	__IMPORT(Input);
	__IMPORT(Kernel);
	__IMPORT(MenuViewers);
	__IMPORT(Modules);
	__IMPORT(Oberon);
	__IMPORT(Out);
	__IMPORT(TextFrames);
	__IMPORT(Texts);
	__IMPORT(Unix);
	__IMPORT(Viewers);
	__REGMOD(System, EnumPtrs);
	__REGCMD("ChangeDirectory", System_ChangeDirectory);
	__REGCMD("ClearLog", System_ClearLog);
	__REGCMD("Close", System_Close);
	__REGCMD("CloseTrack", System_CloseTrack);
	__REGCMD("Collect", System_Collect);
	__REGCMD("Copy", System_Copy);
	__REGCMD("CopyFiles", System_CopyFiles);
	__REGCMD("DeleteFiles", System_DeleteFiles);
	__REGCMD("Directory", System_Directory);
	__REGCMD("Execute", System_Execute);
	__REGCMD("Free", System_Free);
	__REGCMD("Grow", System_Grow);
	__REGCMD("Open", System_Open);
	__REGCMD("OpenLog", System_OpenLog);
	__REGCMD("Quit", System_Quit);
	__REGCMD("Recall", System_Recall);
	__REGCMD("RenameFiles", System_RenameFiles);
	__REGCMD("SetColor", System_SetColor);
	__REGCMD("SetFont", System_SetFont);
	__REGCMD("SetOffset", System_SetOffset);
	__REGCMD("SetUser", System_SetUser);
	__REGCMD("ShowCommands", System_ShowCommands);
	__REGCMD("ShowModules", System_ShowModules);
	__REGCMD("Time", System_Time);
	__REGCMD("Watch", System_Watch);
/* BEGIN */
	Texts_OpenWriter(&System_W, Texts_Writer__typ);
	System_trapLevel = 0;
	Unix_SigSetHandler(2, System_Trap);
	Unix_SigSetHandler(3, System_Trap);
	Unix_SigSetHandler(4, System_Trap);
	Unix_SigSetHandler(7, System_Trap);
	Unix_SigSetHandler(8, System_Trap);
	Unix_SigSetHandler(10, System_Trap);
	Unix_SigSetHandler(11, System_Trap);
	Unix_SigSetHandler(13, System_Trap);
	if (Kernel_LIB[0] == 0x00) {
		__MOVE("OberonV4", Kernel_LIB, 9);
	}
	__ENDMOD;
}
